/*
  Copyright (C) 2002-2017 CERN for the benefit of the ATLAS collaboration
*/

///////////////////////////////////////////////////////////////////
// CaloHit.h
//   Header file for class CaloHit
///////////////////////////////////////////////////////////////////
// Class for Ecal hits
///////////////////////////////////////////////////////////////////

#ifndef CALOSIMEVENT_CALOHIT_H
#define CALOSIMEVENT_CALOHIT_H

// Data members classes
#include "CLHEP/Geometry/Point3D.h"
#include "GeneratorObjects/HepMcParticleLink.h"

class CaloHit  {

  ///////////////////////////////////////////////////////////////////
  // Public methods:
  ///////////////////////////////////////////////////////////////////
public:

  // Constructor with parameters:
  //   local start position of the energy deposit
  //   local end position of the energy deposit
  //   deposited energy
  //   time of energy deposition
  //   number of track which released this energy
  //
  CaloHit(const HepGeom::Point3D<double> &localStartPosition,
           const HepGeom::Point3D<double> &localEndPosition,
           const double energyLoss,
           const double meanTime,
           const int trackNumber,
           const unsigned int id);

  CaloHit(const HepGeom::Point3D<double> &localStartPosition,
           const HepGeom::Point3D<double> &localEndPosition,
           const double energyLoss,
           const double meanTime,
           const int trackNumber,
           const int, const int);
  // Constructor with parameters:
  //   local start position of the energy deposit
  //   local end position of the energy deposit
  //   deposited energy
  //   time of energy deposition
  //   link to particle which released this energy
  //
  CaloHit(const HepGeom::Point3D<double> &localStartPosition,
           const HepGeom::Point3D<double> &localEndPosition,
           const double energyLoss,
           const double meanTime,
           const HepMcParticleLink &track,
           const unsigned int id);

  CaloHit(const HepGeom::Point3D<double> &localStartPosition,
           const HepGeom::Point3D<double> &localEndPosition,
           const double energyLoss,
           const double meanTime,
           const HepMcParticleLink &track,
           const int, const int);
  // needed by athenaRoot
  CaloHit();

  // Destructor:
  virtual ~CaloHit(); 
  
  //move assignment defaulted
  CaloHit & operator = (CaloHit &&) = default;
  //assignment defaulted
  CaloHit & operator = (const CaloHit &) = default;
  //copy c'tor defaulted
  CaloHit(const CaloHit &) = default;

  ///////////////////////////////////////////////////////////////////
  // Const methods:
  ///////////////////////////////////////////////////////////////////

  unsigned int identify() const;

  // local start position of the energy deposit:
  HepGeom::Point3D<double> localStartPosition() const;

  HepGeom::Point3D<double> localEndPosition() const;

  // deposited energy:
  double energyLoss() const;

  // time of energy deposition: FIXME name!
  double meanTime() const;

  // Set the time of energy deposition: FIXME name!
  void setMeanTime(float meanTime);

  // number of track which released this energy:
  int trackNumber() const;

  // link to the particle generating the hit
  const HepMcParticleLink& particleLink() const;

  // Veto, trigger or preshower?
  // bool isVeto() const;
  // bool isTrigger() const;
  // bool isPreshower() const;

  // Row
  int getRow() const;

  // Module
  int getModule() const;

  // some print-out:
  void print() const;

  bool operator < (const CaloHit& rhs) const
  {return m_ID < rhs.m_ID;}

  ///////////////////////////////////////////////////////////////////
  // Non-const methods:
  ///////////////////////////////////////////////////////////////////
  // Scale the length, used to go from cm to mm, of whatever we like.
  void ScaleLength(double);

  ///////////////////////////////////////////////////////////////////
  // Private data:
  ///////////////////////////////////////////////////////////////////
private:

  float m_stX, m_stY, m_stZ;
  float m_enX, m_enY, m_enZ;
  float m_energyLoss; // deposited energy
  float m_meanTime; // time of energy deposition
  HepMcParticleLink m_partLink;
  unsigned int m_ID;
public:
  // enum
  //   { xDep = 2, xPhi = 0, xEta = 1};
};


///////////////////////////////////////////////////////////////////
// Inline methods:
///////////////////////////////////////////////////////////////////

inline unsigned int CaloHit::identify() const
{
  return m_ID;
}

inline double CaloHit::energyLoss() const
{
  return (double) m_energyLoss;
}

inline double CaloHit::meanTime() const
{
  return (double) m_meanTime;
}

inline void CaloHit::setMeanTime(float meanTime)
{
  m_meanTime=meanTime;
}

inline const HepMcParticleLink& CaloHit::particleLink() const
{
  return m_partLink;
}



///////////////////////////////////////////////////////////////////
// open functions:
///////////////////////////////////////////////////////////////////

inline float hitTime(const CaloHit& hit)
{
  return (float) hit.meanTime();
}

#endif // CALOSIMEVENT_CALOHIT_H
