/*
   Copyright (C) 2002-2019 CERN for the benefit of the ATLAS collaboration
   */

/**   @file ClusterReadAlg.cxx
 *   Implementation file for the ClusterReadAlg class.
 *   @author Dave Casper
 *   @date 28 February 2021
 */

#include "ClusterReadAlg.h"

#include "TrackerPrepRawData/FaserSCT_Cluster.h"
#include "FaserDetDescr/FaserDetectorID.h"    
#include "TrackerIdentifier/FaserSCT_ID.h"
#include "TrkEventPrimitives/FitQuality.h"
#include "TrkEventPrimitives/ParticleHypothesis.h"
#include "TrkEventPrimitives/LocalParameters.h"
#include "TrkRIO_OnTrack/RIO_OnTrack.h"
#include "StoreGate/ReadHandle.h"
#include <TH1F.h>
#include <TTree.h>
#include <TBranch.h>

/*
To add as new branches:
POOLCollectiontree - IsSimulation (0 or 1)
- event number
- event time
check issimulation before continuing
- coords, energy deposit, etc
*/


//static const std::string moduleFailureReason{"ClusterReadAlg: Exceeds max clusters"};

// Constructor with parameters:
ClusterReadAlg::ClusterReadAlg(const std::string& name, ISvcLocator* pSvcLocator) :
  AthReentrantAlgorithm(name, pSvcLocator), 
  AthHistogramming( name ),
  m_idHelper{nullptr},
  m_histSvc ( "THistSvc/THistSvc", name ){}

// Initialize method:
StatusCode ClusterReadAlg::initialize() {
  ATH_MSG_INFO("ClusterReadAlg::initialize()");
  ATH_CHECK(m_faserRdoKey.initialize());

  ATH_CHECK( m_VetoWaveformContainer.initialize() );
  ATH_CHECK( m_CaloWaveformContainer.initialize() );
  ATH_CHECK( m_TriggerWaveformContainer.initialize() );
  ATH_CHECK( m_PreshowerWaveformContainer.initialize() );
  ATH_CHECK( m_TestWaveformContainer.initialize() );
  ATH_CHECK( m_ClockWaveformContainer.initialize() );

  ATH_MSG_INFO( "Using FaserSCT RDO Container with key " << m_faserRdoKey.key());//works




  // Tree
  m_tree = new TTree("residTree","Cosmics residuals");
  m_tree->Branch("id", &m_hash, "id/I");
  m_tree->Branch("waveformExists", &m_waveformExists, "waveform/B");
  m_tree->Branch("station", &m_station, "station/I");
  m_tree->Branch("waveformdigichannel", &m_wfdigichannel, "wfdigichannel/I");

  //hist test
  m_hist = new TH1D("Stations", "Stations hit", 4, 0, 3);
  ATH_CHECK(histSvc()->regHist("/HIST/myhist", m_hist));
  m_hist2 = new TH1D("DigiChannels WF", "Channels",10,0,10);
  ATH_CHECK(histSvc()->regHist("/HIST/WfChHist", m_hist2));



  return StatusCode::SUCCESS;
}

// Execute method:
StatusCode ClusterReadAlg::execute(const EventContext& ctx) const
{

  ++m_numberOfEvents;  
  ++m_numberOfTriggeredEvents;
  setFilterPassed(false, ctx);

  SG::ReadHandle<FaserSCT_RDO_Container> h_sctRDO(m_faserRdoKey,ctx);

  SG::ReadHandle<ScintWaveformContainer> caloHandle(m_CaloWaveformContainer, ctx);

  SG::ReadHandle<ScintWaveformContainer> vetoHandle(m_VetoWaveformContainer, ctx);

  SG::ReadHandle<ScintWaveformContainer> triggerHandle(m_TriggerWaveformContainer, ctx);

  SG::ReadHandle<ScintWaveformContainer> preshowerHandle(m_PreshowerWaveformContainer, ctx);

  SG::ReadHandle<ScintWaveformContainer> testHandle(m_TestWaveformContainer, ctx);

  SG::ReadHandle<ScintWaveformContainer> clockHandle(m_ClockWaveformContainer, ctx);
  
  int WFexists=0;
  for( const auto& collection : *h_sctRDO) {    
    for(const auto& rawdata : *collection){
      const auto identifier = rawdata->identify();
      //ATH_MSG_INFO(identifier);
      const FaserSCT_ID* pix;
            ServiceHandle<StoreGateSvc> detStore ("DetectorStore", "SiHitIdHelper");
            if (detStore.retrieve().isSuccess()) {
                if (detStore->retrieve(pix, "FaserSCT_ID").isFailure()) { pix = 0; }
      }
            int station = pix->station(identifier);
            int plane = pix->layer(identifier);
            //int row = pix->phi_module(identifier);
            //int module = pix->eta_module(identifier);
            //int sensor = pix->side(identifier);

            activePlanes[station][plane]=1;

            //ATH_MSG_INFO("S: "<<station<<" P:"<<plane);
    }
  }
 
  for( const auto& collection : *vetoHandle)
    { 
      int ch = collection->channel();
      m_hist2->Fill(ch);
      if(collection->event_counter()>0){WFexists++;}
      
    }
  /*
  for( const auto& collection : *caloHandle)
    { 
      int ch = collection->channel();
      m_hist2->Fill(ch);
      WFexists++;
    }

  for( const auto& collection : *triggerHandle)
    { 
      int ch = collection->channel();
      m_hist2->Fill(ch);
      WFexists++;
    }

  for( const auto& collection : *preshowerHandle)
    { 
      int ch = collection->channel();
      m_hist2->Fill(ch);
      WFexists++;
    }

  for( const auto& collection : *clockHandle)
    { 
      int ch = collection->channel();
      m_hist2->Fill(ch);
      WFexists++;
    }

  for( const auto& collection : *testHandle)
    { 
      int ch = collection->channel();
      m_hist2->Fill(ch);
      WFexists++;
    }

  */
  bool ofInterest=false;
  for (int i=1;i<4;i++){
    int layerCheck=0;
    for(int j=0;j<3;j++){
      if (activePlanes[i][j]){layerCheck++;}
    }
    if (layerCheck==3){ofInterest=true;}
  }
  //if (WFexists>0){ATH_MSG_INFO(">>>There's a WF here: "<<WFexists);}
  if (ofInterest){ATH_MSG_INFO("Something interesting happens!");}
  m_tree->Fill();

  // Done
  return StatusCode::SUCCESS;
}

// Finalize method:
StatusCode ClusterReadAlg::finalize() 
{
  ATH_MSG_INFO("ClusterReadAlg::finalize()");
  ATH_MSG_INFO( m_numberOfEvents << " events processed" );
  ATH_MSG_INFO( m_numberOfTriggeredEvents << " triggered events processed" );
  ATH_MSG_INFO( m_numberOfClusterCollection<< " cluster collections processed" );
  ATH_MSG_INFO( m_numberOfCluster<< " cluster processed" );

  ATH_MSG_INFO("station distribution" << m_station[0] << m_station[1]<< m_station[2]<< m_station[3]);
  


  return StatusCode::SUCCESS;
}


