/*
   Copyright (C) 2021 CERN for the benefit of the FASER collaboration
   */

/**   @file ClusterReadAlg
 *   Header file for the ClusterReadAlg class (an Algorithm).
 *   @author Dave Casper
 *   @date 28 February 2021
 */

#ifndef FASERCLUSTERREAD_CLUSTERREADALG_H
#define FASERCLUSTERREAD_CLUSTERREADALG_H

// Base class
#include "AthenaBaseComps/AthReentrantAlgorithm.h"
#include "AthenaBaseComps/AthHistogramming.h"

////Next contains a typedef so cannot be fwd declared
#include "TrackerPrepRawData/FaserSCT_ClusterContainer.h"
#include "TrackerPrepRawData/FaserSCT_ClusterCollection.h"
#include "TrackerPrepRawData/TrackerClusterContainer.h"
#include "Identifier/IdentifierHash.h"
#include "xAODFaserTrigger/FaserTriggerData.h"
#include "TrkTrack/TrackCollection.h"

//Gaudi
#include "GaudiKernel/ServiceHandle.h"
#include "GaudiKernel/ToolHandle.h"

//STL
#include <map>
#include <string>

//RDO
#include <TH1.h>
#include <math.h>
#include <TProfile.h>
#include "GeneratorObjects/McEventCollection.h"
#include "TrackerSimEvent/FaserSiHitCollection.h"
#include "TrackerRawData/FaserSCT_RDO_Container.h" //new

#include "ScintRawEvent/ScintWaveformContainer.h" //new?
#include "TrackerSimData/TrackerSimDataCollection.h"

class FaserSCT_ID;
class ISvcLocator;
class StatusCode;
class TH1;
class TTree;
namespace Trk
{
    class TrackStateOnSurface;
}



/**
 *    @class SCT_Clusterization
 *    @brief Form clusters from SCT Raw Data Objects
 *    The class loops over an RDO grouping strips and creating collections of clusters, subsequently recorded in StoreGate
 *    Uses SCT_ConditionsTools to determine which strips to include.
 */
class ClusterReadAlg : public AthReentrantAlgorithm, AthHistogramming
 {
  public:
    /// Constructor with parameters:
    ClusterReadAlg(const std::string& name, ISvcLocator* pSvcLocator);
    virtual ~ClusterReadAlg(){};
    /**    @name Usual algorithm methods */
    //@{
    ///Retrieve the tools used and initialize variables
    virtual StatusCode initialize() override;
    ///Form clusters and record them in StoreGate (detector store)
    virtual StatusCode execute(const EventContext& ctx) const override;
    ///Clean up and release the collection containers
    virtual StatusCode finalize() override;
    //Make this algorithm clonable.
    virtual bool isClonable() const override { return true; }
    //@}
    const ServiceHandle<ITHistSvc>& histSvc() const;


  private:
    /**    @name Disallow default instantiation, copy, assignment */
    //@{
    ClusterReadAlg() = delete;
    ClusterReadAlg(const ClusterReadAlg&) = delete;
    ClusterReadAlg &operator=(const ClusterReadAlg&) = delete;
    //@}


    const FaserSCT_ID* m_idHelper;
    ServiceHandle<ITHistSvc> m_histSvc;
    SG::ReadHandleKey<FaserSCT_RDO_Container> m_faserRdoKey { this, "FaserSCT_RDO_Container", "SCT_RDOs"};

    SG::ReadHandleKey<ScintWaveformContainer> m_CaloWaveformContainer
    { this, "CaloWaveformContainerKey", "CaloWaveforms", "ReadHandleKey for CaloWaveforms Container"};
    SG::ReadHandleKey<ScintWaveformContainer> m_VetoWaveformContainer
    { this, "VetoWaveformContainerKey", "VetoWaveforms", "ReadHandleKey for CaloWaveforms Container"};
    SG::ReadHandleKey<ScintWaveformContainer> m_TriggerWaveformContainer
    { this, "TriggerWaveformContainerKey", "TriggerWaveforms", "ReadHandleKey for TriggerWaveforms Container"};
    SG::ReadHandleKey<ScintWaveformContainer> m_PreshowerWaveformContainer
    { this, "PreshowerWaveformContainerKey", "PreshowerWaveforms", "ReadHandleKey for PreshowerWaveforms Container"};
    SG::ReadHandleKey<ScintWaveformContainer> m_TestWaveformContainer
    { this, "TestWaveformContainerKey", "TestWaveforms", "ReadHandleKey for TestWaveforms Container"};
    SG::ReadHandleKey<ScintWaveformContainer> m_ClockWaveformContainer
    { this, "ClockWaveformContainerKey", "ClockWaveforms", "ReadHandleKey for ClockWaveforms Container"};



    mutable TTree* m_tree;
    mutable int   m_hash;
    mutable bool  m_waveformExists;
    mutable int   m_station[5]={}; //replace with coordinate vector later
    mutable float m_energyLoss;
    mutable int   m_wfdigichannel;

    mutable std::atomic<int> m_numberOfEvents{0};
    mutable std::atomic<int> m_numberOfTriggeredEvents{0};
    mutable std::atomic<int> m_numberOfClusterCollection{0};
    mutable std::atomic<int> m_numberOfCluster{0};

    mutable std::vector<std::vector<int>> activePlanes {{ 0, 0, 0},{ 0, 0, 0},{ 0, 0, 0},{ 0, 0, 0}}; // activePlanes[st][layer]= 0/1


    //isvalids
    mutable bool m_triggerisValid = false;
    mutable bool m_clusterisValid = false;

    //RDO
    TH1* m_hist;  // Example histogram
    TH1* m_hist2; 
    TProfile* m_hprof;

  
};

// For the THistSvc
inline const ServiceHandle<ITHistSvc>& ClusterReadAlg::histSvc() const 
{
  return m_histSvc;
}


#endif // FASERCLUSTERREAD_CLUSTERREADLG_H
