BEGIN TRANSACTION;
--
-- ALL sql commands must be terminated with a semi-colon!
--
-- Part 1: Table Schema
--
-- Part 1a: HVS Schema
--
-- This table contains one entry per node.
-- Nodes may be either leaves (data tables) or branches (containers).
-- Branches may contain both branches and leaves.
DROP TABLE IF EXISTS "HVS_NODE";
CREATE TABLE IF NOT EXISTS "HVS_NODE" (
	"NODE_ID"	SLONGLONG UNIQUE,
	"NODE_NAME"	TEXT UNIQUE,
	"PARENT_ID"	SLONGLONG,
	"BRANCH_FLAG"	BOOLEAN,
	"NODE_COMMENT"	TEXT
);
-- This table contains one row per tag.
-- A given tag can apply to a single node.
-- A branch tag "contains" the tags of daughter nodes (branch and leaf).
-- A leaf tag can apply to multiple rows of the table in question.
-- Tags can be locked, which must lock all daughter tags.
-- The REPLICATED field does not appear to be used by ATLAS.
-- The SUPPORTED field corresponds to the last supported software release.
DROP TABLE IF EXISTS "HVS_TAG2NODE";
CREATE TABLE IF NOT EXISTS "HVS_TAG2NODE" (
	"NODE_ID"	SLONGLONG,
	"TAG_NAME"	TEXT UNIQUE,
	"TAG_ID"	SLONGLONG UNIQUE,
	"TAG_COMMENT"	TEXT,
	"LOCKED"	BOOLEAN,
	"REPLICATED"	BOOLEAN,
	"DATE_CREATED"	DATE,
	"DATE_LOCKED"	DATE,
	"SUPPORTED"	INT
);
-- This table implements the many <--> many relationship between child 
-- and parent tags.
-- Each tag may "own" multiple tags at below it, and be owned by multiple
-- tags above it.
-- For example, a given tag on the root FASER node owns concurrent tags
-- for each component (tracker, calorimeter) of the experiment.
-- And each tag for a component may be owned by multiple different root
-- level tags.
DROP TABLE IF EXISTS "HVS_LTAG2LTAG";
CREATE TABLE IF NOT EXISTS "HVS_LTAG2LTAG" (
	"PARENT_NODE"	SLONGLONG,
	"PARENT_TAG"	SLONGLONG,
	"CHILD_NODE"	SLONGLONG,
	"CHILD_TAG"	SLONGLONG
);
-- This table holds all the tags directly or indirectly contained by
-- each root level tag.
-- In principle it should be populated from the other HVS tables which
-- contain the same information.
DROP TABLE IF EXISTS "HVS_TAGCACHE";
CREATE TABLE IF NOT EXISTS "HVS_TAGCACHE" (
	"ROOTTAG"	TEXT,
	"CHILDNODE"	TEXT,
	"CHILDTAG"	TEXT,
	"CHILDTAGID"	SLONGLONG
);
--
-- Part 1b: Content schema (Leaf nodes)
--
-- These tables hold standard material and element definitions
-- The actual data is copied from ATLAS
DROP TABLE IF EXISTS "STDMATERIALS_DATA";
CREATE TABLE IF NOT EXISTS "STDMATERIALS_DATA" (
	"STDMATERIALS_DATA_ID"	SLONGLONG UNIQUE,
	"NAME"	TEXT,
	"DENSITY"	DOUBLE
);
DROP TABLE IF EXISTS "STDMATERIALS_DATA2TAG";
CREATE TABLE IF NOT EXISTS "STDMATERIALS_DATA2TAG" (
	"STDMATERIALS_TAG_ID"	SLONGLONG,
	"STDMATERIALS_DATA_ID"	SLONGLONG
);
DROP TABLE IF EXISTS "STDMATCOMPONENTS_DATA";
CREATE TABLE IF NOT EXISTS "STDMATCOMPONENTS_DATA" (
	"STDMATCOMPONENTS_DATA_ID"	SLONGLONG UNIQUE,
	"MATERIAL_ID"	SLONGLONG,
	"COMPNAME"	TEXT,
	"FRACTION"	DOUBLE
);
DROP TABLE IF EXISTS "STDMATCOMPONENTS_DATA2TAG";
CREATE TABLE IF NOT EXISTS "STDMATCOMPONENTS_DATA2TAG" (
	"STDMATCOMPONENTS_TAG_ID"	SLONGLONG,
	"STDMATCOMPONENTS_DATA_ID"	SLONGLONG
);
DROP TABLE IF EXISTS "ELEMENTS_DATA";
CREATE TABLE IF NOT EXISTS "ELEMENTS_DATA" (
	"ELEMENTS_DATA_ID"	SLONGLONG UNIQUE,
	"NAME"	TEXT,
	"SYMBOL"	TEXT,
	"A"	DOUBLE,
	"Z"	DOUBLE
);
DROP TABLE IF EXISTS "ELEMENTS_DATA2TAG";
CREATE TABLE IF NOT EXISTS "ELEMENTS_DATA2TAG" (
	"ELEMENTS_TAG_ID"	SLONGLONG,
	"ELEMENTS_DATA_ID"	SLONGLONG
);
--
-- Tables for describing the emulsion detector
--
DROP TABLE IF EXISTS "EMULSIONTOPLEVEL_DATA";
CREATE TABLE IF NOT EXISTS "EMULSIONTOPLEVEL_DATA" (
	"EMULSIONTOPLEVEL_DATA_ID"	SLONGLONG UNIQUE,
	"POSX" DOUBLE,
	"POSY" DOUBLE,
	"POSZ" DOUBLE,
	"ROTX" DOUBLE,
	"ROTY" DOUBLE,
	"ROTZ" DOUBLE,
	"ROTORDER" INT,
	"LABEL" TEXT
);
--
-- The DATA2TAG tables associate specific rows of the corresponding
-- _DATA table with the referenced tag (from the HVS_TAG2NODE table).
-- This is a many-to-many relationship: each row may belong to
-- several tags, and each tag may apply to several rows.
DROP TABLE IF EXISTS "EMULSIONTOPLEVEL_DATA2TAG";
CREATE TABLE IF NOT EXISTS "EMULSIONTOPLEVEL_DATA2TAG" (
	"EMULSIONTOPLEVEL_TAG_ID" SLONGLONG,
	"EMULSIONTOPLEVEL_DATA_ID" SLONGLONG
);
--
-- Tables for describing Veto scintillator plates (and passive radiators)
--
DROP TABLE IF EXISTS "VETOTOPLEVEL_DATA";
CREATE TABLE IF NOT EXISTS "VETOTOPLEVEL_DATA" (
	"VETOTOPLEVEL_DATA_ID"	SLONGLONG UNIQUE,
	"POSX" DOUBLE,
	"POSY" DOUBLE,
	"POSZ" DOUBLE,
	"ROTX" DOUBLE,
	"ROTY" DOUBLE,
	"ROTZ" DOUBLE,
	"ROTORDER" INT,
	"LABEL" TEXT
);
--
-- The DATA2TAG tables associate specific rows of the corresponding
-- _DATA table with the referenced tag (from the HVS_TAG2NODE table).
-- This is a many-to-many relationship: each row may belong to
-- several tags, and each tag may apply to several rows.
DROP TABLE IF EXISTS "VETOTOPLEVEL_DATA2TAG";
CREATE TABLE IF NOT EXISTS "VETOTOPLEVEL_DATA2TAG" (
	"VETOTOPLEVEL_TAG_ID" SLONGLONG,
	"VETOTOPLEVEL_DATA_ID" SLONGLONG
);
--
DROP TABLE IF EXISTS "VETOSTATIONGENERAL_DATA";
CREATE TABLE IF NOT EXISTS "VETOSTATIONGENERAL_DATA" (
	"VETOSTATIONGENERAL_DATA_ID" SLONGLONG UNIQUE,
	"NUMPLATES" INT,
	"PLATEPITCH" DOUBLE
);
--
DROP TABLE IF EXISTS "VETOSTATIONGENERAL_DATA2TAG";
CREATE TABLE IF NOT EXISTS "VETOSTATIONGENERAL_DATA2TAG" (
	"VETOSTATIONGENERAL_TAG_ID" SLONGLONG,
	"VETOSTATIONGENERAL_DATA_ID" SLONGLONG
);
DROP TABLE IF EXISTS "VETOPLATEGENERAL_DATA";
CREATE TABLE IF NOT EXISTS "VETOPLATEGENERAL_DATA" (
	"VETOPLATEGENERAL_DATA_ID" SLONGLONG UNIQUE,
	"NUMPMTS" INT,
	"WIDTH" DOUBLE,
	"LENGTH" DOUBLE,
	"THICKNESS" DOUBLE,
	"MATERIAL" TEXT
);
--
DROP TABLE IF EXISTS "VETOPLATEGENERAL_DATA2TAG";
CREATE TABLE IF NOT EXISTS "VETOPLATEGENERAL_DATA2TAG" (
	"VETOPLATEGENERAL_TAG_ID" SLONGLONG,
	"VETOPLATEGENERAL_DATA_ID" SLONGLONG
);
--
-- Tables for describing Trigger scintillator plates (and passive radiators)
--
DROP TABLE IF EXISTS "TRIGGERTOPLEVEL_DATA";
CREATE TABLE IF NOT EXISTS "TRIGGERTOPLEVEL_DATA" (
	"TRIGGERTOPLEVEL_DATA_ID"	SLONGLONG UNIQUE,
	"POSX" DOUBLE,
	"POSY" DOUBLE,
	"POSZ" DOUBLE,
	"ROTX" DOUBLE,
	"ROTY" DOUBLE,
	"ROTZ" DOUBLE,
	"ROTORDER" INT,
	"LABEL" TEXT
);
--
-- The DATA2TAG tables associate specific rows of the corresponding
-- _DATA table with the referenced tag (from the HVS_TAG2NODE table).
-- This is a many-to-many relationship: each row may belong to
-- several tags, and each tag may apply to several rows.
DROP TABLE IF EXISTS "TRIGGERTOPLEVEL_DATA2TAG";
CREATE TABLE IF NOT EXISTS "TRIGGERTOPLEVEL_DATA2TAG" (
	"TRIGGERTOPLEVEL_TAG_ID" SLONGLONG,
	"TRIGGERTOPLEVEL_DATA_ID" SLONGLONG
);
--
DROP TABLE IF EXISTS "TRIGGERSTATIONGENERAL_DATA";
CREATE TABLE IF NOT EXISTS "TRIGGERSTATIONGENERAL_DATA" (
	"TRIGGERSTATIONGENERAL_DATA_ID" SLONGLONG UNIQUE,
	"NUMPLATES" INT,
	"PLATEPITCHY" DOUBLE,
	"PLATEPITCHZ" DOUBLE
);
--
DROP TABLE IF EXISTS "TRIGGERSTATIONGENERAL_DATA2TAG";
CREATE TABLE IF NOT EXISTS "TRIGGERSTATIONGENERAL_DATA2TAG" (
	"TRIGGERSTATIONGENERAL_TAG_ID" SLONGLONG,
	"TRIGGERSTATIONGENERAL_DATA_ID" SLONGLONG
);
DROP TABLE IF EXISTS "TRIGGERPLATEGENERAL_DATA";
CREATE TABLE IF NOT EXISTS "TRIGGERPLATEGENERAL_DATA" (
	"TRIGGERPLATEGENERAL_DATA_ID" SLONGLONG UNIQUE,
	"NUMPMTS" INT,
	"WIDTH" DOUBLE,
	"LENGTH" DOUBLE,
	"THICKNESS" DOUBLE,
	"MATERIAL" TEXT
);
--
DROP TABLE IF EXISTS "TRIGGERPLATEGENERAL_DATA2TAG";
CREATE TABLE IF NOT EXISTS "TRIGGERPLATEGENERAL_DATA2TAG" (
	"TRIGGERPLATEGENERAL_TAG_ID" SLONGLONG,
	"TRIGGERPLATEGENERAL_DATA_ID" SLONGLONG
);
--
-- Tables for describing Preshower scintillator plates (and passive radiators)
--
DROP TABLE IF EXISTS "PRESHOWERTOPLEVEL_DATA";
CREATE TABLE IF NOT EXISTS "PRESHOWERTOPLEVEL_DATA" (
	"PRESHOWERTOPLEVEL_DATA_ID"	SLONGLONG UNIQUE,
	"POSX" DOUBLE,
	"POSY" DOUBLE,
	"POSZ" DOUBLE,
	"ROTX" DOUBLE,
	"ROTY" DOUBLE,
	"ROTZ" DOUBLE,
	"ROTORDER" INT,
	"LABEL" TEXT
);
--
-- The DATA2TAG tables associate specific rows of the corresponding
-- _DATA table with the referenced tag (from the HVS_TAG2NODE table).
-- This is a many-to-many relationship: each row may belong to
-- several tags, and each tag may apply to several rows.
DROP TABLE IF EXISTS "PRESHOWERTOPLEVEL_DATA2TAG";
CREATE TABLE IF NOT EXISTS "PRESHOWERTOPLEVEL_DATA2TAG" (
	"PRESHOWERTOPLEVEL_TAG_ID" SLONGLONG,
	"PRESHOWERTOPLEVEL_DATA_ID" SLONGLONG
);
--
DROP TABLE IF EXISTS "PRESHOWERSTATIONGENERAL_DATA";
CREATE TABLE IF NOT EXISTS "PRESHOWERSTATIONGENERAL_DATA" (
	"PRESHOWERSTATIONGENERAL_DATA_ID" SLONGLONG UNIQUE,
	"NUMPLATES" INT,
	"PLATEPITCH" DOUBLE
);
--
DROP TABLE IF EXISTS "PRESHOWERSTATIONGENERAL_DATA2TAG";
CREATE TABLE IF NOT EXISTS "PRESHOWERSTATIONGENERAL_DATA2TAG" (
	"PRESHOWERSTATIONGENERAL_TAG_ID" SLONGLONG,
	"PRESHOWERSTATIONGENERAL_DATA_ID" SLONGLONG
);
DROP TABLE IF EXISTS "PRESHOWERPLATEGENERAL_DATA";
CREATE TABLE IF NOT EXISTS "PRESHOWERPLATEGENERAL_DATA" (
	"PRESHOWERPLATEGENERAL_DATA_ID" SLONGLONG UNIQUE,
	"NUMPMTS" INT,
	"WIDTH" DOUBLE,
	"LENGTH" DOUBLE,
	"THICKNESS" DOUBLE,
	"MATERIAL" TEXT
);
--
DROP TABLE IF EXISTS "PRESHOWERPLATEGENERAL_DATA2TAG";
CREATE TABLE IF NOT EXISTS "PRESHOWERPLATEGENERAL_DATA2TAG" (
	"PRESHOWERPLATEGENERAL_TAG_ID" SLONGLONG,
	"PRESHOWERPLATEGENERAL_DATA_ID" SLONGLONG
);
--
-- Ecal top level tables
--
DROP TABLE IF EXISTS "ECALTOPLEVEL_DATA";
CREATE TABLE IF NOT EXISTS "ECALTOPLEVEL_DATA" (
	"ECALTOPLEVEL_DATA_ID"	SLONGLONG UNIQUE,
	"POSX" DOUBLE,
	"POSY" DOUBLE,
	"POSZ" DOUBLE,
	"ROTX" DOUBLE,
	"ROTY" DOUBLE,
	"ROTZ" DOUBLE,
	"ROTORDER" INT,
	"LABEL" TEXT
);
--
-- The DATA2TAG tables associate specific rows of the corresponding
-- _DATA table with the referenced tag (from the HVS_TAG2NODE table).
-- This is a many-to-many relationship: each row may belong to
-- several tags, and each tag may apply to several rows.
DROP TABLE IF EXISTS "ECALTOPLEVEL_DATA2TAG";
CREATE TABLE IF NOT EXISTS "ECALTOPLEVEL_DATA2TAG" (
	"ECALTOPLEVEL_TAG_ID" SLONGLONG,
	"ECALTOPLEVEL_DATA_ID" SLONGLONG
);
--
--
-- Ecal row general tables
--
DROP TABLE IF EXISTS "ECALROWGENERAL_DATA";
CREATE TABLE IF NOT EXISTS "ECALROWGENERAL_DATA" (
	"ECALROWGENERAL_DATA_ID" SLONGLONG UNIQUE,
	"NUMMODULES" INT,
	"XGAP" DOUBLE,
	"YAWANGLE" DOUBLE
);
--
DROP TABLE IF EXISTS "ECALROWGENERAL_DATA2TAG";
CREATE TABLE IF NOT EXISTS "ECALROWGENERAL_DATA2TAG" (
	"ECALROWGENERAL_TAG_ID" SLONGLONG,
	"ECALROWGENERAL_DATA_ID" SLONGLONG
);
--
-- Tables for scintillator materials.
-- Materials are recorded in two related nodes.
-- One stores the name and density of the material.
-- The second stores the composition in terms of other materials.
--
-- Neutrino detector materials
DROP TABLE IF EXISTS "NEUTRINOMATERIALS_DATA";
CREATE TABLE IF NOT EXISTS "NEUTRINOMATERIALS_DATA" (
	"NEUTRINOMATERIALS_DATA_ID" SLONGLONG UNIQUE,
	"NAME" TEXT,
	"DENSITY" DOUBLE
);
DROP TABLE IF EXISTS "NEUTRINOMATERIALS_DATA2TAG";
CREATE TABLE IF NOT EXISTS "NEUTRINOMATERIALS_DATA2TAG" (
	"NEUTRINOMATERIALS_TAG_ID" SLONGLONG,
	"NEUTRINOMATERIALS_DATA_ID" SLONGLONG
);
--
DROP TABLE IF EXISTS "NEUTRINOMATCOMPONENTS_DATA";
CREATE TABLE IF NOT EXISTS "NEUTRINOMATCOMPONENTS_DATA" (
	"NEUTRINOMATCOMPONENTS_DATA_ID" SLONGLONG UNIQUE,
	"MATERIAL_ID" SLONGLONG,
	"COMPNAME" TEXT,
	"FRACTION" DOUBLE
);
DROP TABLE IF EXISTS "NEUTRINOMATCOMPONENTS_DATA2TAG";
CREATE TABLE IF NOT EXISTS "NEUTRINOMATCOMPONENTS_DATA2TAG" (
	"NEUTRINOMATCOMPONENTS_TAG_ID" SLONGLONG,
	"NEUTRINOMATCOMPONENTS_DATA_ID" SLONGLONG
);
-- Scintillator materials
DROP TABLE IF EXISTS "SCINTMATERIALS_DATA";
CREATE TABLE IF NOT EXISTS "SCINTMATERIALS_DATA" (
	"SCINTMATERIALS_DATA_ID" SLONGLONG UNIQUE,
	"NAME" TEXT,
	"DENSITY" DOUBLE
);
DROP TABLE IF EXISTS "SCINTMATERIALS_DATA2TAG";
CREATE TABLE IF NOT EXISTS "SCINTMATERIALS_DATA2TAG" (
	"SCINTMATERIALS_TAG_ID" SLONGLONG,
	"SCINTMATERIALS_DATA_ID" SLONGLONG
);
--
DROP TABLE IF EXISTS "SCINTMATCOMPONENTS_DATA";
CREATE TABLE IF NOT EXISTS "SCINTMATCOMPONENTS_DATA" (
	"SCINTMATCOMPONENTS_DATA_ID" SLONGLONG UNIQUE,
	"MATERIAL_ID" SLONGLONG,
	"COMPNAME" TEXT,
	"FRACTION" DOUBLE
);
DROP TABLE IF EXISTS "SCINTMATCOMPONENTS_DATA2TAG";
CREATE TABLE IF NOT EXISTS "SCINTMATCOMPONENTS_DATA2TAG" (
	"SCINTMATCOMPONENTS_TAG_ID" SLONGLONG,
	"SCINTMATCOMPONENTS_DATA_ID" SLONGLONG
);
-- Tracker materials
DROP TABLE IF EXISTS "TRACKERMATERIALS_DATA";
CREATE TABLE IF NOT EXISTS "TRACKERMATERIALS_DATA" (
	"TRACKERMATERIALS_DATA_ID" SLONGLONG UNIQUE,
	"NAME" TEXT,
	"DENSITY" DOUBLE
);
DROP TABLE IF EXISTS "TRACKERMATERIALS_DATA2TAG";
CREATE TABLE IF NOT EXISTS "TRACKERMATERIALS_DATA2TAG" (
	"TRACKERMATERIALS_TAG_ID" SLONGLONG,
	"TRACKERMATERIALS_DATA_ID" SLONGLONG
);
DROP TABLE IF EXISTS "TRACKERMATCOMPONENTS_DATA";
CREATE TABLE IF NOT EXISTS "TRACKERMATCOMPONENTS_DATA" (
	"TRACKERMATCOMPONENTS_DATA_ID" SLONGLONG UNIQUE,
	"MATERIAL_ID" SLONGLONG,
	"COMPNAME" TEXT,
	"FRACTION" DOUBLE
);
DROP TABLE IF EXISTS "TRACKERMATCOMPONENTS_DATA2TAG";
CREATE TABLE IF NOT EXISTS "TRACKERMATCOMPONENTS_DATA2TAG" (
	"TRACKERMATCOMPONENTS_TAG_ID" SLONGLONG,
	"TRACKERMATCOMPONENTS_DATA_ID" SLONGLONG
);
-- Calo materials
DROP TABLE IF EXISTS "CALOMATERIALS_DATA";
CREATE TABLE IF NOT EXISTS "CALOMATERIALS_DATA" (
	"CALOMATERIALS_DATA_ID" SLONGLONG UNIQUE,
	"NAME" TEXT,
	"DENSITY" DOUBLE
);
DROP TABLE IF EXISTS "CALOMATERIALS_DATA2TAG";
CREATE TABLE IF NOT EXISTS "CALOMATERIALS_DATA2TAG" (
	"CALOMATERIALS_TAG_ID" SLONGLONG,
	"CALOMATERIALS_DATA_ID" SLONGLONG
);
DROP TABLE IF EXISTS "CALOMATCOMPONENTS_DATA";
CREATE TABLE IF NOT EXISTS "CALOMATCOMPONENTS_DATA" (
	"CALOMATCOMPONENTS_DATA_ID" SLONGLONG UNIQUE,
	"MATERIAL_ID" SLONGLONG,
	"COMPNAME" TEXT,
	"FRACTION" DOUBLE
);
DROP TABLE IF EXISTS "CALOMATCOMPONENTS_DATA2TAG";
CREATE TABLE IF NOT EXISTS "CALOMATCOMPONENTS_DATA2TAG" (
	"CALOMATCOMPONENTS_TAG_ID" SLONGLONG,
	"CALOMATCOMPONENTS_DATA_ID" SLONGLONG
);
--
DROP TABLE IF EXISTS "FASERCOMMON_DATA";
CREATE TABLE IF NOT EXISTS "FASERCOMMON_DATA" ( 
	"FASERCOMMON_DATA_ID" SLONGLONG UNIQUE,
	"CONFIG" TEXT ,
	"TRACKERTYPE" TEXT ,
	"FASERNUTYPE" TEXT 
);
DROP TABLE IF EXISTS "FASERCOMMON_DATA2TAG";
CREATE TABLE IF NOT EXISTS "FASERCOMMON_DATA2TAG" (
	"FASERCOMMON_TAG_ID" SLONGLONG,
	"FASERCOMMON_DATA_ID" SLONGLONG
);
--
DROP TABLE IF EXISTS "EMULSIONSWITCHES_DATA";
CREATE TABLE IF NOT EXISTS "EMULSIONSWITCHES_DATA" ( 
	"EMULSIONSWITCHES_DATA_ID" SLONGLONG UNIQUE,
	"DETECTORNAME" TEXT ,
	"USEMAGFIELDSVC" INT ,
	"COSMICLAYOUT" INT ,
	"VERSIONNAME" TEXT ,
	"LAYOUT" TEXT ,
	"DESCRIPTION" TEXT
);
DROP TABLE IF EXISTS "EMULSIONSWITCHES_DATA2TAG";
CREATE TABLE IF NOT EXISTS "EMULSIONSWITCHES_DATA2TAG" (
	"EMULSIONSWITCHES_TAG_ID" SLONGLONG,
	"EMULSIONSWITCHES_DATA_ID" SLONGLONG
);
--
DROP TABLE IF EXISTS "VETOSWITCHES_DATA";
CREATE TABLE IF NOT EXISTS "VETOSWITCHES_DATA" ( 
	"VETOSWITCHES_DATA_ID" SLONGLONG UNIQUE,
	"DETECTORNAME" TEXT ,
	"USEMAGFIELDSVC" INT ,
	"COSMICLAYOUT" INT ,
	"VERSIONNAME" TEXT ,
	"LAYOUT" TEXT ,
	"DESCRIPTION" TEXT
);
DROP TABLE IF EXISTS "VETOSWITCHES_DATA2TAG";
CREATE TABLE IF NOT EXISTS "VETOSWITCHES_DATA2TAG" (
	"VETOSWITCHES_TAG_ID" SLONGLONG,
	"VETOSWITCHES_DATA_ID" SLONGLONG
);
--
DROP TABLE IF EXISTS "TRIGGERSWITCHES_DATA";
CREATE TABLE IF NOT EXISTS "TRIGGERSWITCHES_DATA" ( 
	"TRIGGERSWITCHES_DATA_ID" SLONGLONG UNIQUE,
	"DETECTORNAME" TEXT ,
	"USEMAGFIELDSVC" INT ,
	"COSMICLAYOUT" INT ,
	"VERSIONNAME" TEXT ,
	"LAYOUT" TEXT ,
	"DESCRIPTION" TEXT
);
DROP TABLE IF EXISTS "TRIGGERSWITCHES_DATA2TAG";
CREATE TABLE IF NOT EXISTS "TRIGGERSWITCHES_DATA2TAG" (
	"TRIGGERSWITCHES_TAG_ID" SLONGLONG,
	"TRIGGERSWITCHES_DATA_ID" SLONGLONG
);
-- 
DROP TABLE IF EXISTS "PRESHOWERSWITCHES_DATA";
CREATE TABLE IF NOT EXISTS "PRESHOWERSWITCHES_DATA" ( 
	"PRESHOWERSWITCHES_DATA_ID" SLONGLONG UNIQUE,
	"DETECTORNAME" TEXT ,
	"USEMAGFIELDSVC" INT ,
	"COSMICLAYOUT" INT ,
	"VERSIONNAME" TEXT ,
	"LAYOUT" TEXT ,
	"DESCRIPTION" TEXT
);
DROP TABLE IF EXISTS "PRESHOWERSWITCHES_DATA2TAG";
CREATE TABLE IF NOT EXISTS "PRESHOWERSWITCHES_DATA2TAG" (
	"PRESHOWERSWITCHES_TAG_ID" SLONGLONG,
	"PRESHOWERSWITCHES_DATA_ID" SLONGLONG
);
--
DROP TABLE IF EXISTS "ECALSWITCHES_DATA";
CREATE TABLE IF NOT EXISTS "ECALSWITCHES_DATA" ( 
	"ECALSWITCHES_DATA_ID" SLONGLONG UNIQUE,
	"DETECTORNAME" TEXT ,
	"USEMAGFIELDSVC" INT ,
	"COSMICLAYOUT" INT ,
	"VERSIONNAME" TEXT ,
	"LAYOUT" TEXT ,
	"DESCRIPTION" TEXT,
	"GDMLFILE" TEXT
);
DROP TABLE IF EXISTS "ECALSWITCHES_DATA2TAG";
CREATE TABLE IF NOT EXISTS "ECALSWITCHES_DATA2TAG" (
	"ECALSWITCHES_TAG_ID" SLONGLONG,
	"ECALSWITCHES_DATA_ID" SLONGLONG
);
--
DROP TABLE IF EXISTS "NEUTRINOIDENTIFIER_DATA";
CREATE TABLE IF NOT EXISTS "NEUTRINOIDENTIFIER_DATA" ( 
	"NEUTRINOIDENTIFIER_DATA_ID" SLONGLONG UNIQUE,
	"DICT_NAME" TEXT ,
	"DICT_FILENAME" TEXT ,
	"DICT_TAG" TEXT 
);
DROP TABLE IF EXISTS "NEUTRINOIDENTIFIER_DATA2TAG";
CREATE TABLE IF NOT EXISTS "NEUTRINOIDENTIFIER_DATA2TAG" (
	"NEUTRINOIDENTIFIER_TAG_ID" SLONGLONG,
	"NEUTRINOIDENTIFIER_DATA_ID" SLONGLONG
);
--
DROP TABLE IF EXISTS "SCINTIDENTIFIER_DATA";
CREATE TABLE IF NOT EXISTS "SCINTIDENTIFIER_DATA" ( 
	"SCINTIDENTIFIER_DATA_ID" SLONGLONG UNIQUE,
	"DICT_NAME" TEXT ,
	"DICT_FILENAME" TEXT ,
	"DICT_TAG" TEXT 
);
DROP TABLE IF EXISTS "SCINTIDENTIFIER_DATA2TAG";
CREATE TABLE IF NOT EXISTS "SCINTIDENTIFIER_DATA2TAG" (
	"SCINTIDENTIFIER_TAG_ID" SLONGLONG,
	"SCINTIDENTIFIER_DATA_ID" SLONGLONG
);
-- 
DROP TABLE IF EXISTS "TRACKERIDENTIFIER_DATA";
CREATE TABLE IF NOT EXISTS "TRACKERIDENTIFIER_DATA" ( 
	"TRACKERIDENTIFIER_DATA_ID" SLONGLONG UNIQUE,
	"DICT_NAME" TEXT ,
	"DICT_FILENAME" TEXT ,
	"DICT_TAG" TEXT 
);
DROP TABLE IF EXISTS "TRACKERIDENTIFIER_DATA2TAG";
CREATE TABLE IF NOT EXISTS "TRACKERIDENTIFIER_DATA2TAG" (
	"TRACKERIDENTIFIER_TAG_ID" SLONGLONG,
	"TRACKERIDENTIFIER_DATA_ID" SLONGLONG
);
-- 
DROP TABLE IF EXISTS "CALOIDENTIFIER_DATA";
CREATE TABLE IF NOT EXISTS "CALOIDENTIFIER_DATA" ( 
	"CALOIDENTIFIER_DATA_ID" SLONGLONG UNIQUE,
	"DICT_NAME" TEXT ,
	"DICT_FILENAME" TEXT ,
	"DICT_TAG" TEXT 
);
DROP TABLE IF EXISTS "CALOIDENTIFIER_DATA2TAG";
CREATE TABLE IF NOT EXISTS "CALOIDENTIFIER_DATA2TAG" (
	"CALOIDENTIFIER_TAG_ID" SLONGLONG,
	"CALOIDENTIFIER_DATA_ID" SLONGLONG
);
--
-- Part 2a: HVS data
--
-- Data for the HVS_NODE table
INSERT INTO "HVS_NODE" VALUES (0,    "FASER", 0, 1, NULL);
INSERT INTO "HVS_NODE" VALUES (90,   "FaserCommon", 0, 0, NULL);
INSERT INTO "HVS_NODE" VALUES (9000, "Materials", 0, 1, NULL);
INSERT INTO "HVS_NODE" VALUES (9001, "StdMaterials", 9000, 0, NULL);
INSERT INTO "HVS_NODE" VALUES (9002, "StdMatComponents", 9000, 0, NULL);
INSERT INTO "HVS_NODE" VALUES (9003, "Elements", 9000, 0, NULL);
INSERT INTO "HVS_NODE" VALUES (1,    "Neutrino", 0, 1, NULL);
INSERT INTO "HVS_NODE" VALUES (1003, "NeutrinoMaterials", 1, 0, NULL);
INSERT INTO "HVS_NODE" VALUES (1004, "NeutrinoMatComponents", 1, 0, NULL);
INSERT INTO "HVS_NODE" VALUES (1005, "NeutrinoIdentifier", 1, 0, NULL);
INSERT INTO "HVS_NODE" VALUES (11,   "Emulsion", 1, 1, NULL);
INSERT INTO "HVS_NODE" VALUES (110,  "EmulsionTopLevel", 11, 0, NULL);
INSERT INTO "HVS_NODE" VALUES (114,  "EmulsionSwitches", 11, 0, NULL);
INSERT INTO "HVS_NODE" VALUES (2,    "Scintillator", 0, 1, NULL);
INSERT INTO "HVS_NODE" VALUES (2003, "ScintMaterials", 2, 0, NULL);
INSERT INTO "HVS_NODE" VALUES (2004, "ScintMatComponents", 2, 0, NULL);
INSERT INTO "HVS_NODE" VALUES (2005, "ScintIdentifier", 2, 0, NULL);
INSERT INTO "HVS_NODE" VALUES (21,   "Veto", 2, 1, NULL);
INSERT INTO "HVS_NODE" VALUES (210,  "VetoTopLevel", 21, 0, NULL);
INSERT INTO "HVS_NODE" VALUES (211,  "VetoStationGeneral", 21, 0, NULL);
INSERT INTO "HVS_NODE" VALUES (212,  "VetoPlateGeneral", 21, 0, NULL);
INSERT INTO "HVS_NODE" VALUES (214,  "VetoSwitches", 21, 0, NULL );
INSERT INTO "HVS_NODE" VALUES (22,   "Trigger", 2, 1, NULL);
INSERT INTO "HVS_NODE" VALUES (220,  "TriggerTopLevel", 22, 0, NULL);
INSERT INTO "HVS_NODE" VALUES (221,  "TriggerStationGeneral", 22, 0, NULL);
INSERT INTO "HVS_NODE" VALUES (222,  "TriggerPlateGeneral", 22, 0, NULL);
INSERT INTO "HVS_NODE" VALUES (224,  "TriggerSwitches", 22, 0, NULL );
INSERT INTO "HVS_NODE" VALUES (23,   "Preshower", 2, 1, NULL);
INSERT INTO "HVS_NODE" VALUES (230,  "PreshowerTopLevel", 23, 0, NULL);
INSERT INTO "HVS_NODE" VALUES (231,  "PreshowerStationGeneral", 23, 0, NULL);
INSERT INTO "HVS_NODE" VALUES (232,  "PreshowerPlateGeneral", 23, 0, NULL);
INSERT INTO "HVS_NODE" VALUES (234,  "PreshowerSwitches", 23, 0, NULL );
INSERT INTO "HVS_NODE" VALUES (3,    "Tracker", 0, 1, NULL);
INSERT INTO "HVS_NODE" VALUES (3003, "TrackerMaterials", 3, 0, NULL);
INSERT INTO "HVS_NODE" VALUES (3004, "TrackerMatComponents", 3, 0, NULL);
INSERT INTO "HVS_NODE" VALUES (3005, "TrackerIdentifier", 3, 0, NULL);
INSERT INTO "HVS_NODE" VALUES (31,   "SCT", 3, 1, NULL);
INSERT INTO "HVS_NODE" VALUES (310,  "SctTopLevel", 31, 0, NULL);
INSERT INTO "HVS_NODE" VALUES (311,  "SctBrlModule", 31, 0, NULL);
INSERT INTO "HVS_NODE" VALUES (312,  "SctBrlSensor", 31, 0, NULL);
INSERT INTO "HVS_NODE" VALUES (313,  "SctFaserGeneral", 31, 0, NULL);
INSERT INTO "HVS_NODE" VALUES (314,  "SctSwitches", 31, 0, NULL);
INSERT INTO "HVS_NODE" VALUES (315,  "SCTMaterials", 31, 0, NULL);
INSERT INTO "HVS_NODE" VALUES (316,  "SCTMatComponents", 31, 0, NULL);
INSERT INTO "HVS_NODE" VALUES (317,  "SctConditions", 31, 0, NULL);
INSERT INTO "HVS_NODE" VALUES (318,  "SctFrame", 31, 1, NULL);
INSERT INTO "HVS_NODE" VALUES (3181, "SctFrameGeneral", 318, 0, NULL);
INSERT INTO "HVS_NODE" VALUES (3182, "SctFrameShape", 318, 0, NULL);
INSERT INTO "HVS_NODE" VALUES (32,   "Dipole", 3, 1, NULL);
INSERT INTO "HVS_NODE" VALUES (320,  "DipoleTopLevel", 32, 0, NULL);
INSERT INTO "HVS_NODE" VALUES (321,  "DipoleGeneral", 32, 0, NULL);
INSERT INTO "HVS_NODE" VALUES (324,  "DipoleSwitches", 32, 0, NULL);
INSERT INTO "HVS_NODE" VALUES (4,    "Calorimeter", 0, 1, NULL);
INSERT INTO "HVS_NODE" VALUES (41,   "Ecal", 4, 1, NULL);
INSERT INTO "HVS_NODE" VALUES (410,  "EcalTopLevel", 41, 0, NULL);
INSERT INTO "HVS_NODE" VALUES (411,  "EcalRowGeneral", 41, 0, NULL);
INSERT INTO "HVS_NODE" VALUES (414,  "EcalSwitches", 41, 0, NULL);
INSERT INTO "HVS_NODE" VALUES (4003, "CaloMaterials", 4, 0, NULL);
INSERT INTO "HVS_NODE" VALUES (4004, "CaloMatComponents", 4, 0, NULL);
INSERT INTO "HVS_NODE" VALUES (4005, "CaloIdentifier", 4, 0, NULL);
-- Data for the HVS_TAG2NODE table
-- Command to get midnight of current date in UT:
-- date -u --date '00:00 today' +%s
INSERT INTO "HVS_TAG2NODE" VALUES (0, "FASER-00", 100000, NULL, 0, 0, 1549238400000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (0, "FASER-01", 100039, NULL, 0, 0, 1590796800000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (0, "FASER-CR", 107784, NULL, 0, 0, 1598400000000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (90,   "FaserCommon-00", 100013, NULL, 0, 0, 1549324800000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (9000, "Materials-00", 100005, NULL, 0, 0, 1549238400000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (9001, "StdMaterials-00", 100006, NULL, 0, 0, 1549238400000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (9002, "StdMatComponents-00", 100007, NULL, 0, 0, 1549238400000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (9003, "Elements-00", 100008, NULL, 0, 0, 1549238400000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (1, "Neutrino-00", 100031, NULL, 0, 0, 1582416000000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (11,"Emulsion-00", 100034, NULL, 0, 0, 1582416000000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (110, "EmulsionTopLevel-00", 100035, NULL, 0, 0, 1582416000000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (114, "EmulsionSwitches-00", 100036, NULL, 0, 0, 1582416000000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (2, "Scintillator-00", 100001, NULL, 0, 0, 1549238400000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (2, "Scintillator-01", 100042, NULL, 0, 0, 1590796800000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (3, "Tracker-00", 100002, NULL, 0, 0, 1549238400000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (3, "Tracker-01", 100038, NULL, 0, 0, 1590796800000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (3, "Tracker-CR", 107783, NULL, 0, 0, 1598400000000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (31, "SCT-00", 100026, NULL, 0, 0, 1567987200000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (31, "SCT-01", 100037, NULL, 0, 0, 1159079680000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (31, "SCT-CR", 107781, NULL, 0, 0, 1598400000000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (32, "Dipole-00", 100027, NULL, 0, 0, 1568678400000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (32, "Dipole-01", 100041, NULL, 0, 0, 1590796800000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (310, "SctTopLevel-00", 106788, NULL, 0, 0, 1567987200000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (310, "SctTopLevel-01", 106790, NULL, 0, 0, 1590796800000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (310, "SctTopLevel-CR", 107786, NULL, 0, 0, 1598400000000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (311, "SctBarrelModule-00", 107003, NULL, 0, 0, 1567987200000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (312, "SctBarrelSensor-00", 106730, NULL, 0, 0, 1567987200000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (313, "SctFaserGeneral-00", 106789, NULL, 0, 0, 1567987200000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (313, "SctFaserGeneral-01", 106791, NULL, 0, 0, 1567987200000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (313, "SctFaserGeneral-CR", 107780, NULL, 0, 0, 1598400000000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (314, "SctSwitches-00", 107782, NULL, 0, 0, 1567987200000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (315, "SctMaterials-00", 107777, NULL, 0, 0, 1567987200000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (316, "SctMatComponents-00", 107778, NULL, 0, 0, 1567987200000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (317, "SctConditions-00", 107779, NULL, 0, 0, 1579564800000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (318, "SctFrame-00", 100053, NULL, 0, 0,        1591574400000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (3181,"SctFrameGeneral-00", 100054, NULL, 0, 0, 1591574400000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (3182,"SctFrameShape-00", 100055, NULL, 0, 0,   1591574400000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (320, "DipoleTopLevel-00", 100029, NULL, 0, 0, 1568678400000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (320, "DipoleTopLevel-01", 100040, NULL, 0, 0, 1590796800000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (321, "DipoleGeneral-00", 100004, NULL, 0, 0, 1568678400000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (324, "DipoleSwitches-00", 100028, NULL, 0, 0, 1568678400000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (4, "Calorimeter-00", 100003, NULL, 0, 0, 1549238400000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (41, "Ecal-00", 100056, NULL, 0, 0,    1593907200000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (410, "EcalTopLevel-00", 100058, NULL, 0, 0, 1599350400000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (411, "EcalRowGeneral-00", 100059, NULL, 0, 0, 1599350400000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (414, "EcalSwitches-00", 100057, NULL, 0, 0, 1593907200000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (210,  "VetoTopLevel-00", 100009, NULL, 0, 0, 1567123200000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (210,  "VetoTopLevel-01", 100046, NULL, 0, 0, 1590796800000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (211,  "VetoStationGeneral-00", 100010, NULL, 0, 0, 1567123200000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (212,  "VetoPlateGeneral-00", 100025, NULL, 0, 0, 1567209600000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (211,  "VetoStationGeneral-01", 100049, NULL, 0, 0, 1590796800000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (212,  "VetoPlateGeneral-01", 100050, NULL, 0, 0, 1590796800000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (220,  "TriggerTopLevel-00", 110009, NULL, 0, 0, 1581292800000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (220,  "TriggerTopLevel-01", 100047, NULL, 0, 0, 1590796800000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (221,  "TriggerStationGeneral-00", 110010, NULL, 0, 0, 1581292800000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (221,  "TriggerStationGeneral-01", 100051, NULL, 0, 0, 1590796800000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (222,  "TriggerPlateGeneral-00", 110025, NULL, 0, 0, 1581292800000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (230,  "PreshowerTopLevel-00", 120009, NULL, 0, 0, 1581292800000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (230,  "PreshowerTopLevel-01", 100048, NULL, 0, 0, 1590796800000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (231,  "PreshowerStationGeneral-00", 120010, NULL, 0, 0, 1581292800000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (231,  "PreshowerStationGeneral-01", 100052, NULL, 0, 0, 1590796800000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (232,  "PreshowerPlateGeneral-00", 120025, NULL, 0, 0, 1581292800000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (1003, "NeutrinoMaterials-00", 100032, NULL, 0, 0, 1582416000000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (1004, "NeutrinoMatComponents-00", 100033, NULL, 0, 0, 1582416000000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (2003, "ScintMaterials-00", 100011, NULL, 0, 0, 1549238400000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (2004, "ScintMatComponents-00", 100012, NULL, 0, 0, 1549238400000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (3003, "TrackerMaterials-00", 100021, NULL, 0, 0, 1550448000000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (3004, "TrackerMatComponents-00", 100022, NULL, 0, 0, 1550448000000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (4003, "CaloMaterials-00", 100023, NULL, 0, 0, 1550448000000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (4004, "CaloMatComponents-00", 100024, NULL, 0, 0, 1550448000000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (214,  "VetoSwitches-00", 100014, NULL, 0, 0, 1550361600000000000 ,NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (224,  "TriggerSwitches-00", 110014, NULL, 0, 0, 1581292800000000000 ,NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (234,  "PreshowerSwitches-00", 120014, NULL, 0, 0, 1550361600000000000 ,NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (21, "Veto-00", 100015, NULL, 0, 0, 1550448000000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (22, "Trigger-00", 100019, NULL, 0, 0, 1550448000000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (23, "Preshower-00", 100020, NULL, 0, 0, 1550448000000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (21, "Veto-01", 100043, NULL, 0, 0,      1590796800000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (22, "Trigger-01", 100044, NULL, 0, 0,   1590796800000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (23, "Preshower-01", 100045, NULL, 0, 0, 1590796800000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (1005, "NeutrinoIdentifier-00", 100030, NULL, 0, 0, 1582416000000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (2005, "ScintIdentifier-00", 100016, NULL, 0, 0, 1550448000000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (3005, "TrackerIdentifier-00", 100017, NULL, 0, 0, 1550448000000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (3005, "TrackerIdentifier-CR", 107785, NULL, 0, 0, 1598400000000000000, NULL, 22);
INSERT INTO "HVS_TAG2NODE" VALUES (4005, "CaloIdentifier-00", 100018, NULL, 0, 0, 1550448000000000000, NULL, 22);
-- Data for the HVS_LTAG2LTAG table
INSERT INTO "HVS_LTAG2LTAG" VALUES (0,   100000, 2,    100001);
INSERT INTO "HVS_LTAG2LTAG" VALUES (0,   100039, 2,    100042);
INSERT INTO "HVS_LTAG2LTAG" VALUES (0,   107784, 2,    100042);
INSERT INTO "HVS_LTAG2LTAG" VALUES (0,   100000, 3,    100002);
INSERT INTO "HVS_LTAG2LTAG" VALUES (0,   100039, 3,    100038);
INSERT INTO "HVS_LTAG2LTAG" VALUES (0,   107784, 3,    107783);
INSERT INTO "HVS_LTAG2LTAG" VALUES (0,   100000, 4,    100003);
INSERT INTO "HVS_LTAG2LTAG" VALUES (0,   100039, 4,    100003);
INSERT INTO "HVS_LTAG2LTAG" VALUES (0,   107784, 4,    100003);
INSERT INTO "HVS_LTAG2LTAG" VALUES (0,    100000, 9000,  100005);
INSERT INTO "HVS_LTAG2LTAG" VALUES (0,    100039, 9000,  100005);
INSERT INTO "HVS_LTAG2LTAG" VALUES (0,    107784, 9000,  100005);
INSERT INTO "HVS_LTAG2LTAG" VALUES (9000, 100005, 9001,  100006);
INSERT INTO "HVS_LTAG2LTAG" VALUES (9000, 100005, 9002,  100007);
INSERT INTO "HVS_LTAG2LTAG" VALUES (9000, 100005, 9003,  100008);
INSERT INTO "HVS_LTAG2LTAG" VALUES (1,   100031,  11,  100034);
INSERT INTO "HVS_LTAG2LTAG" VALUES (11,  100034, 110,  100035);
INSERT INTO "HVS_LTAG2LTAG" VALUES (11,  100034, 114,  100036);
INSERT INTO "HVS_LTAG2LTAG" VALUES (1,   100031, 1003, 100032);
INSERT INTO "HVS_LTAG2LTAG" VALUES (1,   100031, 1004, 100033);
INSERT INTO "HVS_LTAG2LTAG" VALUES (21,  100015, 210,  100009);
INSERT INTO "HVS_LTAG2LTAG" VALUES (21,  100043, 210,  100046);
INSERT INTO "HVS_LTAG2LTAG" VALUES (21,  100015, 211,  100010);
INSERT INTO "HVS_LTAG2LTAG" VALUES (21,  100015, 212,  100025);
INSERT INTO "HVS_LTAG2LTAG" VALUES (21,  100043, 211,  100049);
INSERT INTO "HVS_LTAG2LTAG" VALUES (21,  100043, 212,  100050);
INSERT INTO "HVS_LTAG2LTAG" VALUES (22,  100019, 220,  110009);
INSERT INTO "HVS_LTAG2LTAG" VALUES (22,  100044, 220,  100047);
INSERT INTO "HVS_LTAG2LTAG" VALUES (22,  100019, 221,  110010);
INSERT INTO "HVS_LTAG2LTAG" VALUES (22,  100044, 221,  100051);
INSERT INTO "HVS_LTAG2LTAG" VALUES (22,  100019, 222,  110025);
INSERT INTO "HVS_LTAG2LTAG" VALUES (22,  100044, 221,  110010);
INSERT INTO "HVS_LTAG2LTAG" VALUES (22,  100044, 222,  110025);
INSERT INTO "HVS_LTAG2LTAG" VALUES (23,  100020, 230,  120009);
INSERT INTO "HVS_LTAG2LTAG" VALUES (23,  100045, 230,  120048);
INSERT INTO "HVS_LTAG2LTAG" VALUES (23,  100020, 231,  120010);
INSERT INTO "HVS_LTAG2LTAG" VALUES (23,  100045, 231,  100052);
INSERT INTO "HVS_LTAG2LTAG" VALUES (23,  100020, 232,  120025);
INSERT INTO "HVS_LTAG2LTAG" VALUES (23,  100045, 232,  120025);
INSERT INTO "HVS_LTAG2LTAG" VALUES (2,   100001, 2003, 100011);
INSERT INTO "HVS_LTAG2LTAG" VALUES (2,   100001, 2004, 100012);
INSERT INTO "HVS_LTAG2LTAG" VALUES (2,   100042, 2003, 100011);
INSERT INTO "HVS_LTAG2LTAG" VALUES (2,   100042, 2004, 100012);
INSERT INTO "HVS_LTAG2LTAG" VALUES (3,   100002, 3003, 100021);
INSERT INTO "HVS_LTAG2LTAG" VALUES (3,   100002, 3004, 100022);
INSERT INTO "HVS_LTAG2LTAG" VALUES (3,   100038, 3003, 100021);
INSERT INTO "HVS_LTAG2LTAG" VALUES (3,   100038, 3004, 100022);
INSERT INTO "HVS_LTAG2LTAG" VALUES (3,   100002, 31,   100026);
INSERT INTO "HVS_LTAG2LTAG" VALUES (3,   100038, 31,   100037);
INSERT INTO "HVS_LTAG2LTAG" VALUES (3,   107783, 31,   107781);
INSERT INTO "HVS_LTAG2LTAG" VALUES (31,  100026, 310,  106788);
INSERT INTO "HVS_LTAG2LTAG" VALUES (31,  100037, 310,  106790);
INSERT INTO "HVS_LTAG2LTAG" VALUES (31,  100081, 310,  107786);
INSERT INTO "HVS_LTAG2LTAG" VALUES (31,  100026, 311,  107003);
INSERT INTO "HVS_LTAG2LTAG" VALUES (31,  100026, 312,  106730);
INSERT INTO "HVS_LTAG2LTAG" VALUES (31,  100026, 313,  106789);
INSERT INTO "HVS_LTAG2LTAG" VALUES (31,  100026, 314,  107782);
INSERT INTO "HVS_LTAG2LTAG" VALUES (31,  100026, 315,  107777);
INSERT INTO "HVS_LTAG2LTAG" VALUES (31,  100026, 316,  107778);
INSERT INTO "HVS_LTAG2LTAG" VALUES (31,  100026, 317,  107779);
INSERT INTO "HVS_LTAG2LTAG" VALUES (31,  100026, 318,  100053);
INSERT INTO "HVS_LTAG2LTAG" VALUES (31,  100037, 311,  107003);
INSERT INTO "HVS_LTAG2LTAG" VALUES (31,  100037, 312,  106730);
INSERT INTO "HVS_LTAG2LTAG" VALUES (31,  100037, 313,  106791);
INSERT INTO "HVS_LTAG2LTAG" VALUES (31,  107781, 313,  107780);
INSERT INTO "HVS_LTAG2LTAG" VALUES (31,  100037, 314,  107782);
INSERT INTO "HVS_LTAG2LTAG" VALUES (31,  100037, 315,  107777);
INSERT INTO "HVS_LTAG2LTAG" VALUES (31,  100037, 316,  107778);
INSERT INTO "HVS_LTAG2LTAG" VALUES (31,  100037, 317,  107779);
INSERT INTO "HVS_LTAG2LTAG" VALUES (31,  100037, 318,  100053);
INSERT INTO "HVS_LTAG2LTAG" VALUES (318, 100053, 3181, 100054);
INSERT INTO "HVS_LTAG2LTAG" VALUES (318, 100053, 3182, 100055);
INSERT INTO "HVS_LTAG2LTAG" VALUES (3,   100002, 32,   100027);
INSERT INTO "HVS_LTAG2LTAG" VALUES (3,   100038, 32,   100041);
INSERT INTO "HVS_LTAG2LTAG" VALUES (32,  100027, 320,  100029);
INSERT INTO "HVS_LTAG2LTAG" VALUES (32,  100027, 321,  100004);
INSERT INTO "HVS_LTAG2LTAG" VALUES (32,  100027, 324,  100028);
INSERT INTO "HVS_LTAG2LTAG" VALUES (32,  100041, 320,  100040);
INSERT INTO "HVS_LTAG2LTAG" VALUES (32,  100041, 321,  100004);
INSERT INTO "HVS_LTAG2LTAG" VALUES (32,  100041, 324,  100028);
INSERT INTO "HVS_LTAG2LTAG" VALUES (4,   100003, 41,   100056);
INSERT INTO "HVS_LTAG2LTAG" VALUES (41,  100056, 410,  100058);
INSERT INTO "HVS_LTAG2LTAG" VALUES (41,  100056, 411,  100059);
INSERT INTO "HVS_LTAG2LTAG" VALUES (41,  100056, 414,  100057);
INSERT INTO "HVS_LTAG2LTAG" VALUES (4,   100003, 4003, 100023);
INSERT INTO "HVS_LTAG2LTAG" VALUES (4,   100003, 4004, 100024);
INSERT INTO "HVS_LTAG2LTAG" VALUES (0,   100000, 90,   100013);
INSERT INTO "HVS_LTAG2LTAG" VALUES (0,   100039, 90,   100013);
INSERT INTO "HVS_LTAG2LTAG" VALUES (2,   100001, 21,   100015);
INSERT INTO "HVS_LTAG2LTAG" VALUES (2,   100001, 22,   100019);
INSERT INTO "HVS_LTAG2LTAG" VALUES (2,   100001, 23,   100020);
INSERT INTO "HVS_LTAG2LTAG" VALUES (2,   100042, 21,   100043);
INSERT INTO "HVS_LTAG2LTAG" VALUES (2,   100042, 22,   100044);
INSERT INTO "HVS_LTAG2LTAG" VALUES (2,   100042, 23,   100045);
INSERT INTO "HVS_LTAG2LTAG" VALUES (21,  100015, 214,  100014);
INSERT INTO "HVS_LTAG2LTAG" VALUES (21,  100043, 214,  100014);
INSERT INTO "HVS_LTAG2LTAG" VALUES (22,  100019, 224,  110014);
INSERT INTO "HVS_LTAG2LTAG" VALUES (22,  100044, 224,  110014);
INSERT INTO "HVS_LTAG2LTAG" VALUES (23,  100020, 234,  120014);
INSERT INTO "HVS_LTAG2LTAG" VALUES (23,  100045, 234,  120014);
INSERT INTO "HVS_LTAG2LTAG" VALUES (1,   100031, 1005, 100030);
INSERT INTO "HVS_LTAG2LTAG" VALUES (2,   100001, 2005, 100016);
INSERT INTO "HVS_LTAG2LTAG" VALUES (2,   100042, 2005, 100016);
INSERT INTO "HVS_LTAG2LTAG" VALUES (3,   100002, 3005, 100017);
INSERT INTO "HVS_LTAG2LTAG" VALUES (3,   100038, 3005, 100017);
INSERT INTO "HVS_LTAG2LTAG" VALUES (3,   107781, 3005, 107785);
INSERT INTO "HVS_LTAG2LTAG" VALUES (4,   100003, 4005, 100018);
-- Data for the HVS_TAGCACHE table
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "FASER",              "FASER-00",                100000);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "FaserCommon",        "FaserCommon-00",          100013);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "Materials",          "Materials-00",            100005);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "StdMaterials",       "StdMaterials-00",         100006);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "StdMatComponents",   "StdMatComponents-00",     100007);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "Elements",           "Elements-00",             100008);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "Neutrino",           "Neutrino-00",             100031);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "Emulsion",           "Emulsion-00",             100034);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "EmulsionTopLevel",   "EmulsionTopLevel-00",     100035);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "EmulsionSwitches",   "EmulsionSwitches-00",     100036);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "Scintillator",       "Scintillator-00",         100001);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "Tracker",            "Tracker-00",              100002);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "SCT",                "SCT-00",                  100026);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "SctTopLevel",        "SCTTopLevel-00",          106788);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "SctBrlModule",       "SCTBrlModule-00",         107003);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "SctBrlSensor",       "SCTBrlSensor-00",         106730);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "SctFaserGeneral",    "SCTFaserGeneral-00",      106789);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "SctSwitches",        "SCTSwitches-00",          107782);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "SCTMaterials",       "SCTMaterials-00",         107777);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "SCTMatComponents",   "SCTMatComponents-00",     107778);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "SctConditions",      "SctConditions-00",        107779);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "SctFrame",           "SctFrame-00",             100053);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "SctFrameGeneral",    "SctFrameGeneral-00",      100054);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "SctFrameShape",      "SctFrameShape-00",        100055);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "Dipole",             "Dipole-00",               100027);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "DipoleTopLevel",     "DipoleTopLevel-00",       100029);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "DipoleGeneral",      "DipoleGeneral-00",        100004);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "DipoleSwitches",     "DipoleSwitches-00",       100028);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "Calorimeter",        "Calorimeter-00",          100003);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "Ecal",               "Ecal-00",                 100056);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "EcalTopLevel",       "EcalTopLevel-00",         100058);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "EcalRowGeneral",     "EcalRowGeneral-00",       100059);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "EcalSwitches",       "EcalSwitches-00",         100057);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "VetoTopLevel",       "VetoTopLevel-00",         100009);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "VetoStationGeneral", "VetoStationGeneral-00",   100010);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "VetoPlateGeneral",   "VetoPlateGeneral-00",     100025);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "TriggerTopLevel",       "TriggerTopLevel-00",         110009);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "TriggerStationGeneral", "TriggerStationGeneral-00",   110010);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "TriggerPlateGeneral",   "TriggerPlateGeneral-00",     110025);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "PreshowerTopLevel",     "PreshowerTopLevel-00",         120009);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "PreshowerStationGeneral", "PreshowerStationGeneral-00",   120010);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "PreshowerPlateGeneral",   "PreshowerPlateGeneral-00",     120025);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "NeutrinoMaterials",     "NeutrinoMaterials-00",       100032);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "NeutrinoMatComponents", "NeutrinoMatComponents-00",   100033);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "ScintMaterials",     "ScintMaterials-00",       100011);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "ScintMatComponents", "ScintMatComponents-00",   100012);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "TrackerMaterials",     "TrackerMaterials-00",     100021);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "TrackerMatComponents", "TrackerMatComponents-00", 100022);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "CaloMaterials",     "CaloMaterials-00",         100023);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "CaloMatComponents", "CaloMatComponents-00",     100024);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "VetoSwitches",       "VetoSwitches-00",         100014);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "TriggerSwitches",    "TriggerSwitches-00",      110014);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "PreshowerSwitches",  "PreshowerSwitches-00",    120014);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "Veto",               "Veto-00",                 100015);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "Trigger",            "Trigger-00",              100019);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "Preshower",          "Preshower-00",            100020);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "NeutrinoIdentifier", "NeutrinoIdentifier-00",   100030);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "ScintIdentifier",    "ScintIdentifier-00",      100016);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "TrackerIdentifier",  "TrackerIdentifier-00",    100017);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-00", "CaloIdentifier",     "CaloIdentifier-00",       100018);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "FASER",              "FASER-01",                100039);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "FaserCommon",        "FaserCommon-00",          100013);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "Materials",          "Materials-00",            100005);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "StdMaterials",       "StdMaterials-00",         100006);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "StdMatComponents",   "StdMatComponents-00",     100007);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "Elements",           "Elements-00",             100008);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "Neutrino",           "Neutrino-00",             100031);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "Emulsion",           "Emulsion-00",             100034);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "EmulsionTopLevel",   "EmulsionTopLevel-00",     100035);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "EmulsionSwitches",   "EmulsionSwitches-00",     100036);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "Scintillator",       "Scintillator-01",         100042);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "Tracker",            "Tracker-01",              100038);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "SCT",                "SCT-01",                  100037);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "SctTopLevel",        "SCTTopLevel-01",          106790);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "SctBrlModule",       "SCTBrlModule-00",         107003);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "SctBrlSensor",       "SCTBrlSensor-00",         106730);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "SctFaserGeneral",    "SCTFaserGeneral-01",      106791);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "SctSwitches",        "SCTSwitches-00",          107782);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "SCTMaterials",       "SCTMaterials-00",         107777);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "SCTMatComponents",   "SCTMatComponents-00",     107778);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "SctConditions",      "SctConditions-00",        107779);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "SctFrame",           "SctFrame-00",             100053);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "SctFrameGeneral",    "SctFrameGeneral-00",      100054);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "SctFrameShape",      "SctFrameShape-00",        100055);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "Dipole",             "Dipole-01",               100041);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "DipoleTopLevel",     "DipoleTopLevel-01",       100040);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "DipoleGeneral",      "DipoleGeneral-00",        100004);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "DipoleSwitches",     "DipoleSwitches-00",       100028);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "Calorimeter",        "Calorimeter-00",          100003);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "Ecal",               "Ecal-00",                 100056);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "EcalTopLevel",       "EcalTopLevel-00",         100058);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "EcalRowGeneral",     "EcalRowGeneral-00",       100059);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "EcalSwitches",       "EcalSwitches-00",         100057);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "VetoTopLevel",       "VetoTopLevel-01",         100046);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "VetoStationGeneral", "VetoStationGeneral-01",   100049);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "VetoPlateGeneral",   "VetoPlateGeneral-01",     100050);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "TriggerTopLevel",       "TriggerTopLevel-01",         100047);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "TriggerStationGeneral", "TriggerStationGeneral-01",   100051);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "TriggerPlateGeneral",   "TriggerPlateGeneral-00",     110025);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "PreshowerTopLevel",     "PreshowerTopLevel-01",         100048);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "PreshowerStationGeneral", "PreshowerStationGeneral-01",   100052);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "PreshowerPlateGeneral",   "PreshowerPlateGeneral-00",     120025);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "NeutrinoMaterials",     "NeutrinoMaterials-00",       100032);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "NeutrinoMatComponents", "NeutrinoMatComponents-00",   100033);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "ScintMaterials",     "ScintMaterials-00",       100011);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "ScintMatComponents", "ScintMatComponents-00",   100012);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "TrackerMaterials",     "TrackerMaterials-00",     100021);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "TrackerMatComponents", "TrackerMatComponents-00", 100022);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "CaloMaterials",     "CaloMaterials-00",         100023);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "CaloMatComponents", "CaloMatComponents-00",     100024);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "VetoSwitches",       "VetoSwitches-00",         100014);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "TriggerSwitches",    "TriggerSwitches-00",      110014);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "PreshowerSwitches",  "PreshowerSwitches-00",    120014);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "Veto",               "Veto-01",                 100043);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "Trigger",            "Trigger-01",              100044);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "Preshower",          "Preshower-01",            100045);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "NeutrinoIdentifier", "NeutrinoIdentifier-00",   100030);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "ScintIdentifier",    "ScintIdentifier-00",      100016);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "TrackerIdentifier",  "TrackerIdentifier-00",    100017);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-01", "CaloIdentifier",     "CaloIdentifier-00",       100018);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "FASER",              "FASER-CR",                107784);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "FaserCommon",        "FaserCommon-00",          100013);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "Materials",          "Materials-00",            100005);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "StdMaterials",       "StdMaterials-00",         100006);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "StdMatComponents",   "StdMatComponents-00",     100007);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "Elements",           "Elements-00",             100008);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "Neutrino",           "Neutrino-00",             100031);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "Emulsion",           "Emulsion-00",             100034);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "EmulsionTopLevel",   "EmulsionTopLevel-00",     100035);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "EmulsionSwitches",   "EmulsionSwitches-00",     100036);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "Scintillator",       "Scintillator-01",         100042);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "Tracker",            "Tracker-CR",              107783);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "SCT",                "SCT-CR",                  107781);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "SctTopLevel",        "SCTTopLevel-CR",          107786);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "SctBrlModule",       "SCTBrlModule-00",         107003);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "SctBrlSensor",       "SCTBrlSensor-00",         106730);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "SctFaserGeneral",    "SCTFaserGeneral-CR",      107780);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "SctSwitches",        "SCTSwitches-00",          107782);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "SCTMaterials",       "SCTMaterials-00",         107777);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "SCTMatComponents",   "SCTMatComponents-00",     107778);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "SctConditions",      "SctConditions-00",        107779);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "SctFrame",           "SctFrame-00",             100053);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "SctFrameGeneral",    "SctFrameGeneral-00",      100054);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "SctFrameShape",      "SctFrameShape-00",        100055);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "Dipole",             "Dipole-01",               100041);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "DipoleTopLevel",     "DipoleTopLevel-01",       100040);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "DipoleGeneral",      "DipoleGeneral-00",        100004);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "DipoleSwitches",     "DipoleSwitches-00",       100028);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "Calorimeter",        "Calorimeter-00",          100003);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "Ecal",               "Ecal-00",                 100056);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "EcalTopLevel",       "EcalTopLevel-00",         100058);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "EcalRowGeneral",     "EcalRowGeneral-00",       100059);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "EcalSwitches",       "EcalSwitches-00",         100057);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "VetoTopLevel",       "VetoTopLevel-01",         100046);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "VetoStationGeneral", "VetoStationGeneral-01",   100049);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "VetoPlateGeneral",   "VetoPlateGeneral-01",     100050);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "TriggerTopLevel",       "TriggerTopLevel-01",         100047);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "TriggerStationGeneral", "TriggerStationGeneral-01",   100051);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "TriggerPlateGeneral",   "TriggerPlateGeneral-00",     110025);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "PreshowerTopLevel",     "PreshowerTopLevel-01",         100048);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "PreshowerStationGeneral", "PreshowerStationGeneral-01",   100052);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "PreshowerPlateGeneral",   "PreshowerPlateGeneral-00",     120025);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "NeutrinoMaterials",     "NeutrinoMaterials-00",       100032);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "NeutrinoMatComponents", "NeutrinoMatComponents-00",   100033);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "ScintMaterials",     "ScintMaterials-00",       100011);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "ScintMatComponents", "ScintMatComponents-00",   100012);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "TrackerMaterials",     "TrackerMaterials-00",     100021);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "TrackerMatComponents", "TrackerMatComponents-00", 100022);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "CaloMaterials",     "CaloMaterials-00",         100023);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "CaloMatComponents", "CaloMatComponents-00",     100024);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "VetoSwitches",       "VetoSwitches-00",         100014);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "TriggerSwitches",    "TriggerSwitches-00",      110014);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "PreshowerSwitches",  "PreshowerSwitches-00",    120014);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "Veto",               "Veto-01",                 100043);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "Trigger",            "Trigger-01",              100044);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "Preshower",          "Preshower-01",            100045);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "NeutrinoIdentifier", "NeutrinoIdentifier-00",   100030);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "ScintIdentifier",    "ScintIdentifier-00",      100016);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "TrackerIdentifier",  "TrackerIdentifier-CR",    107785);
INSERT INTO "HVS_TAGCACHE" VALUES ("FASER-CR", "CaloIdentifier",     "CaloIdentifier-00",       100018);
-- 
--
-- Part 2b: Content (Leaf node) data
--
-- Data for the standard material tables (copied directly from ATLAS geomDB)
INSERT INTO "STDMATERIALS_DATA" VALUES (0,'Air',0.001214);
INSERT INTO "STDMATERIALS_DATA" VALUES (1,'Vacuum',1.0e-06);
INSERT INTO "STDMATERIALS_DATA" VALUES (2,'Carbon',2.265);
INSERT INTO "STDMATERIALS_DATA" VALUES (3,'Aluminium',2.7);
INSERT INTO "STDMATERIALS_DATA" VALUES (4,'Silicon',2.33);
INSERT INTO "STDMATERIALS_DATA" VALUES (5,'Copper',8.96);
INSERT INTO "STDMATERIALS_DATA" VALUES (6,'Beryllia',3.01);
INSERT INTO "STDMATERIALS_DATA" VALUES (7,'Polystyrene',1.032);
INSERT INTO "STDMATERIALS_DATA" VALUES (8,'Polyethylene',0.95);
INSERT INTO "STDMATERIALS_DATA" VALUES (9,'C3F8',1.032);
INSERT INTO "STDMATERIALS_DATA" VALUES (10,'Kapton',1.46);
INSERT INTO "STDMATERIALS_DATA" VALUES (11,'G10',2.0);
INSERT INTO "STDMATERIALS_DATA" VALUES (12,'Bakelite',1.4);
INSERT INTO "STDMATERIALS_DATA" VALUES (13,'Iron',7.87);
INSERT INTO "STDMATERIALS_DATA" VALUES (14,'Beryllium',1.848);
INSERT INTO "STDMATERIALS_DATA" VALUES (15,'LiquidArgon',1.396);
INSERT INTO "STDMATERIALS_DATA" VALUES (16,'Lead',11.35);
INSERT INTO "STDMATERIALS_DATA" VALUES (17,'Quartz',2.62);
INSERT INTO "STDMATERIALS_DATA" VALUES (18,'Feldspar',2.56);
INSERT INTO "STDMATERIALS_DATA" VALUES (19,'Rock',2.33);
INSERT INTO "STDMATERIALS_DATA" VALUES (20,'Concrete',2.4);
INSERT INTO "STDMATERIALS_DATA" VALUES (21,'Mag_Concrete',3.67);
INSERT INTO "STDMATERIALS_DATA" VALUES (22,'Titanium',4.507);
INSERT INTO "STDMATERIALS_DATA" VALUES (23,'Mylar',1.39);
INSERT INTO "STDMATERIALS_DATA" VALUES (24,'Water',1.0);
INSERT INTO "STDMATERIALS_DATA" VALUES (25,'Aerogel',0.2);
INSERT INTO "STDMATERIALS_DATA" VALUES (26,'Pump_Ti',0.885);
INSERT INTO "STDMATERIALS_DATA" VALUES (27,'Aerogel',0.182);
INSERT INTO "STDMATERIALS_DATA" VALUES (28,'Getter',5.8);
INSERT INTO "STDMATERIALS_DATA" VALUES (29,'Inconel',8.47);
INSERT INTO "STDMATERIALS_DATA" VALUES (30,'SSteel',7.87);
INSERT INTO "STDMATERIALS_DATA" VALUES (31,'Brass',1.0);
INSERT INTO "STDMATERIALS_DATA" VALUES (32,'Pump_Ti',1.7);
INSERT INTO "STDMATERIALS_DATA" VALUES (33,'Inconel_H1',4.235);
INSERT INTO "STDMATERIALS_DATA" VALUES (34,'TiO2',4.23);
INSERT INTO "STDMATERIALS_DATA" VALUES (35,'AerogelBase',2.7);
INSERT INTO "STDMATERIALS_DATA" VALUES (36,'AerogelRun2',0.212);
INSERT INTO "STDMATERIALS_DATA" VALUES (37,'Polyester',1.4);
INSERT INTO "STDMATERIALS_DATA" VALUES (38,'Viscose',1.51);
INSERT INTO "STDMATERIALS_DATA" VALUES (39,'Valmat',1.25);
INSERT INTO "STDMATERIALS_DATA" VALUES (40,'Silica',0.7);
INSERT INTO "STDMATERIALS_DATA" VALUES (41,'AlH6O3',2.42);
INSERT INTO "STDMATERIALS_DATA" VALUES (42,'PyrogelXT',0.18);
INSERT INTO "STDMATERIALS_DATA" VALUES (43,'Wolfram',19.3);
INSERT INTO "STDMATERIALS_DATA" VALUES (44,'Nd2Fe14B',5.8);
--
INSERT INTO "STDMATERIALS_DATA2TAG" VALUES (100006, 0);
INSERT INTO "STDMATERIALS_DATA2TAG" VALUES (100006, 1);
INSERT INTO "STDMATERIALS_DATA2TAG" VALUES (100006, 2);
INSERT INTO "STDMATERIALS_DATA2TAG" VALUES (100006, 3);
INSERT INTO "STDMATERIALS_DATA2TAG" VALUES (100006, 4);
INSERT INTO "STDMATERIALS_DATA2TAG" VALUES (100006, 5);
INSERT INTO "STDMATERIALS_DATA2TAG" VALUES (100006, 6);
INSERT INTO "STDMATERIALS_DATA2TAG" VALUES (100006, 7);
INSERT INTO "STDMATERIALS_DATA2TAG" VALUES (100006, 8);
INSERT INTO "STDMATERIALS_DATA2TAG" VALUES (100006, 9);
INSERT INTO "STDMATERIALS_DATA2TAG" VALUES (100006, 10);
INSERT INTO "STDMATERIALS_DATA2TAG" VALUES (100006, 11);
INSERT INTO "STDMATERIALS_DATA2TAG" VALUES (100006, 12);
INSERT INTO "STDMATERIALS_DATA2TAG" VALUES (100006, 13);
INSERT INTO "STDMATERIALS_DATA2TAG" VALUES (100006, 14);
INSERT INTO "STDMATERIALS_DATA2TAG" VALUES (100006, 15);
INSERT INTO "STDMATERIALS_DATA2TAG" VALUES (100006, 16);
INSERT INTO "STDMATERIALS_DATA2TAG" VALUES (100006, 17);
INSERT INTO "STDMATERIALS_DATA2TAG" VALUES (100006, 18);
INSERT INTO "STDMATERIALS_DATA2TAG" VALUES (100006, 19);
INSERT INTO "STDMATERIALS_DATA2TAG" VALUES (100006, 20);
INSERT INTO "STDMATERIALS_DATA2TAG" VALUES (100006, 21);
INSERT INTO "STDMATERIALS_DATA2TAG" VALUES (100006, 22);
INSERT INTO "STDMATERIALS_DATA2TAG" VALUES (100006, 23);
INSERT INTO "STDMATERIALS_DATA2TAG" VALUES (100006, 24);
INSERT INTO "STDMATERIALS_DATA2TAG" VALUES (100006, 25);
INSERT INTO "STDMATERIALS_DATA2TAG" VALUES (100006, 26);
INSERT INTO "STDMATERIALS_DATA2TAG" VALUES (100006, 27);
INSERT INTO "STDMATERIALS_DATA2TAG" VALUES (100006, 28);
INSERT INTO "STDMATERIALS_DATA2TAG" VALUES (100006, 29);
INSERT INTO "STDMATERIALS_DATA2TAG" VALUES (100006, 30);
INSERT INTO "STDMATERIALS_DATA2TAG" VALUES (100006, 31);
INSERT INTO "STDMATERIALS_DATA2TAG" VALUES (100006, 32);
INSERT INTO "STDMATERIALS_DATA2TAG" VALUES (100006, 33);
INSERT INTO "STDMATERIALS_DATA2TAG" VALUES (100006, 34);
INSERT INTO "STDMATERIALS_DATA2TAG" VALUES (100006, 35);
INSERT INTO "STDMATERIALS_DATA2TAG" VALUES (100006, 36);
INSERT INTO "STDMATERIALS_DATA2TAG" VALUES (100006, 37);
INSERT INTO "STDMATERIALS_DATA2TAG" VALUES (100006, 38);
INSERT INTO "STDMATERIALS_DATA2TAG" VALUES (100006, 39);
INSERT INTO "STDMATERIALS_DATA2TAG" VALUES (100006, 40);
INSERT INTO "STDMATERIALS_DATA2TAG" VALUES (100006, 41);
INSERT INTO "STDMATERIALS_DATA2TAG" VALUES (100006, 42);
INSERT INTO "STDMATERIALS_DATA2TAG" VALUES (100006, 43);
INSERT INTO "STDMATERIALS_DATA2TAG" VALUES (100006, 44);
--
--
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (0,0,'Nitrogen',0.7494);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (1,0,'Oxygen',0.2369);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (2,0,'Argon',0.0129);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (3,0,'Hydrogen',0.0008);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (4,1,'Nitrogen',0.7494);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (5,1,'Oxygen',0.2369);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (6,1,'Argon',0.0129);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (7,1,'Hydrogen',0.0008);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (8,2,'Carbon',1.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (9,3,'Aluminium',1.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (10,4,'Silicon',1.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (11,5,'Copper',1.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (12,6,'Beryllium',0.33);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (13,6,'Oxygen',0.66);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (14,7,'Carbon',1.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (15,7,'Hydrogen',1.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (16,8,'Carbon',1.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (17,8,'Hydrogen',2.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (18,9,'Carbon',3.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (19,9,'Fluorine',8.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (20,10,'Carbon',22.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (21,10,'Hydrogen',10.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (22,10,'Oxygen',5.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (23,10,'Nitrogen',2.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (24,11,'Carbon',2.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (25,11,'Hydrogen',2.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (26,11,'Fluorine',4.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (27,12,'Carbon',1.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (28,12,'Hydrogen',4.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (29,12,'Oxygen',2.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (30,13,'Iron',1.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (31,14,'Beryllium',1.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (32,15,'Argon',1.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (33,16,'Lead',1.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (34,17,'Silicon',0.47);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (35,17,'Oxygen',0.53);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (36,18,'Oxygen',0.46);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (37,18,'Silicon',0.3);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (38,18,'Potassium',0.14);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (39,18,'Aluminium',0.1);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (40,19,'std::Quartz',0.83);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (41,19,'std::Feldspar',0.17);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (42,20,'Oxygen',0.5);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (43,20,'Calcium',0.2);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (44,20,'Silicon',0.2);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (45,20,'Carbon',0.03);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (46,20,'Aluminium',0.03);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (47,20,'Iron',0.014);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (48,20,'Potassium',0.01);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (49,20,'Sodium',0.01);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (50,20,'Hydrogen',0.006);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (51,21,'Oxygen',0.35);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (52,21,'Calcium',0.027);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (53,21,'Silicon',0.02);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (54,21,'Aluminium',0.02);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (55,21,'Iron',0.55);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (56,21,'Sodium',0.013);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (57,21,'Hydrogen',0.007);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (58,21,'Boron',0.013);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (59,22,'Titanium',1.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (60,23,'Carbon',5.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (61,23,'Oxygen',2.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (62,23,'Hydrogen',4.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (63,24,'Hydrogen',2.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (64,24,'Oxygen',1.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (65,25,'std::Quartz',0.625);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (66,25,'std::Water',0.374);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (67,25,'Carbon',0.001);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (68,26,'Titanium',1.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (69,27,'std::Quartz',0.97);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (70,27,'std::Water',0.03);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (71,28,'Titanium',3.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (72,28,'Zirconium',3.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (73,28,'Vanadium',4.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (74,29,'Nickel',0.75);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (75,29,'Chromium',0.16);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (76,29,'Iron',0.08);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (77,29,'Manganese',0.005);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (78,29,'Silicon',0.0025);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (79,29,'Copper',0.0025);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (80,30,'Iron',0.74);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (81,30,'Chromium',0.18);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (82,30,'Nickel',0.08);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (83,31,'Copper',0.9);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (84,31,'Zinc',0.1);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (85,32,'Titanium',1.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (86,33,'std::Inconel',1.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (87,34,'Titanium',0.599);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (88,34,'Oxygen',0.401);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (89,35,'std::Quartz',0.919);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (90,35,'std::TiO2',0.081);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (91,36,'std::AerogelBase',0.97);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (92,36,'std::Water',0.03);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (93,37,'Carbon',15.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (94,37,'Hydrogen',12.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (95,37,'Bromine',4.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (96,37,'Oxygen',2.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (97,38,'Carbon',6.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (98,38,'Hydrogen',10.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (99,38,'Oxygen',5.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (100,39,'Polyester',0.3);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (101,39,'Viscose',0.7);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (102,40,'Silicon',1.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (103,40,'Oxygen',3.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (104,40,'Hydrogen',2.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (105,40,'Chlorine',2.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (106,40,'Cobalt',2.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (107,41,'Aluminium',6.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (108,41,'Hydrogen',6.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (109,41,'Oxygen',3.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (110,42,'Silica',0.45);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (111,42,'Quartz',0.45);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (112,42,'TiO2',0.7);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (113,42,'AlH6O3',0.7);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (114,42,'TiO2',0.05);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (115,42,'AlH6O3',0.05);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (116,39,'std::Polyester',0.3);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (117,39,'std:Viscose',0.7);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (118,42,'std::Silica',0.45);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (119,42,'std::Quartz',0.45);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (120,42,'std::TiO2',0.05);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (121,42,'std::AlH6O3',0.05);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (122,39,'std::Viscose',0.7);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (123,43,'Wolfram',1.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (124,44,'Neodymium',2.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (125,44,'Iron',14.0);
INSERT INTO "STDMATCOMPONENTS_DATA" VALUES (126,44,'Boron',1.0);
--
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 0);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 1);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 2);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 3);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 4);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 5);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 6);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 7);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 8);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 9);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 10);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 11);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 12);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 13);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 14);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 15);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 16);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 17);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 18);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 19);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 20);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 21);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 22);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 23);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 24);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 25);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 26);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 27);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 28);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 29);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 30);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 31);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 32);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 33);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 34);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 35);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 36);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 37);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 38);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 39);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 40);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 41);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 42);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 43);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 44);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 45);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 46);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 47);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 48);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 49);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 50);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 51);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 52);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 53);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 54);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 55);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 56);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 57);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 58);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 59);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 60);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 61);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 62);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 63);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 64);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 65);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 66);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 67);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 68);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 69);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 70);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 71);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 72);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 73);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 74);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 75);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 76);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 77);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 78);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 79);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 80);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 81);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 82);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 83);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 84);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 85);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 86);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 87);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 88);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 89);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 90);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 91);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 92);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 93);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 94);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 95);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 96);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 97);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 98);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 99);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 100);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 101);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 102);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 103);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 104);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 105);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 106);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 107);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 108);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 109);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 110);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 111);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 112);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 113);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 114);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 115);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 116);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 117);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 118);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 119);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 120);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 121);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 122);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 123);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 124);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 125);
INSERT INTO "STDMATCOMPONENTS_DATA2TAG" VALUES (100007, 126);
--
--
INSERT INTO "ELEMENTS_DATA" VALUES (0,'Hydrogen','H',1.00797,1.0);
INSERT INTO "ELEMENTS_DATA" VALUES (1,'Helium','He',4.0026,2.0);
INSERT INTO "ELEMENTS_DATA" VALUES (2,'Lithium','Li',6.939,3.0);
INSERT INTO "ELEMENTS_DATA" VALUES (3,'Beryllium','Be',9.0122,4.0);
INSERT INTO "ELEMENTS_DATA" VALUES (4,'Boron','B',10.811,5.0);
INSERT INTO "ELEMENTS_DATA" VALUES (5,'Carbon','C',12.0112,6.0);
INSERT INTO "ELEMENTS_DATA" VALUES (6,'Nitrogen','N',14.0067,7.0);
INSERT INTO "ELEMENTS_DATA" VALUES (7,'Oxygen','O',15.9994,8.0);
INSERT INTO "ELEMENTS_DATA" VALUES (8,'Fluorine','F',18.9984,9.0);
INSERT INTO "ELEMENTS_DATA" VALUES (9,'Neon','Ne',20.183,10.0);
INSERT INTO "ELEMENTS_DATA" VALUES (10,'Sodium','Na',22.9898,11.0);
INSERT INTO "ELEMENTS_DATA" VALUES (11,'Magnesium','Mg',24.312,12.0);
INSERT INTO "ELEMENTS_DATA" VALUES (12,'Aluminium','Al',26.9815,13.0);
INSERT INTO "ELEMENTS_DATA" VALUES (13,'Silicon','Si',28.086,14.0);
INSERT INTO "ELEMENTS_DATA" VALUES (14,'Phosphorus','P',30.984,15.0);
INSERT INTO "ELEMENTS_DATA" VALUES (15,'Sulfur','S',32.064,16.0);
INSERT INTO "ELEMENTS_DATA" VALUES (16,'Chlorine','Cl',35.453,17.0);
INSERT INTO "ELEMENTS_DATA" VALUES (17,'Argon','Ar',39.948,18.0);
INSERT INTO "ELEMENTS_DATA" VALUES (18,'Potassium','K',39.102,19.0);
INSERT INTO "ELEMENTS_DATA" VALUES (19,'Calcium','Ca',40.08,20.0);
INSERT INTO "ELEMENTS_DATA" VALUES (20,'Scandium','Sc',44.956,21.0);
INSERT INTO "ELEMENTS_DATA" VALUES (21,'Titanium','Ti',47.9,22.0);
INSERT INTO "ELEMENTS_DATA" VALUES (22,'Vanadium','V',50.942,23.0);
INSERT INTO "ELEMENTS_DATA" VALUES (23,'Chromium','Cr',51.996,24.0);
INSERT INTO "ELEMENTS_DATA" VALUES (24,'Manganese','Mn',54.938,25.0);
INSERT INTO "ELEMENTS_DATA" VALUES (25,'Iron','Fe',55.847,26.0);
INSERT INTO "ELEMENTS_DATA" VALUES (26,'Cobalt','Co',58.933,27.0);
INSERT INTO "ELEMENTS_DATA" VALUES (27,'Nickel','Ni',58.71,28.0);
INSERT INTO "ELEMENTS_DATA" VALUES (28,'Copper','Cu',63.54,29.0);
INSERT INTO "ELEMENTS_DATA" VALUES (29,'Zinc','Zn',65.37,30.0);
INSERT INTO "ELEMENTS_DATA" VALUES (30,'Gallium','Ga',69.72,31.0);
INSERT INTO "ELEMENTS_DATA" VALUES (31,'Germanium','Ge',72.59,32.0);
INSERT INTO "ELEMENTS_DATA" VALUES (32,'Arsenic','As',74.992,33.0);
INSERT INTO "ELEMENTS_DATA" VALUES (33,'Selenium','Se',78.96,34.0);
INSERT INTO "ELEMENTS_DATA" VALUES (34,'Bromine','Br',79.909,35.0);
INSERT INTO "ELEMENTS_DATA" VALUES (35,'Krypton','Kr',83.8,36.0);
INSERT INTO "ELEMENTS_DATA" VALUES (36,'Rubidium','Rb',85.47,37.0);
INSERT INTO "ELEMENTS_DATA" VALUES (37,'Strontium','Sr',87.62,38.0);
INSERT INTO "ELEMENTS_DATA" VALUES (38,'Yttrium','Y',88.9059,39.0);
INSERT INTO "ELEMENTS_DATA" VALUES (39,'Zirconium','Zr',91.22,40.0);
INSERT INTO "ELEMENTS_DATA" VALUES (40,'Niobium','Nb',92.906,41.0);
INSERT INTO "ELEMENTS_DATA" VALUES (41,'Molybdenum','Mo',95.94,42.0);
INSERT INTO "ELEMENTS_DATA" VALUES (42,'Technetium','Tc',96.9062,43.0);
INSERT INTO "ELEMENTS_DATA" VALUES (43,'Ruthenium','Ru',101.07,44.0);
INSERT INTO "ELEMENTS_DATA" VALUES (44,'Rhodium','Rh',102.905,45.0);
INSERT INTO "ELEMENTS_DATA" VALUES (45,'Palladium','Pd',106.4,46.0);
INSERT INTO "ELEMENTS_DATA" VALUES (46,'Silver','Ag',107.87,47.0);
INSERT INTO "ELEMENTS_DATA" VALUES (47,'Cadmium','Cd',112.4,48.0);
INSERT INTO "ELEMENTS_DATA" VALUES (48,'Indium','In',114.82,49.0);
INSERT INTO "ELEMENTS_DATA" VALUES (49,'Tin','Sn',118.69,50.0);
INSERT INTO "ELEMENTS_DATA" VALUES (50,'Antimony','Sb',121.75,51.0);
INSERT INTO "ELEMENTS_DATA" VALUES (51,'Tellurium','Te',127.6,52.0);
INSERT INTO "ELEMENTS_DATA" VALUES (52,'Iodine','I',126.904,53.0);
INSERT INTO "ELEMENTS_DATA" VALUES (53,'Xenon','Xe',131.3,54.0);
INSERT INTO "ELEMENTS_DATA" VALUES (54,'Cesium','Cs',132.905,55.0);
INSERT INTO "ELEMENTS_DATA" VALUES (55,'Barium','Ba',137.34,56.0);
INSERT INTO "ELEMENTS_DATA" VALUES (56,'Lanthanum','La',138.91,57.0);
INSERT INTO "ELEMENTS_DATA" VALUES (57,'Cerium','Ce',140.12,58.0);
INSERT INTO "ELEMENTS_DATA" VALUES (58,'Praseodymium','Pr',140.907,59.0);
INSERT INTO "ELEMENTS_DATA" VALUES (59,'Neodymium','Nd',144.24,60.0);
INSERT INTO "ELEMENTS_DATA" VALUES (60,'Promethium','Pm',144.9128,61.0);
INSERT INTO "ELEMENTS_DATA" VALUES (61,'Samarium','Sm',150.35,62.0);
INSERT INTO "ELEMENTS_DATA" VALUES (62,'Europium','Eu',151.96,63.0);
INSERT INTO "ELEMENTS_DATA" VALUES (63,'Gadolinium','Gd',157.25,64.0);
INSERT INTO "ELEMENTS_DATA" VALUES (64,'Terbium','Tb',158.924,65.0);
INSERT INTO "ELEMENTS_DATA" VALUES (65,'Dysprosium','Dy',162.5,66.0);
INSERT INTO "ELEMENTS_DATA" VALUES (66,'Holmium','Ho',164.93,67.0);
INSERT INTO "ELEMENTS_DATA" VALUES (67,'Erbium','Er',167.26,68.0);
INSERT INTO "ELEMENTS_DATA" VALUES (68,'Thulium','Tm',168.934,69.0);
INSERT INTO "ELEMENTS_DATA" VALUES (69,'Ytterbium','Yb',173.04,70.0);
INSERT INTO "ELEMENTS_DATA" VALUES (70,'Lutetium','Lu',174.97,71.0);
INSERT INTO "ELEMENTS_DATA" VALUES (71,'Hafnium','Hf',178.49,72.0);
INSERT INTO "ELEMENTS_DATA" VALUES (72,'Tantalum','Ta',180.948,73.0);
INSERT INTO "ELEMENTS_DATA" VALUES (73,'Wolfram','W',183.85,74.0);
INSERT INTO "ELEMENTS_DATA" VALUES (74,'Rhenium','Re',186.2,75.0);
INSERT INTO "ELEMENTS_DATA" VALUES (75,'Osmium','Os',190.2,76.0);
INSERT INTO "ELEMENTS_DATA" VALUES (76,'Iridium','Ir',192.22,77.0);
INSERT INTO "ELEMENTS_DATA" VALUES (77,'Platinum','Pt',195.09,78.0);
INSERT INTO "ELEMENTS_DATA" VALUES (78,'Gold','Au',196.967,79.0);
INSERT INTO "ELEMENTS_DATA" VALUES (79,'Mercury','Hg',200.59,80.0);
INSERT INTO "ELEMENTS_DATA" VALUES (80,'Thallium','Tl',204.37,81.0);
INSERT INTO "ELEMENTS_DATA" VALUES (81,'Lead','Pb',207.19,82.0);
INSERT INTO "ELEMENTS_DATA" VALUES (82,'Bismuth','Bi',208.98,83.0);
INSERT INTO "ELEMENTS_DATA" VALUES (83,'Polonium','Po',208.98,84.0);
INSERT INTO "ELEMENTS_DATA" VALUES (84,'Astatine','At',209.987,85.0);
INSERT INTO "ELEMENTS_DATA" VALUES (85,'Radon','Rn',222.0176,86.0);
INSERT INTO "ELEMENTS_DATA" VALUES (86,'Francium','Fr',223.0189,87.0);
INSERT INTO "ELEMENTS_DATA" VALUES (87,'Radium','Ra',226.0,88.0);
INSERT INTO "ELEMENTS_DATA" VALUES (88,'Actinium','Ac',227.0278,89.0);
INSERT INTO "ELEMENTS_DATA" VALUES (89,'Thorium','Th',232.038,90.0);
INSERT INTO "ELEMENTS_DATA" VALUES (90,'Protactinium','Pa',231.0359,91.0);
INSERT INTO "ELEMENTS_DATA" VALUES (91,'Uranium','U',238.03,92.0);
--
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 0);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 1);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 2);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 3);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 4);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 5);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 6);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 7);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 8);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 9);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 10);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 11);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 12);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 13);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 14);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 15);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 16);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 17);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 18);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 19);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 20);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 21);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 22);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 23);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 24);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 25);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 26);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 27);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 28);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 29);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 30);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 31);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 32);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 33);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 34);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 35);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 36);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 37);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 38);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 39);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 40);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 41);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 42);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 43);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 44);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 45);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 46);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 47);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 48);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 49);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 50);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 51);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 52);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 53);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 54);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 55);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 56);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 57);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 58);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 59);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 60);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 61);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 62);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 63);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 64);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 65);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 66);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 67);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 68);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 69);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 70);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 71);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 72);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 73);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 74);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 75);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 76);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 77);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 78);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 79);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 80);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 81);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 82);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 83);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 84);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 85);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 86);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 87);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 88);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 89);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 90);
INSERT INTO "ELEMENTS_DATA2TAG" VALUES (100008, 91);
--
--
INSERT INTO "VETOTOPLEVEL_DATA" VALUES (0, 0.0, 0.0, -1725.0, 0.0, 0.0, 0.0, 321, "Veto");
INSERT INTO "VETOTOPLEVEL_DATA" VALUES (1, 0.0, 0.0, -160.0, 0.0, 0.0, 0.0, 321, "StationA");
INSERT INTO "VETOTOPLEVEL_DATA" VALUES (2, 0.0, 0.0,  160.0, 0.0, 0.0, 0.0, 321, "StationB");
INSERT INTO "VETOTOPLEVEL_DATA" VALUES (3, 0.0, 0.0, -1709.0, 0.0, 0.0, 0.0, 321, "Veto");
INSERT INTO "VETOTOPLEVEL_DATA" VALUES (4, 0.0, 0.0, -100.0, 0.0, 0.0, 0.0, 321, "StationA");
INSERT INTO "VETOTOPLEVEL_DATA" VALUES (5, 0.0, 0.0,  100.0, 0.0, 0.0, 0.0, 321, "StationB");
INSERT INTO "VETOTOPLEVEL_DATA2TAG" VALUES (100009, 0);
INSERT INTO "VETOTOPLEVEL_DATA2TAG" VALUES (100009, 1);
INSERT INTO "VETOTOPLEVEL_DATA2TAG" VALUES (100009, 2);
INSERT INTO "VETOTOPLEVEL_DATA2TAG" VALUES (100046, 3);
INSERT INTO "VETOTOPLEVEL_DATA2TAG" VALUES (100046, 4);
INSERT INTO "VETOTOPLEVEL_DATA2TAG" VALUES (100046, 5);
--
--
INSERT INTO "VETOSTATIONGENERAL_DATA" VALUES (0, 2, 100.0);
INSERT INTO "VETOSTATIONGENERAL_DATA" VALUES (1, 2, 20.0);
INSERT INTO "VETOSTATIONGENERAL_DATA2TAG" VALUES (100010, 0);
INSERT INTO "VETOSTATIONGENERAL_DATA2TAG" VALUES (100049, 1);
--
INSERT INTO "VETOPLATEGENERAL_DATA" VALUES (0, 1, 300.0, 300.0, 20.0, "scint::Scintillator");
INSERT INTO "VETOPLATEGENERAL_DATA" VALUES (1, 1, 300.0, 300.0, 19.5, "scint::Scintillator");
INSERT INTO "VETOPLATEGENERAL_DATA2TAG" VALUES (100025, 0);
INSERT INTO "VETOPLATEGENERAL_DATA2TAG" VALUES (100050, 1);
--
--
INSERT INTO "TRIGGERTOPLEVEL_DATA" VALUES (0, 0.0, 0.0, 187.0, 0.0, 0.0, 0.0, 321, "Trigger");
INSERT INTO "TRIGGERTOPLEVEL_DATA" VALUES (1, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 321, "StationA");
INSERT INTO "TRIGGERTOPLEVEL_DATA" VALUES (2, 0.0, -5.0, -28.9, 0.0, 0.0, 0.0, 321, "Trigger");
INSERT INTO "TRIGGERTOPLEVEL_DATA2TAG" VALUES (110009, 0);
INSERT INTO "TRIGGERTOPLEVEL_DATA2TAG" VALUES (110009, 1);
INSERT INTO "TRIGGERTOPLEVEL_DATA2TAG" VALUES (100047, 1);
INSERT INTO "TRIGGERTOPLEVEL_DATA2TAG" VALUES (100047, 2);
--
--
INSERT INTO "TRIGGERSTATIONGENERAL_DATA" VALUES (0, 2, 195.0, 11.0);
INSERT INTO "TRIGGERSTATIONGENERAL_DATA" VALUES (1, 2, 196.0, 14.9);
INSERT INTO "TRIGGERSTATIONGENERAL_DATA2TAG" VALUES (110010, 0);
INSERT INTO "TRIGGERSTATIONGENERAL_DATA2TAG" VALUES (100051, 1);
--
INSERT INTO "TRIGGERPLATEGENERAL_DATA" VALUES (0, 2, 400.0, 200.0, 10.0, "scint::Scintillator");
INSERT INTO "TRIGGERPLATEGENERAL_DATA2TAG" VALUES (110025, 0);
--
--
INSERT INTO "PRESHOWERTOPLEVEL_DATA" VALUES (0, 0.0, 0.0, 2626.0, 0.0, 0.0, 0.0, 321, "Preshower");
INSERT INTO "PRESHOWERTOPLEVEL_DATA" VALUES (1, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 321, "StationA");
INSERT INTO "PRESHOWERTOPLEVEL_DATA" VALUES (2, 0.0, 0.0, 2662.7, 0.0, 0.0, 0.0, 321, "Preshower");
INSERT INTO "PRESHOWERTOPLEVEL_DATA2TAG" VALUES (120009, 0);
INSERT INTO "PRESHOWERTOPLEVEL_DATA2TAG" VALUES (120009, 1);
INSERT INTO "PRESHOWERTOPLEVEL_DATA2TAG" VALUES (100048, 1);
INSERT INTO "PRESHOWERTOPLEVEL_DATA2TAG" VALUES (100048, 2);
--
--
INSERT INTO "PRESHOWERSTATIONGENERAL_DATA" VALUES (0, 2, 80.0);
INSERT INTO "PRESHOWERSTATIONGENERAL_DATA" VALUES (1, 2, 82.0);
INSERT INTO "PRESHOWERSTATIONGENERAL_DATA2TAG" VALUES (120010, 0);
INSERT INTO "PRESHOWERSTATIONGENERAL_DATA2TAG" VALUES (100052, 1);
--
INSERT INTO "PRESHOWERPLATEGENERAL_DATA" VALUES (0, 1, 300.0, 300.0, 20.0, "scint::Scintillator");
INSERT INTO "PRESHOWERPLATEGENERAL_DATA2TAG" VALUES (120025, 0);
--
--
INSERT INTO "ECALTOPLEVEL_DATA" VALUES (0, 0.0,   0.0, 3099.2, 0.0, 0.0, 0.0, 321, "Ecal");
INSERT INTO "ECALTOPLEVEL_DATA" VALUES (1, 0.0, -71.6,    0.0, 0.0, 2.8, 0.0, 321, "BottomRow");
INSERT INTO "ECALTOPLEVEL_DATA" VALUES (2, 0.0,  49.6,    0.0, 0.0, 2.8, 0.0, 321, "TopRow");
INSERT INTO "ECALTOPLEVEL_DATA2TAG" VALUES (100058, 0);
INSERT INTO "ECALTOPLEVEL_DATA2TAG" VALUES (100058, 1);
INSERT INTO "ECALTOPLEVEL_DATA2TAG" VALUES (100058, 2);
--
--
INSERT INTO "ECALROWGENERAL_DATA" VALUES (0, 2, 0.0, 2.8);
INSERT INTO "ECALROWGENERAL_DATA2TAG" VALUES (100059, 0);
--
--
INSERT INTO "SCINTMATERIALS_DATA" VALUES (0, "Scintillator", 1.032);
INSERT INTO "SCINTMATERIALS_DATA2TAG" VALUES (100011, 0);
--
--
INSERT INTO "SCINTMATCOMPONENTS_DATA" VALUES (0, 0, "Carbon", 0.475);
INSERT INTO "SCINTMATCOMPONENTS_DATA" VALUES (1, 0, "Hydrogen", 0.525);
INSERT INTO "SCINTMATCOMPONENTS_DATA2TAG" VALUES (100012, 0);
INSERT INTO "SCINTMATCOMPONENTS_DATA2TAG" VALUES (100012, 1);
--
--
INSERT INTO "FASERCOMMON_DATA" VALUES (0, "RUN3", "TRACKER", "NONE");
INSERT INTO "FASERCOMMON_DATA2TAG" VALUES (100013, 0);
--
--
INSERT INTO "VETOSWITCHES_DATA" VALUES  (0, "Veto", 1, 0, "GEO", "Development", "Baseline geometry");
INSERT INTO "VETOSWITCHES_DATA2TAG" VALUES (100014, 0);
--
--
INSERT INTO "TRIGGERSWITCHES_DATA" VALUES  (0, "Trigger", 1, 0, "GEO", "Development", "Baseline geometry");
INSERT INTO "TRIGGERSWITCHES_DATA2TAG" VALUES (110014, 0);
--
--
INSERT INTO "PRESHOWERSWITCHES_DATA" VALUES  (0, "Preshower", 1, 0, "GEO", "Development", "Baseline geometry");
INSERT INTO "PRESHOWERSWITCHES_DATA2TAG" VALUES (120014, 0);
--
--
INSERT INTO "ECALSWITCHES_DATA" VALUES ( 0, "Ecal", 1, 0, "GEO", "Development", "Baseline geometry", "EcalGeoModel/Ecal.gdml");
INSERT INTO "ECALSWITCHES_DATA2TAG" VALUES (100057, 0);
--
--
INSERT INTO "NEUTRINOIDENTIFIER_DATA" VALUES (0, "Neutrino", "NeutrinoIdDictFiles/IdDictNeutrino.xml", "Baseline layout");
INSERT INTO "NEUTRINOIDENTIFIER_DATA2TAG" VALUES (100030, 0);
--
--
INSERT INTO "SCINTIDENTIFIER_DATA" VALUES (0, "Scintillator", "ScintIdDictFiles/IdDictScintillator.xml", "Baseline layout");
INSERT INTO "SCINTIDENTIFIER_DATA2TAG" VALUES (100016, 0);
--
--
INSERT INTO "TRACKERIDENTIFIER_DATA" VALUES (0, "Tracker", "TrackerIdDictFiles/IdDictTracker.xml", "Baseline layout");
INSERT INTO "TRACKERIDENTIFIER_DATA" VALUES (1, "Tracker", "TrackerIdDictFiles/IdDictCosmic.xml", "Cosmic ray test stand");
INSERT INTO "TRACKERIDENTIFIER_DATA2TAG" VALUES (100017, 0);
INSERT INTO "TRACKERIDENTIFIER_DATA2TAG" VALUES (107785, 1);
--
--
INSERT INTO "CALOIDENTIFIER_DATA" VALUES (0, "Calorimeter", "CaloIdDictFiles/IdDictCalorimeter.xml", "Baseline layout");
INSERT INTO "CALOIDENTIFIER_DATA2TAG" VALUES (100018, 0);
--
--
DROP TABLE IF EXISTS "SCTBRLSENSOR_DATA";
CREATE TABLE "SCTBRLSENSOR_DATA" ( "SCTBRLSENSOR_DATA_ID" SLONGLONG ,"THICKNESS" DOUBLE ,"WIDTH" DOUBLE ,"WAFERLENGTH" DOUBLE ,"NUMWAFERS" INT ,"MATERIAL" TEXT ,
"FIDUCIALSEP" DOUBLE ,"CENTERTOFIDUCIAL" DOUBLE ,"STRIPPITCH" DOUBLE ,"NUMSTRIPS" INT ,"NUMREADOUTSTRIPS" INT ,"STRIPSHIFT" INT ,"ACTIVEHALFLENGTH" DOUBLE , UNIQUE ( "SCTBRLSENSOR_DATA_ID" ) );
INSERT INTO "SCTBRLSENSOR_DATA" VALUES(0,0.285,63.56,63.96,2,'std::Silicon',2.19,30.95,0.08,770,768,1,31.0);
DROP TABLE IF EXISTS "SCTBRLSENSOR_DATA2TAG";
CREATE TABLE "SCTBRLSENSOR_DATA2TAG" ( "SCTBRLSENSOR_TAG_ID" SLONGLONG ,"SCTBRLSENSOR_DATA_ID" SLONGLONG  );
INSERT INTO "SCTBRLSENSOR_DATA2TAG" VALUES(106730,0);
--
--
DROP TABLE IF EXISTS "SCTBRLMODULE_DATA";
CREATE TABLE "SCTBRLMODULE_DATA" ( "SCTBRLMODULE_DATA_ID" SLONGLONG ,"BASEBOARDTHICKNESS" DOUBLE ,"BASEBOARDWIDTH" DOUBLE ,"BASEBOARDLENGTH" DOUBLE ,
"BASEBOARDMATERIAL" TEXT ,"BASEBOARDOFFSETY" DOUBLE ,"BASEBOARDOFFSETZ" DOUBLE ,"HYBRIDTHICKNESS" DOUBLE ,"HYBRIDWIDTH" DOUBLE ,"HYBRIDLENGTH" DOUBLE ,
"HYBRIDMATERIAL" TEXT ,"HYBRIDOFFSETX" DOUBLE ,"HYBRIDOFFSETZ" DOUBLE ,"PIGTAILTHICKNESS" DOUBLE ,"PIGTAILWIDTH" DOUBLE ,"PIGTAILLENGTH" DOUBLE ,
"PIGTAILMATERIAL" TEXT ,"STEREOUPPERSIGN" INT ,"STEREOANGLE" DOUBLE ,"SENSORTOSENSORGAP" DOUBLE ,"SIDEUPPER" INT , UNIQUE ( "SCTBRLMODULE_DATA_ID" ) );
INSERT INTO "SCTBRLMODULE_DATA" VALUES(0,0.38,87.2,61.8,'sct::BrlBaseBoard',-5.7,-7.1,0.9,74.6,21.0,'sct::BrlHybrid',1.058,-15.5,0.9,25.4,35.4,'sct::PigTail',-1,40.0,0.6,0);
DROP TABLE IF EXISTS "SCTBRLMODULE_DATA2TAG";
CREATE TABLE "SCTBRLMODULE_DATA2TAG" ( "SCTBRLMODULE_TAG_ID" SLONGLONG ,"SCTBRLMODULE_DATA_ID" SLONGLONG  );
INSERT INTO "SCTBRLMODULE_DATA2TAG" VALUES(107003,0);
--
--
DROP TABLE IF EXISTS "SCTTOPLEVEL_DATA";
CREATE TABLE "SCTTOPLEVEL_DATA" ( "SCTTOPLEVEL_DATA_ID" SLONGLONG ,"POSX" DOUBLE ,"POSY" DOUBLE ,"POSZ" DOUBLE ,"ROTX" DOUBLE ,"ROTY" DOUBLE ,"ROTZ" DOUBLE ,"ROTORDER" INT ,"LABEL" TEXT , UNIQUE ( "SCTTOPLEVEL_DATA_ID" ) );
INSERT INTO "SCTTOPLEVEL_DATA" VALUES(0,0.0,0.0,-1200.0,0.0,0.0,0.0,312,'StationA');
INSERT INTO "SCTTOPLEVEL_DATA" VALUES(1,0.0,0.0,    0.0,0.0,0.0,0.0,312,'StationB');
INSERT INTO "SCTTOPLEVEL_DATA" VALUES(2,0.0,0.0, 1200.0,0.0,0.0,0.0,312,'StationC');
INSERT INTO "SCTTOPLEVEL_DATA" VALUES(3,0.0,0.0, 1290.0,0.0,0.0,0.0,312,'SCT');
INSERT INTO "SCTTOPLEVEL_DATA" VALUES(4,0.0,0.0,-1190.0,0.0,0.0,0.0,312,'StationA');
INSERT INTO "SCTTOPLEVEL_DATA" VALUES(5,0.0,0.0,    0.0,0.0,0.0,0.0,312,'StationB');
INSERT INTO "SCTTOPLEVEL_DATA" VALUES(6,0.0,0.0, 1190.0,0.0,0.0,0.0,312,'StationC');
INSERT INTO "SCTTOPLEVEL_DATA" VALUES(7,0.0,0.0, 1237.7075,0.0,0.0,0.0,312,'SCT');

DROP TABLE IF EXISTS "SCTTOPLEVEL_DATA2TAG";
CREATE TABLE "SCTTOPLEVEL_DATA2TAG" ( "SCTTOPLEVEL_TAG_ID" SLONGLONG ,"SCTTOPLEVEL_DATA_ID" SLONGLONG  );
INSERT INTO "SCTTOPLEVEL_DATA2TAG" VALUES(106788,0);
INSERT INTO "SCTTOPLEVEL_DATA2TAG" VALUES(106788,1);
INSERT INTO "SCTTOPLEVEL_DATA2TAG" VALUES(106788,2);
INSERT INTO "SCTTOPLEVEL_DATA2TAG" VALUES(106788,3);
INSERT INTO "SCTTOPLEVEL_DATA2TAG" VALUES(106790,4);
INSERT INTO "SCTTOPLEVEL_DATA2TAG" VALUES(106790,5);
INSERT INTO "SCTTOPLEVEL_DATA2TAG" VALUES(106790,6);
INSERT INTO "SCTTOPLEVEL_DATA2TAG" VALUES(106790,7);
INSERT INTO "SCTTOPLEVEL_DATA2TAG" VALUES(107786,5);
INSERT INTO "SCTTOPLEVEL_DATA2TAG" VALUES(107786,7);
--
--
DROP TABLE IF EXISTS "SCTFASERGENERAL_DATA";
CREATE TABLE "SCTFASERGENERAL_DATA" ("SCTFASERGENERAL_DATA_ID" SLONGLONG, "NUMLAYERS" INT, "LAYERPITCH" DOUBLE, "ETAHALFPITCH" DOUBLE, "PHIHALFPITCH" DOUBLE, "DEPTHHALFPITCH" DOUBLE, "SIDEHALFPITCH" DOUBLE, "PHISTAGGER" DOUBLE, UNIQUE ( "SCTFASERGENERAL_DATA_ID") );
INSERT INTO "SCTFASERGENERAL_DATA" VALUES (0, 3, 65.0, 60.0, 30.0, 2.5, 5.0, 0.0);
INSERT INTO "SCTFASERGENERAL_DATA" VALUES (1, 3, 31.5, 62.0, 31.87, 1.985, 3.9725, -5.0);
INSERT INTO "SCTFASERGENERAL_DATA" VALUES (2, 3, 31.5, 62.0, 31.87, 1.985, 3.9725, 0.0);
DROP TABLE IF EXISTS "SCTFASERGENERAL_DATA2TAG";
CREATE TABLE "SCTFASERGENERAL_DATA2TAG" ( "SCTFASERGENERAL_TAG_ID" SLONGLONG ,"SCTFASERGENERAL_DATA_ID" SLONGLONG  );
INSERT INTO "SCTFASERGENERAL_DATA2TAG" VALUES(106789, 0);
INSERT INTO "SCTFASERGENERAL_DATA2TAG" VALUES(106791, 1);
INSERT INTO "SCTFASERGENERAL_DATA2TAG" VALUES(107780, 2);
--
--
DROP TABLE IF EXISTS "SCTSWITCHES_DATA";
CREATE TABLE "SCTSWITCHES_DATA" ( "SCTSWITCHES_DATA_ID" SLONGLONG ,"DETECTORNAME" TEXT ,"G3COMPATIBLE" INT ,"G3DIGITSCOMPATIBLE" INT ,"DC2GEOMETRY" INT ,"USEMAGFIELDSVC" INT ,"COSMICLAYOUT" INT ,"VERSIONNAME" TEXT ,"LAYOUT" TEXT ,"DESCRIPTION" TEXT , UNIQUE ( "SCTSWITCHES_DATA_ID" ) );
INSERT INTO "SCTSWITCHES_DATA" VALUES(0,'SCT',0,0,0,1,0,'GEO','Final','Atlas Geometry 2008');
DROP TABLE IF EXISTS "SCTSWITCHES_DATA2TAG";
CREATE TABLE "SCTSWITCHES_DATA2TAG" ( "SCTSWITCHES_TAG_ID" SLONGLONG ,"SCTSWITCHES_DATA_ID" SLONGLONG  );
INSERT INTO "SCTSWITCHES_DATA2TAG" VALUES(107782,0);
--
--
DROP TABLE IF EXISTS "SCTCONDITIONS_DATA";
CREATE TABLE "SCTCONDITIONS_DATA" ( "SCTCONDITIONS_DATA_ID" SLONGLONG, "TEMPERATURE" DOUBLE, "BIASVOLT" DOUBLE, "DEPLETIONVOLT" DOUBLE, "USECONDDB" INT, UNIQUE ( "SCTCONDITIONS_DATA_ID" ) );
INSERT INTO "SCTCONDITIONS_DATA" VALUES(0, 4.6, 70.0, 150.0, 0);
DROP TABLE IF EXISTS "SCTCONDITIONS_DATA2TAG";
CREATE TABLE "SCTCONDITIONS_DATA2TAG" ( "SCTCONDITIONS_TAG_ID" SLONGLONG , "SCTCONDITIONS_DATA_ID" SLONGLONG );
INSERT INTO "SCTCONDITIONS_DATA2TAG" VALUES(107779, 0);
--
--
DROP TABLE IF EXISTS "SCTFRAMEGENERAL_DATA";
CREATE TABLE "SCTFRAMEGENERAL_DATA" ( "SCTFRAMEGENERAL_DATA_ID" SLONGLONG , "MATERIAL" TEXT , "WIDTH" DOUBLE , "HEIGHT" DOUBLE , "THICKNESS" DOUBLE , UNIQUE( "SCTFRAMEGENERAL_DATA_ID") );
INSERT INTO "SCTFRAMEGENERAL_DATA" VALUES(0, "sct::AW5083", 320.0, 320.0, 31.0 );
DROP TABLE IF EXISTS "SCTFRAMEGENERAL_DATA2TAG";
CREATE TABLE "SCTFRAMEGENERAL_DATA2TAG" ( "SCTFRAMEGENERAL_TAG_ID" SLONGLONG , "SCTFRAMEGENERAL_DATA_ID" SLONGLONG );
INSERT INTO "SCTFRAMEGENERAL_DATA2TAG" VALUES(100054, 0);
--
--
DROP TABLE IF EXISTS "SCTFRAMESHAPE_DATA";
CREATE TABLE "SCTFRAMESHAPE_DATA" ( "SCTFRAMESHAPE_DATA_ID" SLONGLONG , "DX" DOUBLE , "DY" DOUBLE , "DZ" DOUBLE , "X" DOUBLE , "Y" DOUBLE , "Z" DOUBLE , "STAGGERED" INT, UNIQUE( "SCTFRAMESHAPE_DATA_ID" ) );
INSERT INTO "SCTFRAMESHAPE_DATA" VALUES(1,  100.0,  285.0, 100.0,    0.0,   0.0,   0.0,  0);
INSERT INTO "SCTFRAMESHAPE_DATA" VALUES(2,   99.5,  285.0,  26.0,  -79.75,  0.0,  14.5,  0);
INSERT INTO "SCTFRAMESHAPE_DATA" VALUES(3,   99.5,  285.0,  22.0,   79.75,  0.00, 16.5,  0);
INSERT INTO "SCTFRAMESHAPE_DATA" VALUES(4,   99.5,  285.0,  26.0,   79.75,  0.0, -14.5,  0);
INSERT INTO "SCTFRAMESHAPE_DATA" VALUES(5,   99.5,  285.0,  22.0,  -79.75,  0.00,-15.5,  0);
INSERT INTO "SCTFRAMESHAPE_DATA" VALUES(6,  28.0,   20.0, 100.0,  -78.0,   0.0,   0.00, 1);
INSERT INTO "SCTFRAMESHAPE_DATA" VALUES(7,  28.0,   20.0, 100.0,   78.0,   0.0,   0.00, 1);
INSERT INTO "SCTFRAMESHAPE_DATA" VALUES(8,  28.0,   12.5, 100.0,  -78.0,  63.75,  0.00, 1);
INSERT INTO "SCTFRAMESHAPE_DATA" VALUES(9,  28.0,   12.5, 100.0,  -78.0, -63.75,  0.00, 1);
INSERT INTO "SCTFRAMESHAPE_DATA" VALUES(10,  28.0,   12.5, 100.0,   78.0, -63.75,  0.00, 1);
INSERT INTO "SCTFRAMESHAPE_DATA" VALUES(11,  28.0,   12.5, 100.0,   78.0,  63.75,  0.00, 1);
INSERT INTO "SCTFRAMESHAPE_DATA" VALUES(12,  129.5,   68.2,  51.5,  -64.75, 32.0,  24.25, 1);
INSERT INTO "SCTFRAMESHAPE_DATA" VALUES(13,  129.5,   68.2,  51.5,  -64.75,-95.5,  24.25, 1);
INSERT INTO "SCTFRAMESHAPE_DATA" VALUES(14,  129.5,   68.2,  47.0,   64.75,-32.0,  26.5,  1);
INSERT INTO "SCTFRAMESHAPE_DATA" VALUES(15,  129.5,   68.2,  47.0,   64.75, 95.5,  26.5,  1);
INSERT INTO "SCTFRAMESHAPE_DATA" VALUES(16, 129.5,   68.2,  51.5,   64.75, 32.0, -24.25, 1);
INSERT INTO "SCTFRAMESHAPE_DATA" VALUES(17, 129.5,   68.2,  51.5,   64.75,-95.5, -24.25, 1);
INSERT INTO "SCTFRAMESHAPE_DATA" VALUES(18, 129.5,   68.2,  47.0,  -64.75,-32.0, -26.5,  1);
INSERT INTO "SCTFRAMESHAPE_DATA" VALUES(19, 129.5,   68.2,  47.0,  -64.75, 95.5, -26.5,  1);
INSERT INTO "SCTFRAMESHAPE_DATA" VALUES(20,  28.0,   50.0,  53.0,  -78.0,  33.75, 23.5,  1);
INSERT INTO "SCTFRAMESHAPE_DATA" VALUES(21,  28.0,   68.5,  53.0,  -78.0,-103.25, 23.5,  1);
INSERT INTO "SCTFRAMESHAPE_DATA" VALUES(22,  28.0,   50.0,  48.5,   78.0, -33.75, 25.75, 1);
INSERT INTO "SCTFRAMESHAPE_DATA" VALUES(23,  28.0,   68.5,  48.5,   78.0, 103.25, 25.75, 1);
INSERT INTO "SCTFRAMESHAPE_DATA" VALUES(24,  28.0,   50.0,  53.0,   78.0,  33.75,-23.5,  1);
INSERT INTO "SCTFRAMESHAPE_DATA" VALUES(25,  28.0,   68.5,  53.0,   78.0,-103.25,-23.5,  1);
INSERT INTO "SCTFRAMESHAPE_DATA" VALUES(26,  28.0,   50.0,  48.5,  -78.0, -33.75,-25.75, 1);
INSERT INTO "SCTFRAMESHAPE_DATA" VALUES(27,  28.0,   68.5,  48.5,  -78.0, 103.25,-25.75, 1);



DROP TABLE IF EXISTS "SCTFRAMESHAPE_DATA2TAG";
CREATE TABLE "SCTFRAMESHAPE_DATA2TAG" ( "SCTFRAMESHAPE_TAG_ID" SLONGLONG , "SCTFRAMESHAPE_DATA_ID" SLONGLONG );
INSERT INTO "SCTFRAMESHAPE_DATA2TAG" VALUES(100055, 1);
INSERT INTO "SCTFRAMESHAPE_DATA2TAG" VALUES(100055, 2);
INSERT INTO "SCTFRAMESHAPE_DATA2TAG" VALUES(100055, 3);
INSERT INTO "SCTFRAMESHAPE_DATA2TAG" VALUES(100055, 4);
INSERT INTO "SCTFRAMESHAPE_DATA2TAG" VALUES(100055, 5);
INSERT INTO "SCTFRAMESHAPE_DATA2TAG" VALUES(100055, 6);
INSERT INTO "SCTFRAMESHAPE_DATA2TAG" VALUES(100055, 7);
INSERT INTO "SCTFRAMESHAPE_DATA2TAG" VALUES(100055, 8);
INSERT INTO "SCTFRAMESHAPE_DATA2TAG" VALUES(100055, 9);
INSERT INTO "SCTFRAMESHAPE_DATA2TAG" VALUES(100055, 10);
INSERT INTO "SCTFRAMESHAPE_DATA2TAG" VALUES(100055, 11);
INSERT INTO "SCTFRAMESHAPE_DATA2TAG" VALUES(100055, 12);
INSERT INTO "SCTFRAMESHAPE_DATA2TAG" VALUES(100055, 13);
INSERT INTO "SCTFRAMESHAPE_DATA2TAG" VALUES(100055, 14);
INSERT INTO "SCTFRAMESHAPE_DATA2TAG" VALUES(100055, 15);
INSERT INTO "SCTFRAMESHAPE_DATA2TAG" VALUES(100055, 16);
INSERT INTO "SCTFRAMESHAPE_DATA2TAG" VALUES(100055, 17);
INSERT INTO "SCTFRAMESHAPE_DATA2TAG" VALUES(100055, 18);
INSERT INTO "SCTFRAMESHAPE_DATA2TAG" VALUES(100055, 19);
INSERT INTO "SCTFRAMESHAPE_DATA2TAG" VALUES(100055, 20);
INSERT INTO "SCTFRAMESHAPE_DATA2TAG" VALUES(100055, 21);
INSERT INTO "SCTFRAMESHAPE_DATA2TAG" VALUES(100055, 22);
INSERT INTO "SCTFRAMESHAPE_DATA2TAG" VALUES(100055, 23);
INSERT INTO "SCTFRAMESHAPE_DATA2TAG" VALUES(100055, 24);
INSERT INTO "SCTFRAMESHAPE_DATA2TAG" VALUES(100055, 25);
INSERT INTO "SCTFRAMESHAPE_DATA2TAG" VALUES(100055, 26);
INSERT INTO "SCTFRAMESHAPE_DATA2TAG" VALUES(100055, 27);
--
--
DROP TABLE IF EXISTS "DIPOLESWITCHES_DATA";
CREATE TABLE "DIPOLESWITCHES_DATA" ( "DIPOLESWITCHES_DATA_ID" SLONGLONG ,"DETECTORNAME" TEXT ,"G3COMPATIBLE" INT ,"G3DIGITSCOMPATIBLE" INT ,"DC2GEOMETRY" INT ,"USEMAGFIELDSVC" INT ,"COSMICLAYOUT" INT ,"VERSIONNAME" TEXT ,"LAYOUT" TEXT ,"DESCRIPTION" TEXT , UNIQUE ( "DIPOLESWITCHES_DATA_ID" ) );
INSERT INTO "DIPOLESWITCHES_DATA" VALUES(0,'Dipole',0,0,0,1,0,'GEO','Prelim','Faser Geometry 2019');
DROP TABLE IF EXISTS "DIPOLESWITCHES_DATA2TAG";
CREATE TABLE "DIPOLESWITCHES_DATA2TAG" ( "DIPOLESWITCHES_TAG_ID" SLONGLONG ,"DIPOLESWITCHES_DATA_ID" SLONGLONG  );
INSERT INTO "DIPOLESWITCHES_DATA2TAG" VALUES(100028,0);
--
--
DROP TABLE IF EXISTS "DIPOLETOPLEVEL_DATA";
CREATE TABLE "DIPOLETOPLEVEL_DATA" ( "DIPOLETOPLEVEL_DATA_ID" SLONGLONG ,"POSX" DOUBLE ,"POSY" DOUBLE ,"POSZ" DOUBLE ,"ROTX" DOUBLE ,"ROTY" DOUBLE ,"ROTZ" DOUBLE ,"ROTORDER" INT ,"LABEL" TEXT , UNIQUE ( "DIPOLETOPLEVEL_DATA_ID" ) );
INSERT INTO "DIPOLETOPLEVEL_DATA" VALUES(0,0.0,0.0,-750.0,0.0,0.0,0.0,312,'UpstreamDipole');
INSERT INTO "DIPOLETOPLEVEL_DATA" VALUES(1,0.0,0.0, 700.0,0.0,0.0,0.0,312,'CentralDipole');
INSERT INTO "DIPOLETOPLEVEL_DATA" VALUES(2,0.0,0.0,1900.0,0.0,0.0,0.0,312,'DownstreamDipole');
INSERT INTO "DIPOLETOPLEVEL_DATA" VALUES(3,0.0,0.0, 0.0,0.0,0.0,0.0,312,'Dipole');
INSERT INTO "DIPOLETOPLEVEL_DATA" VALUES(4,0.0,0.0,-812.32,0.0,0.0,0.0,312,'UpstreamDipole');
INSERT INTO "DIPOLETOPLEVEL_DATA" VALUES(5,0.0,0.0, 637.726,0.0,0.0,0.0,312,'CentralDipole');
INSERT INTO "DIPOLETOPLEVEL_DATA" VALUES(6,0.0,0.0,1837.726,0.0,0.0,0.0,312,'DownstreamDipole');

DROP TABLE IF EXISTS "DIPOLETOPLEVEL_DATA2TAG";
CREATE TABLE "DIPOLETOPLEVEL_DATA2TAG" ( "DIPOLETOPLEVEL_TAG_ID" SLONGLONG ,"DIPOLETOPLEVEL_DATA_ID" SLONGLONG  );
INSERT INTO "DIPOLETOPLEVEL_DATA2TAG" VALUES(100029,0);
INSERT INTO "DIPOLETOPLEVEL_DATA2TAG" VALUES(100029,1);
INSERT INTO "DIPOLETOPLEVEL_DATA2TAG" VALUES(100029,2);
INSERT INTO "DIPOLETOPLEVEL_DATA2TAG" VALUES(100029,3);
INSERT INTO "DIPOLETOPLEVEL_DATA2TAG" VALUES(100040,3);
INSERT INTO "DIPOLETOPLEVEL_DATA2TAG" VALUES(100040,4);
INSERT INTO "DIPOLETOPLEVEL_DATA2TAG" VALUES(100040,5);
INSERT INTO "DIPOLETOPLEVEL_DATA2TAG" VALUES(100040,6);
--
--
DROP TABLE IF EXISTS "DIPOLEGENERAL_DATA";
CREATE TABLE "DIPOLEGENERAL_DATA" ( "DIPOLEGENERAL_DATA_ID" SLONGLONG, "MATERIAL" TEXT, "INNERRAD" DOUBLE, "OUTERRAD" DOUBLE, "SHORTLENGTH" DOUBLE, "LONGLENGTH" DOUBLE, UNIQUE ( "DIPOLEGENERAL_DATA_ID") );
INSERT INTO "DIPOLEGENERAL_DATA" VALUES(0, 'std::Nd2Fe14B', 100.0, 215.0, 1000.0, 1500.0);
DROP TABLE IF EXISTS "DIPOLEGENERAL_DATA2TAG";
CREATE TABLE "DIPOLEGENERAL_DATA2TAG" ( "DIPOLEGENERAL_TAG_ID" SLONGLONG, "DIPOLEGENERAL_DATA_ID" SLONGLONG );
INSERT INTO "DIPOLEGENERAL_DATA2TAG" VALUES(100004, 0);
--
--
DROP TABLE IF EXISTS "SCTMATCOMPONENTS_DATA";
CREATE TABLE "SCTMATCOMPONENTS_DATA" ( "SCTMATCOMPONENTS_DATA_ID" SLONGLONG ,"MATERIAL_ID" SLONGLONG ,"COMPNAME" TEXT ,"FRACTION" DOUBLE , UNIQUE ( "SCTMATCOMPONENTS_DATA_ID" ) );
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(201,50,'Carbon',0.4816);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(202,50,'sct::AlNitride',0.1576);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(203,50,'sct::Elastosil',0.0056);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(204,50,'sct::Epoxy',0.117);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(205,50,'Aluminium',0.0212);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(206,50,'sct::FR4',0.0644);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(207,50,'sct::Glass',0.1526);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(208,51,'Carbon',0.4081);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(209,51,'sct::AlNitride',0.1528);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(210,51,'sct::Elastosil',0.0111);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(211,51,'sct::Epoxy',0.1908);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(212,51,'Aluminium',0.0156);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(213,51,'sct::FR4',0.083);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(214,51,'sct::Glass',0.1386);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(544,124,'Iron',0.187);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(545,124,'Aluminium',0.491);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(546,124,'sct::BeCu',0.322);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(897,175,'sct::CN107',0.226);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(161,42,'sct::Peek',0.0251);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(168,43,'sct::Peek',0.0327);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(174,44,'sct::Peek',0.0192);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(178,45,'sct::Peek',0.0175);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(898,175,'sct::C6F14',0.076);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(0,0,'Carbon',1.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(1,1,'Carbon',1.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(2,2,'Carbon',1.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(3,3,'Carbon',1.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(4,4,'Copper',0.863);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(5,4,'std::Kapton',0.137);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(6,5,'Aluminium',1.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(7,6,'Aluminium',0.369);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(24,14,'sct::AirexR82',0.057);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(9,7,'Copper',0.7);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(10,7,'Nickel',0.3);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(11,8,'sct::CN107',0.09);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(12,8,'std::C3F8',0.91);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(13,9,'Carbon',1.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(14,10,'Carbon',1.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(15,11,'Carbon',1.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(16,12,'std::Polystyrene',1.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(17,13,'sct::CFRP',0.367);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(18,13,'std::Aluminium',0.495);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(19,13,'std::C3F8',0.116);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(20,13,'std::Kapton',0.022);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(21,14,'sct::CFRP',0.508);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(22,14,'std::Aluminium',0.193);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(23,14,'std::C3F8',0.242);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(27,15,'sct::AirexR82',0.005);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(25,15,'sct::CFRP',0.9);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(26,15,'std::Aluminium',0.095);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(28,16,'std::Aluminium',0.588);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(29,16,'std::Kapton',0.412);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(30,17,'Silicon',1.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(31,18,'Carbon',1.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(32,19,'Carbon',1.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(33,20,'Hydrogen',0.092);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(34,20,'Oxygen',0.366);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(35,20,'Aluminium',0.542);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(36,21,'Aluminium',1.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(37,22,'std::Polystyrene',1.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(38,23,'std::Polyethylene',1.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(8,6,'std::Kapton',0.631);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(338,77,'Carbon',37.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(339,77,'Hydrogen',24.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(340,77,'Oxygen',4.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(341,77,'Nitrogen',2.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(342,78,'Hydrogen',0.026);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(343,78,'Carbon',0.692);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(344,78,'Nitrogen',0.073);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(345,78,'Oxygen',0.209);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(346,79,'Aluminium',0.0496);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(347,79,'Iron',0.0468);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(348,79,'Copper',0.185);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(349,79,'Nickel',0.0027);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(350,79,'Silicon',0.0012);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(351,79,'sct::Solder',0.0091);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(352,79,'std::Kapton',0.119);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(353,79,'sct::Peek',0.0409);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(354,79,'sct::PEI',0.0126);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(355,79,'sct::Glass',0.0136);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(356,79,'sct::Plastic',0.0085);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(357,79,'sct::CFRP',0.1192);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(358,79,'sct::FR4',0.1344);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(359,79,'sct::Adhesive',0.119);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(360,79,'sct::AlNitride',0.1384);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(490,107,'sct::Epoxy',0.6925);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(491,107,'sct::BoronNitride',0.3075);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(533,122,'sct::Peek',0.4);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(534,122,'sct::CFRP',0.45);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(535,122,'sct::PEI',0.11);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(536,122,'sct::Epoxy',0.04);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(547,125,'Iron',0.0417);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(548,125,'Copper',0.333);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(549,125,'sct::CN107',0.6249);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(550,125,'sct::solder',0.0004);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(561,128,'Silver',0.15);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(562,128,'Nickel',0.6);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(563,128,'sct::Epoxy',0.25);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(684,148,'Copper',0.9);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(685,148,'Zinc',0.1);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(686,149,'Aluminium',0.2722);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(687,149,'Iron',0.0168);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(688,149,'sct::CFRP',0.6314);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(689,149,'sct::Peek',0.0701);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(690,149,'sct::EPDM',0.0018);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(691,149,'sct::Brass',0.0077);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(737,157,'Carbon',1.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(738,158,'Carbon',1.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(826,142,'sct::CFRP',0.613);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(827,142,'sct::CuKaptonEC',0.053);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(828,142,'std::Kapton',0.04);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(829,142,'Copper',0.065);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(830,142,'sct::Araldite',0.072);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(831,142,'sct::CuOnAlWire',0.007);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(832,142,'std::G10',0.15);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(906,174,'Copper',0.696);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(907,174,'std::C3F8',0.052);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(908,174,'Iron',0.1);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(909,174,'sct::CN107',0.024);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(910,174,'sct::Rubber',0.128);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(937,184,'Aluminium',0.004);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(938,184,'sct::Peek',0.6584);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(939,184,'sct::Vectra',0.2431);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(940,184,'sct::Glass',0.0026);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(941,184,'sct::Plastic',0.0374);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(942,184,'sct::Silica',0.0545);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(943,185,'sct::Peek',0.6677);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(944,185,'sct::Glass',0.1638);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(945,185,'sct::Plastic',0.102);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(946,185,'sct::Araldite',0.0665);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(947,186,'Aluminium',0.0293);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(948,186,'sct::Peek',0.7988);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(949,186,'sct::Vectra',0.1237);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(950,186,'sct::Glass',0.0153);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(951,186,'sct::Silica',0.0053);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(952,186,'sct::Araldite',0.0276);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(953,187,'Aluminium',0.0021);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(954,187,'sct::Peek',0.2746);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(955,187,'sct::Vectra',0.418);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(956,187,'sct::Glass',0.026);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(957,187,'sct::Plastic',0.2038);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(958,187,'sct::Silica',0.0247);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(959,187,'std::Kapton',0.001);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(960,187,'sct::Araldite',0.0498);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(510,116,'sct::CN107',0.785);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(511,116,'std::C3F8',0.127);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(512,116,'Copper',0.088);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(513,117,'std::Kapton',0.27);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(514,117,'Copper',0.61);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(515,117,'sct::Epoxy',0.12);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(516,118,'sct::CN107',0.52);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(517,118,'std::C3F8',0.21);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(518,118,'Aluminium',0.1);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(519,118,'std::Kapton',0.03);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(520,118,'sct::BeCu',0.14);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(521,119,'sct::Plastic',0.33);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(522,119,'sct::Epoxy',0.06);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(523,119,'sct::BeCu',0.14);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(524,119,'std::Polyethylene',0.47);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(525,120,'Aluminium',0.219);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(526,120,'sct::Peek',0.522);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(527,120,'Copper',0.125);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(528,120,'Iron',0.134);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(529,121,'sct::Peek',0.4);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(530,121,'sct::CFRP',0.45);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(531,121,'sct::PEI',0.11);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(532,121,'sct::Epoxy',0.04);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(537,123,'Aluminium',0.044);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(538,123,'sct::CFRP',0.821);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(539,123,'sct::CuKapton',0.079);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(540,123,'sct::Solder',0.001);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(541,123,'Copper',0.006);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(542,123,'sct::Plastic',0.036);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(543,123,'sct::Peek',0.013);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(264,34,'sct::Plastic',0.23);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(265,34,'Copper',0.32);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(266,34,'sct::FR4',0.27);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(267,34,'sct::CuOnAlWire',0.04);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(268,34,'std::Kapton',0.14);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(604,132,'sct::Silica',0.99);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(619,138,'Copper',0.76);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(620,138,'std::Kapton',0.24);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(621,139,'sct::PEI',0.373);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(622,139,'sct::AlKapton',0.028);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(623,139,'sct::CuKaptonEC',0.166);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(624,139,'sct::Araldite',0.31);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(625,139,'std::Kapton',0.084);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(626,139,'Copper',0.024);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(627,139,'sct::CuOnAlWire',0.015);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(628,140,'sct::CFRP',0.512);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(629,140,'sct::PEI',0.163);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(630,140,'sct::CuKapton',0.117);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(631,140,'sct::Epoxy',0.208);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(632,141,'sct::Epoxy',0.106);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(633,141,'sct::PEI',0.342);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(634,141,'sct::CuKapton',0.552);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(635,142,'sct::CFRP',0.835);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(636,142,'sct::CuKapton',0.072);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(637,142,'std::Kapton',0.054);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(638,142,'Copper',0.009);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(639,142,'sct::Araldite',0.02);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(640,142,'sct::CuOnAlWire',0.01);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(641,143,'Iron',0.284);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(642,143,'sct::PEI',0.062);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(643,143,'sct::CFRP',0.522);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(644,143,'sct::Araldite',0.007);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(645,143,'std::G10',0.125);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(646,144,'sct::PEI',0.106);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(647,144,'sct::CuKapton',0.093);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(648,144,'Aluminium',0.008);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(649,144,'sct::CFRP',0.763);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(650,144,'Titanium',0.03);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(651,145,'sct::CFRP',0.796);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(652,145,'sct::PEI',0.069);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(653,145,'sct::AlKapton',0.007);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(654,145,'sct::Epoxy',0.061);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(655,145,'sct::CuKapton',0.037);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(656,145,'std::Kapton',0.012);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(657,145,'Copper',0.003);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(658,145,'sct::Araldite',0.007);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(659,145,'sct::CuOnAlWire',0.005);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(660,145,'Aluminium',0.002);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(661,145,'sct::Peek',0.001);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(662,146,'Iron',0.645);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(663,146,'sct::Peek',0.241);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(664,146,'Aluminium',0.114);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(665,147,'sct::CFRP',0.377);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(666,147,'Aluminium',0.04);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(667,147,'Iron',0.05);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(668,147,'sct::PEI',0.454);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(669,147,'sct::Silicone',0.079);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(750,145,'sct::CuKaptonEC',0.045);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(833,171,'Carbon',6.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(834,171,'Fluorine',14.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(835,172,'Carbon',15.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(836,172,'Hydrogen',6.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(837,172,'Oxygen',3.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(838,173,'sct::Polycarbonate',1.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(839,174,'Copper',0.684);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(840,174,'std::C3F8',0.069);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(841,174,'Iron',0.098);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(842,174,'sct::CN107',0.023);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(843,174,'sct::Rubber',0.126);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(844,175,'std::Kapton',0.18);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(845,175,'Copper',0.406);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(846,175,'sct::Epoxy',0.08);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(847,175,'sct:CN107',0.226);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(848,175,'sct:C6F14',0.076);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(849,175,'Aluminium',0.032);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(850,176,'Aluminium',0.925);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(851,176,'Iron',0.043);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(852,176,'Copper',0.002);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(853,176,'std::Kapton',0.03);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(854,155,'Aluminium',0.235);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(855,155,'Copper',0.118);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(856,155,'Iron',0.427);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(857,155,'std::Polyethylene',0.201);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(858,155,'sct::Polyester',0.019);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(859,177,'Copper',0.549);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(860,177,'sct::CN107',0.152);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(861,177,'Iron',0.265);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(862,177,'sct::Peek',0.008);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(863,177,'sct::Solder',0.026);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(864,178,'sct::CN107',0.702);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(865,178,'Copper',0.073);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(866,178,'std::C3F8',0.126);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(867,178,'sct::Rubber',0.099);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(868,179,'std::Kapton',0.258);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(869,179,'Copper',0.578);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(870,179,'sct::Epoxy',0.114);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(871,179,'Aluminium',0.05);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(872,153,'Aluminium',0.326);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(873,153,'Copper',0.196);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(874,153,'std::Polyethylene',0.478);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(875,180,'sct::Polyimide',0.021);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(876,180,'Iron',0.634);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(877,180,'sct::Polyester',0.239);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(878,180,'Aluminium',0.002);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(879,180,'Copper',0.1);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(880,180,'sct::Peek',0.004);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(881,181,'sct::PEI',0.533);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(882,181,'sct::CFRP',0.181);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(883,181,'sct::Araldite',0.041);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(884,181,'sct::Polyimide',0.002);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(885,181,'Aluminium',0.076);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(886,181,'Iron',0.03);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(887,181,'sct::Torlon',0.015);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(888,181,'sct::Silicone',0.066);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(889,181,'std::Kapton',0.021);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(890,181,'Copper',0.004);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(891,181,'sct::Polyimide',0.025);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(892,181,'sct::CuKaptonEC',0.006);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(893,182,'Aluminium',0.904);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(894,182,'Iron',0.053);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(895,182,'std::G10',0.031);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(896,182,'sct::Araldite',0.012);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(249,64,'Carbon',10.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(250,64,'Hydrogen',6.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(251,64,'Oxygen',1.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(252,65,'Copper',1.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(253,66,'sct::Plastic',1.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(254,67,'sct::Plastic',1.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(255,68,'sct::Plastic',1.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(256,69,'sct::Silica',1.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(257,70,'sct::Plastic',1.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(258,71,'Copper',1.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(222,52,'sct::Solder',0.0554);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(165,42,'sct::Araldite',0.0099);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(171,43,'sct::Araldite',0.0098);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(551,125,'sct::Solder',0.0004);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(925,149,'Aluminium',0.2152);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(926,149,'Iron',0.0013);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(116,35,'std::C3F8',0.1636);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(928,149,'sct::Peek',0.0403);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(929,149,'sct::EPDM',0.0017);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(930,183,'Aluminium',0.4829);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(931,183,'Iron',0.0733);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(932,183,'sct::CFRP',0.2265);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(933,183,'sct::Peek',0.1792);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(934,183,'sct::EPDM',0.0021);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(935,183,'sct::Epoxy',0.0002);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(936,183,'sct::Brass',0.0358);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(927,149,'sct::CFRP',0.7415);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(327,74,'Silicon',0.637);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(328,74,'Oxygen',0.363);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(329,75,'Aluminium',0.6103);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(330,75,'Copper',0.0126);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(331,75,'sct::Solder',0.1652);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(332,75,'sct::Peek',0.184);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(333,75,'sct::Silicone',0.0279);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(451,105,'sct::CN107',0.4811);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(452,105,'sct::Solder',0.0027);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(453,105,'std::C3F8',0.5162);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(558,127,'Aluminium',0.27);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(559,127,'sct::CFRP',0.72);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(560,127,'sct::Araldite',0.01);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(605,137,'Aluminium',0.4822);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(606,137,'Iron',0.0106);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(607,137,'Copper',0.0152);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(608,137,'sct::CN107',0.0012);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(609,137,'std::Kapton',0.1966);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(610,137,'sct::Peek',0.0023);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(611,137,'std::G10',0.2794);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(612,137,'sct::BeCu',0.0034);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(613,137,'sct::Glass',0.004);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(614,137,'sct::Plastic',0.0051);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(615,120,'Aluminium',0.242);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(616,120,'sct::Peek',0.471);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(617,120,'Copper',0.138);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(618,120,'Iron',0.149);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(260,53,'sct::Plastic',0.64);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(93,30,'Oxygen',0.4418);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(259,53,'sct::Glass',0.08);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(261,53,'sct::Rubber',0.23);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(262,53,'sct::Ceramic',0.03);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(263,53,'Aluminium',0.02);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(899,161,'Aluminium',0.3033);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(900,161,'Fluorine',0.3965);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(901,161,'Copper',0.1013);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(902,161,'sct::CN107',0.0797);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(903,161,'sct::Airex',0.02);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(904,161,'sct::Peek',0.0012);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(905,161,'std::C3F8',0.098);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(39,24,'Hydrogen',0.0063);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(40,24,'Boron',0.005);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(41,24,'Carbon',0.4703);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(42,24,'Nitrogen',0.0143);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(43,24,'Oxygen',0.0922);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(44,24,'Aluminium',0.0581);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(45,24,'Silicon',0.1351);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(46,24,'Chlorine',0.0033);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(47,24,'Nickel',0.0038);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(48,24,'Copper',0.1621);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(49,24,'Tin',0.0094);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(50,24,'Silver',0.0329);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(51,24,'Gold',0.0017);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(52,24,'Lead',0.0055);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(53,25,'Hydrogen',0.0032);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(54,25,'Beryllium',0.0892);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(55,25,'Boron',0.0014);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(56,25,'Carbon',0.7198);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(57,25,'Nitrogen',0.0019);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(58,25,'Oxygen',0.166);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(59,25,'Aluminium',0.0175);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(60,25,'Silver',0.001);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(61,26,'Hydrogen',0.0289);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(62,26,'Carbon',0.3713);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(63,26,'Nitrogen',0.0112);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(64,26,'Oxygen',0.091);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(65,26,'Nickel',0.0009);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(66,26,'Copper',0.4063);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(67,26,'Tin',0.0606);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(68,26,'Gold',0.0004);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(69,26,'Lead',0.0294);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(70,27,'Hydrogen',0.0348);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(71,27,'Carbon',0.5257);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(72,27,'Oxygen',0.1733);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(73,27,'Aluminium',0.0929);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(74,27,'Silicon',0.0742);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(75,27,'Iron',0.0346);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(76,27,'Copper',0.0531);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(77,27,'Silver',0.0001);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(78,27,'Tin',0.0034);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(79,27,'Lead',0.0079);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(80,28,'Hydrogen',0.0187);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(81,28,'Carbon',0.3659);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(82,28,'Nitrogen',0.0538);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(83,28,'Oxygen',0.1104);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(84,28,'Aluminium',0.1224);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(85,28,'Silicon',0.032);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(86,28,'Iron',0.0385);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(87,28,'Copper',0.2487);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(88,28,'Nickel',0.0022);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(89,28,'Tin',0.0022);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(90,28,'Lead',0.0052);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(91,29,'Aluminium',0.3595);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(92,29,'std::Kapton',0.6405);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(502,114,'sct::Plastic',0.15);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(94,30,'Silicon',0.2743);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(95,30,'Boron',0.0166);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(96,30,'Aluminium',0.0207);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(97,30,'Sodium',0.0448);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(98,30,'Potassium',0.0822);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(99,30,'Zinc',0.0882);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(100,30,'Titanium',0.0292);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(101,30,'Antimony',0.0022);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(102,31,'Copper',0.6142);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(103,31,'std::Kapton',0.3858);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(104,32,'Aluminium',1.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(105,32,'Nitrogen',1.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(106,33,'Copper',0.2571);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(107,33,'Aluminium',0.5199);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(108,33,'std::Kapton',0.223);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(109,34,'Sct:Plastic',0.346);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(110,34,'Copper',0.4613);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(111,34,'Sct::FR4',0.118);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(112,34,'Sct::CuOnAlWire',0.0157);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(113,34,'std::Kapton',0.059);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(114,35,'sct::CN107',0.747);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(115,35,'sct::Solder',0.0894);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(503,114,'Copper',0.09);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(504,114,'sct::FR4',0.5);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(505,114,'sct::CuOnAlWire',0.09);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(506,114,'std::Kapton',0.17);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(508,115,'Titanium',0.22);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(509,115,'Aluminium',0.3);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(507,115,'sct::CN107',0.48);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(552,16,'Aluminium',0.27);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(553,16,'sct::CFRP',0.72);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(554,16,'sct::Araldite',0.01);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(555,126,'Silver',0.15);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(556,126,'Nickel',0.6);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(557,126,'sct::Epoxy',0.25);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(159,42,'sct::Torlon',0.7947);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(160,42,'sct::Silica',0.0585);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(162,42,'sct::Vectra',0.0472);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(163,42,'sct::Plastic',0.0622);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(164,42,'Aluminium',0.0024);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(166,43,'sct::Torlon',0.8186);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(167,43,'sct::Silica',0.0131);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(169,43,'sct::Vectra',0.0114);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(170,43,'sct::Plastic',0.1144);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(172,44,'sct::Torlon',0.8208);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(173,44,'sct::Silica',0.1454);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(175,44,'Aluminium',0.0146);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(176,45,'sct::Torlon',0.76);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(177,45,'sct::Silica',0.196);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(179,45,'Aluminium',0.0265);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(194,49,'Carbon',0.4895);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(195,49,'sct::AlNitride',0.1511);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(196,49,'sct::Elastosil',0.0052);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(197,49,'sct::Epoxy',0.1179);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(198,49,'Aluminium',0.021);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(199,49,'sct::FR4',0.0639);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(200,49,'sct::Glass',0.1514);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(215,52,'Copper',0.2873);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(216,52,'Carbon',0.2376);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(217,52,'Silicon',0.1024);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(218,52,'std::Kapton',0.105);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(219,52,'sct::Epoxy',0.1013);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(220,52,'sct::Al2O3',0.0531);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(221,52,'sct::Plastic',0.0226);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(223,52,'sct::AlNitride',0.0179);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(224,52,'sct::EotiteP102',0.0128);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(225,52,'sct::AIT',0.0046);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(226,53,'sct::Glass',0.0825);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(227,53,'sct::Plastic',0.5408);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(228,53,'sct::Rubber',0.2291);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(229,53,'sct::Ceramic',0.0261);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(230,53,'Aluminium',0.0215);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(231,54,'Silicon',1.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(232,54,'Oxygen',2.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(233,55,'Tin',0.3642);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(234,55,'Lead',0.6358);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(235,56,'sct::Epoxy',0.2064);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(236,56,'sct::Glass',0.7936);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(237,57,'Aluminium',2.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(238,57,'Oxygen',3.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(239,58,'sct::Plastic',1.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(240,59,'sct::Plastic',1.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(241,60,'Carbon',6.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(242,60,'Hydrogen',6.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(243,60,'Oxygen',1.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(244,61,'sct::AIT',1.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(245,62,'sct::Epoxy',0.6514);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(246,62,'Silver',0.3486);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(247,63,'Silicon',1.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(248,63,'Hydrogen',2.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(564,129,'sct::PEI',1.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(565,130,'sct::Silicone',1.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(566,131,'Aluminium',0.15);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(567,131,'std::Kapton',0.85);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(568,132,'Silica',0.99);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(569,132,'std::Air',0.01);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(570,133,'sct::Rubber',1.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(571,134,'Aluminium',0.4334);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(572,134,'std::Kapton',0.0064);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(573,134,'sct::AlKapton',0.0293);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(574,134,'sct::Pyrogel',0.1619);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(575,134,'sct::CFRP',0.261);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(576,134,'sct::Airex',0.015);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(577,134,'sct::Araldite',0.0864);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(578,134,'sct::Epoxy',0.0014);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(579,134,'sct::Techsil',0.001);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(580,134,'Nickel',0.0034);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(581,134,'Silver',0.0008);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(582,135,'Aluminium',0.7898);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(583,135,'Iron',0.002);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(584,135,'sct::CN107',0.0622);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(585,135,'sct::Peek',0.0682);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(586,135,'sct::EPDM',0.0258);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(587,135,'sct::Araldite',0.0109);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(588,135,'sct::Silicone',0.0003);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(589,135,'sct::Techsil',0.0408);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(590,136,'Aluminium',0.4126);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(591,136,'Iron',0.0096);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(592,136,'Copper',0.0085);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(593,136,'std::Kapton',0.007);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(594,136,'sct::AlKapton',0.0126);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(595,136,'sct::CuKapton',0.0114);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(596,136,'sct::Pyrogel',0.0788);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(597,136,'sct::CFRP',0.241);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(598,136,'sct::Airex',0.0066);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(599,136,'sct::Peek',0.0099);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(600,136,'sct::EPDM',0.0142);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(601,136,'sct::Araldite',0.1864);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(602,136,'sct::Epoxy',0.0013);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(603,136,'sct::Solder',0.0001);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(739,145,'sct::CFRP',0.75);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(740,145,'sct::PEI',0.085);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(741,145,'sct::AlKapton',0.008);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(742,145,'sct::Epoxy',0.075);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(743,145,'sct::CuKapton',0.045);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(744,145,'std::Kapton',0.015);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(745,145,'Copper',0.004);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(746,145,'sct::Araldite',0.009);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(747,145,'sct::CuOnAlWire',0.006);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(748,145,'Aluminium',0.002);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(749,145,'sct::Peek',0.001);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(816,137,'Aluminium',0.4985);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(817,137,'Iron',0.0103);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(818,137,'Copper',0.0147);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(819,137,'sct::CN107',0.0011);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(820,137,'std::Kapton',0.1904);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(821,137,'sct::Peek',0.0022);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(822,137,'std::G10',0.2705);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(823,137,'sct::BeCu',0.0034);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(824,137,'sct::Glass',0.0039);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(825,137,'sct::Plastic',0.005);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(361,80,'Copper',0.95);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(362,80,'Gold',0.05);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(363,81,'Aluminium',0.0087);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(364,81,'sct::CFRP',0.9913);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(365,82,'Aluminium',0.0086);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(366,82,'sct::CFRP',0.9914);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(367,83,'Aluminium',0.007);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(368,83,'sct::CFRP',0.993);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(369,84,'Aluminium',0.0077);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(370,84,'sct::CFRP',0.9923);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(371,85,'sct::Peek',0.0511);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(372,85,'sct::CFRP',0.9472);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(373,85,'sct::CuAu',0.0017);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(374,86,'sct::Peek',0.0373);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(375,86,'sct::CFRP',0.9616);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(376,86,'sct::CuAu',0.0011);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(377,87,'sct::Peek',0.0311);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(378,87,'sct::CFRP',0.968);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(379,87,'sct::CuAu',0.0009);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(380,88,'sct::Peek',0.0273);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(381,88,'sct::CFRP',0.9719);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(382,88,'sct::CuAu',0.0008);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(383,89,'Aluminium',0.0311);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(384,89,'Copper',0.0347);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(385,89,'sct::Peek',0.6797);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(386,89,'sct::Glass',0.1296);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(387,89,'sct::Plastic',0.0807);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(388,89,'sct::BeCu',0.0442);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(389,90,'Aluminium',0.0297);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(390,90,'Copper',0.0322);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(391,90,'sct::Peek',0.6468);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(392,90,'sct::Glass',0.1537);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(393,90,'sct::Plastic',0.0957);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(394,90,'sct::BeCu',0.0419);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(395,91,'Aluminium',0.0313);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(396,91,'Copper',0.0295);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(397,91,'sct::Peek',0.6211);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(398,91,'sct::Glass',0.172);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(399,91,'sct::Plastic',0.1071);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(400,91,'sct::BeCu',0.039);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(401,92,'Aluminium',0.028);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(402,92,'Copper',0.0285);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(403,92,'sct::Peek',0.5872);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(404,92,'sct::Glass',0.1961);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(405,92,'sct::Plastic',0.1221);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(406,92,'sct::BeCu',0.0381);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(670,123,'sct::CuKaptonEC',0.079);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(671,140,'sct::CuKaptonEC',0.117);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(672,141,'sct::CuKaptonEC',0.552);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(673,142,'sct::CuKaptonEC',0.072);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(674,144,'sct::CuKaptonEC',0.093);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(675,145,'sct::CuKaptonEC',0.037);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(676,71,'Copper',0.27);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(677,71,'Aluminium',0.73);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(678,74,'Carbon',2.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(679,74,'Hydrogen',6.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(680,74,'Oxygen',1.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(681,74,'Silicon',1.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(682,133,'Carbon',3.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(683,133,'Hydrogen',10.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(729,156,'sct::PEI',0.222);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(730,156,'std::Kapton',0.106);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(731,156,'Aluminium',0.003);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(732,156,'Copper',0.114);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(733,156,'sct::CFRP',0.191);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(734,156,'sct::Plastic',0.016);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(735,156,'sct::Araldite',0.218);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(736,156,'sct::CuOnAlWire',0.13);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(751,159,'Copper',0.054);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(752,159,'sct::CN107',0.7994);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(753,159,'std::C3F8',0.1466);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(754,160,'Aluminium',0.2755);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(755,160,'Fluorine',0.0016);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(756,160,'Copper',0.3159);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(757,160,'sct::CN107',0.2966);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(758,160,'sct::Airex',0.0182);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(759,160,'sct::Peek',0.0032);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(760,160,'std::C3F8',0.089);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(761,161,'Aluminium',0.2737);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(762,161,'Fluorine',0.4783);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(763,161,'Copper',0.0831);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(764,161,'sct::CN107',0.0654);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(765,161,'sct::Airex',0.0181);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(766,161,'sct::Peek',0.001);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(767,161,'std::C3F8',0.0804);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(768,162,'Carbon',22.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(769,162,'Hydrogen',10.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(770,162,'Oxygen',5.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(771,162,'Nitrogen',2.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(772,163,'Aluminium',0.776);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(773,163,'Iron',0.0154);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(774,163,'Copper',0.0826);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(775,163,'sct::Peek',0.0069);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(776,163,'sct::FR4',0.0962);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(777,163,'sct::Polyimide',0.0213);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(778,163,'sct::Al2O3',0.0016);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(779,164,'Aluminium',0.3856);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(780,164,'Iron',0.0548);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(781,164,'Copper',0.2619);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(782,164,'sct::Peek',0.0019);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(783,164,'sct::FR4',0.2128);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(784,164,'sct::Polyimide',0.083);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(785,165,'Aluminium',0.4381);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(786,165,'Iron',0.0425);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(787,165,'Copper',0.2352);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(788,165,'sct::FR4',0.1912);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(789,165,'sct::Polyimide',0.0746);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(790,165,'sct::Plastic',0.0056);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(791,165,'sct::BeCu',0.0103);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(792,165,'sct::Ceramic',0.0025);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(793,166,'Aluminium',0.6489);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(794,166,'Iron',0.0145);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(795,166,'Copper',0.2967);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(796,166,'sct::FR4',0.0399);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(797,167,'Aluminium',0.5534);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(798,167,'Copper',0.3901);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(799,167,'sct::Araldite',0.0109);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(800,167,'sct::Polyimide',0.0456);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(801,168,'Aluminium',0.3385);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(802,168,'Copper',0.4877);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(803,168,'sct::Araldite',0.0061);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(804,168,'sct::Polyimide',0.1677);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(805,169,'Aluminium',0.4169);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(806,169,'Copper',0.4079);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(807,169,'sct::Araldite',0.0049);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(808,169,'sct::Polyimide',0.1363);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(809,169,'sct::Glass',0.0143);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(810,169,'sct::Plastic',0.0197);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(811,170,'Aluminium',0.6634);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(812,170,'Copper',0.2682);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(813,170,'sct::CN107',0.0421);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(814,170,'sct::Araldite',0.0112);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(815,170,'sct::Polyimide',0.0151);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(911,136,'Aluminium',0.4165);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(912,136,'Iron',0.0095);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(913,136,'Copper',0.0084);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(914,136,'std::Kapton',0.007);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(915,136,'sct::AlKapton',0.0125);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(916,136,'sct::CuKapton',0.0113);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(917,136,'sct::Pyrogel',0.0783);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(918,136,'sct::CFRP',0.2394);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(919,136,'sct::Airex',0.0066);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(920,136,'sct::Peek',0.0098);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(921,136,'sct::EPDM',0.0141);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(922,136,'sct::Araldite',0.1852);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(923,136,'sct::Epoxy',0.0013);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(924,136,'sct::Solder',0.0001);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(117,36,'sct::CN107',0.7245);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(118,36,'sct::Solder',0.1189);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(119,36,'std::C3F8',0.1566);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(120,37,'sct::CN107',0.7761);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(121,37,'sct::Solder',0.048);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(122,37,'std::C3F8',0.1759);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(123,38,'Carbon',0.7909);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(124,38,'sct::Peek',0.0351);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(125,38,'Copper',0.0175);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(126,38,'Aluminium',0.1059);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(127,38,'Gold',0.0125);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(128,38,'Lead',0.0108);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(129,38,'Tin',0.0162);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(130,38,'sct::BeCu',0.0093);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(131,38,'sct::Grease',0.0018);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(132,39,'Carbon',0.7267);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(133,39,'sct::Peek',0.0288);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(134,39,'Copper',0.0229);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(135,39,'Aluminium',0.1425);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(136,39,'Gold',0.0165);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(137,39,'Lead',0.0178);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(138,39,'Tin',0.0266);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(139,39,'sct::BeCu',0.0153);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(140,39,'sct::Grease',0.0029);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(141,40,'Carbon',0.7194);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(142,40,'sct::Peek',0.069);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(143,40,'Copper',0.0232);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(144,40,'Aluminium',0.1107);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(145,40,'Gold',0.0166);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(146,40,'Lead',0.0204);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(147,40,'Tin',0.0306);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(148,40,'sct::BeCu',0.0078);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(149,40,'sct::Grease',0.0023);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(150,41,'Carbon',0.6606);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(151,41,'sct::Peek',0.0494);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(152,41,'Copper',0.0283);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(153,41,'Aluminium',0.1446);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(154,41,'Gold',0.0203);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(155,41,'Lead',0.0323);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(156,41,'Tin',0.0485);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(157,41,'sct::BeCu',0.0124);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(158,41,'sct::Grease',0.0036);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(180,46,'Aluminium',0.5957);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(181,46,'sct::Peek',0.0618);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(182,46,'sct::Plastic',0.05);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(183,46,'Copper',0.0297);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(184,46,'sct::FR4',0.1728);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(185,46,'sct::CuOnAlWire',0.0313);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(186,46,'std::Kapton',0.0587);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(187,47,'Aluminium',0.1778);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(188,47,'sct::Plastic',0.6784);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(189,47,'sct::BeCu',0.1438);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(190,48,'Aluminium',0.4371);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(191,48,'sct::CN107',0.3171);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(192,48,'sct::Solder',0.0415);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(193,48,'sct::PCB',0.2043);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(492,112,'Aluminium',0.0127);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(493,112,'Copper',0.2054);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(494,112,'sct::Solder',0.0113);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(495,112,'std::Kapton',0.0186);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(496,112,'sct::Peek',0.4896);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(497,112,'sct::FR4',0.2624);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(498,113,'Aluminium',0.3326);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(499,113,'std::Kapton',0.4874);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(500,113,'sct::Glass',0.1109);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(501,113,'sct::Plastic',0.0691);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(692,150,'sct::CFRP',0.953);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(693,150,'std::Kapton',0.025);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(694,150,'Copper',0.007);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(695,150,'sct::Araldite',0.015);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(696,151,'sct::CFRP',0.971);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(697,151,'std::Kapton',0.015);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(698,151,'Copper',0.005);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(699,151,'sct::Araldite',0.009);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(700,152,'Aluminium',0.472);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(701,152,'Iron',0.528);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(702,153,'Aluminium',0.423);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(703,153,'Iron',0.017);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(704,153,'sct::Torlon',0.006);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(705,153,'sct::PEI',0.2);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(706,153,'Carbon',0.005);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(707,153,'std::Kapton',0.04);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(708,153,'Copper',0.133);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(709,153,'sct::Epoxy',0.018);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(710,153,'sct::Silicone',0.011);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(711,153,'sct::CuKaptonEC',0.001);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(712,153,'sct::CN107',0.139);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(713,153,'sct::Solder',0.002);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(714,153,'std::Polyethylene',0.005);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(715,154,'Aluminium',0.315);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(716,154,'Copper',0.312);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(717,154,'sct::CN107',0.013);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(718,154,'Iron',0.269);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(719,154,'std::Kapton',0.058);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(720,154,'sct::Epoxy',0.026);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(721,154,'std::Polyethylene',0.007);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(270,51,'sct::AlNitride',0.2196);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(271,51,'sct::Elastosil',0.0078);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(272,51,'sct::Epoxy',0.0969);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(273,51,'Aluminium',0.0296);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(274,51,'sct::FR4',0.0898);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(275,51,'sct::Glass',0.2127);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(276,35,'sct::CN107',0.5);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(277,35,'sct::Solder',0.06);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(278,35,'std::C3F8',0.44);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(279,36,'sct::CN107',0.49);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(280,36,'sct::Solder',0.08);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(281,36,'std::C3F8',0.43);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(282,37,'sct::CN107',0.51);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(283,37,'sct::Solder',0.03);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(284,37,'std::C3F8',0.46);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(285,18,'Carbon',0.883);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(286,18,'std::Kapton',0.013);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(287,18,'sct::Epoxy',0.004);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(288,18,'Aluminium',0.02);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(289,18,'Nickel',0.016);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(290,18,'Gold',0.006);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(291,18,'Copper',0.008);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(292,18,'sct::Peek',0.05);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(293,34,'sct::Plastic',0.19);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(294,34,'Copper',0.46);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(295,34,'sct::FR4',0.21);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(296,34,'sct::CuOnAlWire',0.03);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(297,34,'std::Kapton',0.11);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(269,51,'Carbon',0.3436);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(722,155,'Aluminium',0.315);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(723,155,'Copper',0.312);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(724,155,'sct::CN107',0.013);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(725,155,'Iron',0.269);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(726,155,'std::Kapton',0.058);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(727,155,'sct::Epoxy',0.026);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(728,155,'std::Polyethylene',0.007);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(298,62,'sct::Epoxy',0.29);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(299,62,'Silver',0.71);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(300,59,'Carbon',18.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(301,59,'Hydrogen',20.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(302,59,'Oxygen',3.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(303,64,'Carbon',19.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(304,64,'Hydrogen',12.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(305,64,'Oxygen',3.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(306,58,'sct::Epoxy',1.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(307,60,'sct::Peek',1.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(308,23,'Carbon',5.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(309,23,'Hydrogen',8.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(310,55,'Tin',0.63);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(311,55,'Lead',0.37);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(312,65,'Beryllium',0.01);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(313,65,'Copper',0.99);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(314,2,'sct::Epoxy',0.3);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(315,2,'Carbon',0.7);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(316,30,'sct::Silica',0.81);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(317,30,'sct::B2O3',0.13);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(318,30,'sct::Na2O',0.04);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(319,30,'sct::Al2O3',0.02);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(320,72,'Boron',2.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(321,72,'Oxygen',3.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(322,73,'Sodium',2.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(323,73,'Oxygen',1.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(324,56,'sct::Epoxy',0.23);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(325,56,'sct::Glass',0.77);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(326,66,'sct::FR4',1.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(334,76,'Aluminium',0.1528);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(335,76,'Iron',0.057);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(336,76,'sct::Peek',0.5327);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(337,76,'sct::CFRP',0.2575);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(407,93,'Aluminium',0.0087);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(408,93,'sct::CFRP',0.9913);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(409,94,'Aluminium',0.0086);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(410,94,'sct::CFRP',0.9914);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(411,95,'Aluminium',0.007);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(412,95,'sct::CFRP',0.993);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(413,96,'Aluminium',0.0077);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(414,96,'sct::CFRP',0.9923);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(415,97,'sct::Peek',0.0511);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(416,97,'sct::CFRP',0.9472);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(417,97,'sct::CuAu',0.0017);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(418,98,'sct::Peek',0.0373);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(419,98,'sct::CFRP',0.9616);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(420,98,'sct::CuAu',0.0011);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(421,99,'sct::Peek',0.0311);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(422,99,'sct::CFRP',0.968);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(423,99,'sct::CuAu',0.0009);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(424,100,'sct::Peek',0.0273);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(425,100,'sct::CFRP',0.9719);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(426,100,'sct::CuAu',0.0008);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(427,101,'Aluminium',0.0311);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(428,101,'Copper',0.0347);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(429,101,'sct::Peek',0.6797);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(430,101,'sct::Glass',0.1296);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(431,101,'sct::Plastic',0.0807);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(432,101,'sct::BeCu',0.0442);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(433,102,'Aluminium',0.0297);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(434,102,'Copper',0.0322);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(435,102,'sct::Peek',0.6468);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(436,102,'sct::Glass',0.1537);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(437,102,'sct::Plastic',0.0957);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(438,102,'sct::BeCu',0.0419);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(439,103,'Aluminium',0.0313);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(440,103,'Copper',0.0295);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(441,103,'sct::Peek',0.6211);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(442,103,'sct::Glass',0.172);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(443,103,'sct::Plastic',0.1071);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(444,103,'sct::BeCu',0.039);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(445,104,'Aluminium',0.028);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(446,104,'Copper',0.0285);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(447,104,'sct::Peek',0.5872);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(448,104,'sct::Glass',0.1961);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(449,104,'sct::Plastic',0.1221);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(450,104,'sct::BeCu',0.0381);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(454,106,'Boron',1.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(455,106,'Nitrogen',1.0);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(456,107,'Epoxy',0.6925);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(457,107,'BoronNitride',0.3075);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(458,108,'Aluminium',0.0435);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(459,108,'Iron',0.0006);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(460,108,'Copper',0.0033);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(461,108,'sct::CN107',0.8631);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(462,108,'sct::Solder',0.0043);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(463,108,'std::Kapton',0.0019);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(464,108,'std::C3F8',0.083);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(465,108,'sct::ThermalGlue',0.0003);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(466,109,'Aluminium',0.043);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(467,109,'Iron',0.0006);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(468,109,'Copper',0.0032);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(469,109,'sct::CN107',0.8646);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(470,109,'sct::Solder',0.0043);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(471,109,'std::Kapton',0.0019);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(472,109,'std::C3F8',0.0821);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(473,109,'sct::ThermalGlue',0.0003);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(474,110,'Aluminium',0.0421);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(475,110,'Iron',0.0006);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(476,110,'Copper',0.0031);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(477,110,'sct::CN107',0.8675);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(478,110,'sct::Solder',0.0042);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(479,110,'std::Kapton',0.0018);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(480,110,'std::C3F8',0.0804);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(481,110,'sct::ThermalGlue',0.0003);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(482,111,'Aluminium',0.041);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(483,111,'Iron',0.0006);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(484,111,'Copper',0.003);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(485,111,'sct::CN107',0.871);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(486,111,'sct::Solder',0.0041);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(487,111,'std::Kapton',0.0018);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(488,111,'std::C3F8',0.0782);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(489,111,'sct::ThermalGlue',0.0003);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(3001,3000,'Silicon',0.004);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(3002,3000,'Iron',0.004);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(3003,3000,'Copper',0.001);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(3004,3000,'Manganese',0.007);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(3005,3000,'Magnesium',0.0445);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(3006,3000,'Zinc',0.0025);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(3007,3000,'Titanium',0.0015);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(3008,3000,'Chromium',0.0015);
INSERT INTO "SCTMATCOMPONENTS_DATA" VALUES(3009,3000,'Aluminium',0.934);
DROP TABLE IF EXISTS "SCTMATCOMPONENTS_DATA2TAG";
CREATE TABLE "SCTMATCOMPONENTS_DATA2TAG" ( "SCTMATCOMPONENTS_TAG_ID" SLONGLONG ,"SCTMATCOMPONENTS_DATA_ID" SLONGLONG  );
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,9);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,10);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,13);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,14);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,15);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,16);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,17);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,18);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,19);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,20);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,21);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,22);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,23);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,24);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,25);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,26);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,27);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,32);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,33);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,34);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,35);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,37);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,39);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,40);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,41);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,42);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,43);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,44);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,45);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,46);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,47);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,48);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,49);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,50);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,51);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,52);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,53);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,54);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,55);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,56);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,57);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,58);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,59);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,60);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,61);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,62);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,63);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,64);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,65);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,66);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,67);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,68);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,69);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,102);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,103);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,104);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,105);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,106);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,107);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,108);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,123);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,124);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,125);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,126);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,127);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,128);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,129);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,130);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,131);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,132);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,133);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,134);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,135);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,136);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,137);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,138);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,139);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,140);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,141);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,142);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,143);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,144);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,145);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,146);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,147);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,148);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,149);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,150);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,151);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,152);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,153);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,154);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,155);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,156);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,157);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,158);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,159);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,160);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,161);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,162);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,163);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,164);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,165);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,166);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,167);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,168);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,169);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,170);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,171);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,172);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,173);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,174);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,175);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,176);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,177);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,178);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,179);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,180);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,181);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,182);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,183);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,184);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,185);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,186);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,187);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,188);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,189);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,190);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,191);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,192);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,193);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,194);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,195);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,196);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,197);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,198);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,199);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,200);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,201);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,202);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,203);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,204);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,205);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,206);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,207);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,215);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,216);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,217);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,218);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,219);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,220);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,221);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,222);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,223);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,224);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,225);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,231);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,232);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,237);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,238);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,244);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,247);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,248);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,254);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,255);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,256);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,257);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,259);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,260);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,261);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,262);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,263);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,269);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,270);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,271);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,272);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,273);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,274);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,275);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,276);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,277);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,278);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,279);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,280);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,281);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,282);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,283);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,284);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,285);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,286);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,287);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,288);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,289);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,290);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,291);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,292);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,293);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,294);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,295);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,296);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,297);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,298);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,299);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,300);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,301);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,302);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,303);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,304);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,305);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,306);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,307);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,308);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,309);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,310);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,311);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,312);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,313);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,314);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,315);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,316);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,317);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,318);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,319);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,320);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,321);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,322);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,323);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,324);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,325);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,326);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,329);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,330);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,331);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,332);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,333);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,334);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,335);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,336);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,337);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,338);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,339);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,340);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,341);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,342);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,343);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,344);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,345);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,346);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,347);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,348);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,349);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,350);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,351);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,352);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,353);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,354);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,355);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,356);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,357);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,358);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,359);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,360);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,361);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,362);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,407);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,408);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,409);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,410);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,411);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,412);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,413);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,414);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,415);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,416);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,417);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,418);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,419);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,420);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,421);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,422);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,423);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,424);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,425);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,426);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,427);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,428);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,429);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,430);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,431);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,432);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,433);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,434);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,435);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,436);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,437);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,438);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,439);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,440);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,441);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,442);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,443);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,444);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,445);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,446);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,447);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,448);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,449);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,450);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,451);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,452);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,453);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,454);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,455);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,458);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,459);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,460);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,461);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,462);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,463);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,464);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,465);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,466);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,467);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,468);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,469);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,470);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,471);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,472);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,473);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,474);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,475);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,476);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,477);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,478);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,479);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,480);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,481);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,482);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,483);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,484);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,485);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,486);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,487);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,488);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,489);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,490);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,491);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,492);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,493);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,494);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,495);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,496);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,497);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,498);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,499);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,500);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,501);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,502);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,503);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,504);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,505);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,506);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,507);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,508);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,509);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,510);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,511);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,512);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,513);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,514);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,515);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,516);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,517);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,518);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,519);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,520);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,521);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,522);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,523);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,524);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,533);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,534);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,535);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,536);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,537);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,538);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,540);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,541);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,542);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,543);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,544);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,545);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,546);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,547);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,548);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,549);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,551);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,552);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,553);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,554);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,558);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,559);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,560);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,561);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,562);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,563);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,564);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,565);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,566);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,567);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,569);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,571);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,572);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,573);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,574);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,575);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,576);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,577);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,578);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,579);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,580);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,581);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,582);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,583);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,584);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,585);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,586);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,587);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,588);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,589);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,604);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,615);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,616);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,617);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,618);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,619);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,620);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,621);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,622);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,623);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,624);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,625);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,626);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,627);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,628);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,629);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,631);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,632);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,633);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,641);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,642);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,643);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,644);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,645);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,646);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,648);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,649);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,650);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,662);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,663);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,664);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,665);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,666);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,667);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,668);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,669);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,670);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,671);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,672);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,674);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,676);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,677);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,678);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,679);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,680);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,681);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,167);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,168);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,169);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,682);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,683);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,684);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,685);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,692);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,693);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,694);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,695);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,696);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,697);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,698);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,699);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,700);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,701);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,729);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,730);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,731);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,732);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,733);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,734);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,735);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,736);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,737);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,738);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,739);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,740);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,741);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,742);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,744);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,745);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,746);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,747);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,748);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,749);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,750);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,751);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,752);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,753);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,754);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,755);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,756);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,757);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,758);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,759);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,760);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,768);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,769);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,770);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,771);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,772);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,773);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,774);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,775);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,776);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,777);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,778);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,779);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,780);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,781);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,782);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,783);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,784);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,785);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,786);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,787);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,788);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,789);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,790);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,791);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,792);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,793);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,794);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,795);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,796);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,797);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,798);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,799);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,800);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,801);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,802);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,803);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,804);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,805);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,806);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,807);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,808);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,809);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,810);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,811);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,812);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,813);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,814);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,815);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,816);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,817);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,818);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,819);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,820);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,821);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,822);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,823);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,824);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,825);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,826);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,827);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,828);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,829);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,830);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,831);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,832);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,833);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,834);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,835);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,836);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,837);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,838);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,844);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,845);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,846);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,849);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,850);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,851);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,852);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,853);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,854);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,855);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,856);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,857);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,858);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,859);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,860);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,861);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,862);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,863);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,864);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,865);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,866);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,867);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,868);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,869);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,870);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,871);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,872);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,873);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,874);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,875);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,876);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,877);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,878);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,879);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,880);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,881);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,882);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,883);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,884);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,885);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,886);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,887);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,888);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,889);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,890);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,891);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,892);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,893);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,894);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,895);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,896);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,897);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,898);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,899);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,900);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,901);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,902);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,903);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,904);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,905);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,906);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,907);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,908);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,909);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,910);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,911);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,912);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,913);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,914);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,915);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,916);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,917);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,918);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,919);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,920);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,921);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,922);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,923);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,924);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,925);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,926);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,927);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,928);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,929);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,930);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,931);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,932);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,933);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,934);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,935);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,936);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,937);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,938);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,939);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,940);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,941);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,942);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,943);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,944);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,945);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,946);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,947);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,948);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,949);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,950);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,951);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,952);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,953);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,954);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,955);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,956);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,957);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,958);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,959);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,960);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,3001);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,3002);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,3003);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,3004);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,3005);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,3006);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,3007);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,3008);
INSERT INTO "SCTMATCOMPONENTS_DATA2TAG" VALUES(107778,3009);
DROP TABLE IF EXISTS "SCTMATERIALS_DATA";
CREATE TABLE "SCTMATERIALS_DATA" ( "SCTMATERIALS_DATA_ID" SLONGLONG ,"NAME" TEXT ,"DENSITY" DOUBLE , UNIQUE ( "SCTMATERIALS_DATA_ID" ) );
INSERT INTO "SCTMATERIALS_DATA" VALUES(34,'ModuleConnBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(35,'DiscCoolOutBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(36,'DiscCoolMidBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(37,'DiscCoolInnBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(38,'CoolBlkMHiBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(39,'CoolBlkMLoBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(40,'CoolBlkSHiBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(41,'CoolBlkSLoBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(42,'FSIBHBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(43,'FSIBLBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(44,'FSIFHBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(45,'FSIFLBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(46,'PPF0eBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(47,'PPF0oBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(48,'PPF0cBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(49,'FwdSpineOutBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(50,'FwdSpineMidBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(51,'FwdSpineInnBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(52,'FwdHybridBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(124,'DiscFixBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(188,'FwdCryCool2Base',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(0,'Hybrid',3.97);
INSERT INTO "SCTMATERIALS_DATA" VALUES(1,'TPG',2.247);
INSERT INTO "SCTMATERIALS_DATA" VALUES(2,'CFRP',1.78);
INSERT INTO "SCTMATERIALS_DATA" VALUES(3,'Bracket',1.48);
INSERT INTO "SCTMATERIALS_DATA" VALUES(4,'CopperKapton',5.19);
INSERT INTO "SCTMATERIALS_DATA" VALUES(5,'CoolingBlock',3.25);
INSERT INTO "SCTMATERIALS_DATA" VALUES(6,'AluminiumKapton',1.7);
INSERT INTO "SCTMATERIALS_DATA" VALUES(7,'CN107',8.95);
INSERT INTO "SCTMATERIALS_DATA" VALUES(8,'CoolingPipe',1.83);
INSERT INTO "SCTMATERIALS_DATA" VALUES(9,'CFiberSupport',0.189);
INSERT INTO "SCTMATERIALS_DATA" VALUES(10,'Flange',1.6);
INSERT INTO "SCTMATERIALS_DATA" VALUES(11,'CFiberInterLink',1.6);
INSERT INTO "SCTMATERIALS_DATA" VALUES(12,'AirexR82',0.08);
INSERT INTO "SCTMATERIALS_DATA" VALUES(13,'TSC',1.97);
INSERT INTO "SCTMATERIALS_DATA" VALUES(14,'TSE',0.795);
INSERT INTO "SCTMATERIALS_DATA" VALUES(15,'TSP',0.569);
INSERT INTO "SCTMATERIALS_DATA" VALUES(16,'EMI',1.97);
INSERT INTO "SCTMATERIALS_DATA" VALUES(17,'FwdHybrid',1.59);
INSERT INTO "SCTMATERIALS_DATA" VALUES(18,'DiscSupport',0.1265);
INSERT INTO "SCTMATERIALS_DATA" VALUES(19,'CFiberFwdSupportFrame',0.214);
INSERT INTO "SCTMATERIALS_DATA" VALUES(20,'FwdCoolingPipe',1.52);
INSERT INTO "SCTMATERIALS_DATA" VALUES(21,'KaptonAl',1.6);
INSERT INTO "SCTMATERIALS_DATA" VALUES(22,'Foam',0.04377);
INSERT INTO "SCTMATERIALS_DATA" VALUES(23,'Rubber',2.982);
INSERT INTO "SCTMATERIALS_DATA" VALUES(77,'PEI',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(78,'Adhesive',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(79,'Dogleg',5.3051);
INSERT INTO "SCTMATERIALS_DATA" VALUES(122,'FwdRailBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(125,'SpiderBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(128,'EMIJointBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(148,'Brass',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(149,'InterlinkBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(157,'PPB1Tmp',1.8064);
INSERT INTO "SCTMATERIALS_DATA" VALUES(158,'BrlCryoServTmp',0.7742);
INSERT INTO "SCTMATERIALS_DATA" VALUES(184,'FSIFlangeBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(185,'FSIFibMaskBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(186,'FSIEndJewelBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(187,'FSIScorpionBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(116,'FwdCoolPipeBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(117,'FwdLMTBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(118,'FwdLMTCoolBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(119,'FwdFibresBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(120,'FwdNPipeBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(121,'FwdNPipeBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(123,'FwdSupportBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(138,'CuKaptonEC',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(139,'FwdOTEBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(140,'FwdITEBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(141,'FwdShShieldBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(142,'FwdFrontSupBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(143,'FwdFlangeFOBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(144,'FwdFlangeFIBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(145,'FwdRearSupBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(146,'FwdFlangeROBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(147,'FwdFlangeRIBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(171,'C6F14',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(172,'Polycarbonate',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(173,'Polyester',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(174,'FwdCryoCoolBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(175,'FwdCryoLMTBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(176,'FwdCCTBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(177,'FwdHEXConnBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(178,'FwdRadHEXBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(179,'FwdRadLMTBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(180,'FwdRadDryNBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(181,'FwdSTFTBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(182,'FwdRCTBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(64,'Peek',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(65,'BeCu',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(66,'PCB',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(67,'Vectra',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(68,'Torlon',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(69,'Ceramic',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(70,'Grease',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(71,'CuOnAlWire',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(55,'Solder',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(183,'BearingBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(74,'Silicone',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(75,'CoolingBlock',5.0107);
INSERT INTO "SCTMATERIALS_DATA" VALUES(105,'CoolingPipeBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(127,'EMIBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(137,'HSPBase',1.41104);
INSERT INTO "SCTMATERIALS_DATA" VALUES(61,'EotiteP102',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(24,'BrlHybrid',2.241);
INSERT INTO "SCTMATERIALS_DATA" VALUES(25,'BrlBaseBoard',2.133);
INSERT INTO "SCTMATERIALS_DATA" VALUES(26,'PigTail',1.135);
INSERT INTO "SCTMATERIALS_DATA" VALUES(27,'BrlBracket',3.03);
INSERT INTO "SCTMATERIALS_DATA" VALUES(28,'Dogleg',6.44);
INSERT INTO "SCTMATERIALS_DATA" VALUES(29,'AlKapton',1.712);
INSERT INTO "SCTMATERIALS_DATA" VALUES(30,'Glass',2.51);
INSERT INTO "SCTMATERIALS_DATA" VALUES(31,'CuKapton',2.94);
INSERT INTO "SCTMATERIALS_DATA" VALUES(32,'AlNitride',3.3);
INSERT INTO "SCTMATERIALS_DATA" VALUES(33,'DiscPowerTape',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(115,'PPCoolingBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(126,'EMIJoint',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(53,'OptoHarness',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(54,'Silica',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(56,'FR4',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(57,'Al2O3',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(58,'Araldite',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(59,'Epoxy',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(60,'Plastic',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(114,'PPConnectorBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(62,'AIT',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(63,'Elastosil',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(129,'Airex',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(130,'Techsil',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(131,'AlKapton',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(132,'Pyrogel',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(133,'EPDM',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(134,'TSCylinderBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(135,'TSBulkheadBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(136,'TSEndPanelBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(80,'CuAu',9.18);
INSERT INTO "SCTMATERIALS_DATA" VALUES(81,'SupportCyl0',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(82,'SupportCyl1',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(83,'SupportCyl2',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(84,'SupportCyl3',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(85,'Flange0',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(86,'Flange1',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(87,'Flange2',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(88,'Flange3',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(89,'Clamp0',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(90,'Clamp1',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(91,'Clamp2',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(92,'Clamp3',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(156,'FwdOTEFlngBase',0.34);
INSERT INTO "SCTMATERIALS_DATA" VALUES(159,'HEXInnerBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(160,'HEXMiddleBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(161,'HEXOuterBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(162,'Polyimide',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(163,'PPB1BBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(164,'PPB1ADBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(165,'PPB1A1CBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(166,'PPB1EFEGBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(167,'CabTrayBBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(168,'CabTrayADBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(169,'CabTrayA1CBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(170,'CabTrayEFEGBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(112,'HarnessBase',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(113,'PowerTape',1.4912);
INSERT INTO "SCTMATERIALS_DATA" VALUES(150,'FwdFrntWingBase',0.33);
INSERT INTO "SCTMATERIALS_DATA" VALUES(151,'FwdRearWingBase',0.24);
INSERT INTO "SCTMATERIALS_DATA" VALUES(152,'FwdMechBase',6.54);
INSERT INTO "SCTMATERIALS_DATA" VALUES(153,'FwdRadServBase',0.815);
INSERT INTO "SCTMATERIALS_DATA" VALUES(154,'FwdRadServBase',2.107);
INSERT INTO "SCTMATERIALS_DATA" VALUES(155,'FwdCryoServBase',2.107);
INSERT INTO "SCTMATERIALS_DATA" VALUES(72,'B2O3',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(73,'Na2O',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(76,'BrlBracket',1.6177);
INSERT INTO "SCTMATERIALS_DATA" VALUES(93,'SupportCyl0Base',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(94,'SupportCyl1Base',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(95,'SupportCyl2Base',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(96,'SupportCyl3Base',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(97,'Flange0Base',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(98,'Flange1Base',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(99,'Flange2Base',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(100,'Flange3Base',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(101,'Clamp0Base',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(102,'Clamp1Base',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(103,'Clamp2Base',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(104,'Clamp3Base',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(106,'BoronNitride',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(107,'ThermalGlue',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(108,'CoolingEnd0Base',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(109,'CoolingEnd1Base',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(110,'CoolingEnd2Base',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(111,'CoolingEnd3Base',1.0);
INSERT INTO "SCTMATERIALS_DATA" VALUES(3000, 'AW5083', 2.650);
DROP TABLE IF EXISTS "SCTMATERIALS_DATA2TAG";
CREATE TABLE "SCTMATERIALS_DATA2TAG" ( "SCTMATERIALS_TAG_ID" SLONGLONG ,"SCTMATERIALS_DATA_ID" SLONGLONG  );
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,2);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,7);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,9);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,10);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,11);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,12);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,13);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,14);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,15);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,18);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,19);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,20);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,22);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,23);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,24);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,25);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,26);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,30);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,31);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,32);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,33);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,34);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,35);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,36);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,37);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,38);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,39);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,40);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,41);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,42);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,43);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,44);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,45);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,46);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,47);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,48);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,49);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,50);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,51);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,52);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,53);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,54);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,55);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,56);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,57);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,58);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,59);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,60);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,61);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,62);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,63);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,64);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,65);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,66);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,67);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,68);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,69);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,70);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,71);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,72);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,73);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,74);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,75);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,76);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,77);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,78);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,79);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,80);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,93);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,94);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,95);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,96);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,97);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,98);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,99);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,100);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,101);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,102);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,103);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,104);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,105);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,106);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,107);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,108);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,109);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,110);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,111);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,112);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,113);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,114);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,115);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,116);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,117);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,118);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,119);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,120);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,122);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,123);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,124);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,125);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,127);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,128);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,129);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,130);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,131);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,132);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,133);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,134);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,135);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,136);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,137);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,138);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,139);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,140);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,141);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,142);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,143);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,144);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,145);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,146);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,147);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,148);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,149);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,150);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,151);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,152);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,153);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,155);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,156);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,157);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,158);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,159);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,160);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,161);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,162);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,163);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,164);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,165);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,166);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,167);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,168);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,169);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,170);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,171);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,172);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,173);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,174);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,175);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,176);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,177);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,178);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,179);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,180);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,181);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,182);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,183);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,184);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,185);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,186);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777,187);
INSERT INTO "SCTMATERIALS_DATA2TAG" VALUES(107777, 3000);
--
--
COMMIT;
